#include <QDebug>
#include <QMessageBox>
#include <math.h>

#include "camera.h"
#include "videocamera.h"
#include "videofile.h"
#include "image.h"
#include "cameracontrols.h"

#ifdef Q_OS_LINUX
#include "video1394.h"
#include "sys1394.h"
#endif

#include "cameramanager.h"
#include "imageprocessing.h"
#include "calibrationdialog.h"
#include "videosource.h"
#include "framewriter.h"
#include <string.h>

Camera::Camera(QString name)
{
    cameraMatrix = cvCreateMat(3, 4, CV_32FC1);
    distortionCoeff = cvCreateMat(5, 1, CV_32F);
    intrinsicMatrix = cvCreateMat(3, 3, CV_32F);
    calibrationDialog = NULL;
    cvZero(cameraMatrix);
    videoInitialized = false;
    processNextFrame = false;
    doFinishThread = false;
    intrinsicParamsSet = false;
    firstFrameCaptured = false;
    setCheckImageActive = false;
    cameraManager = NULL;
    checkImage = cvLoadImage("check-image.png", 1);
    rotatedCamera = false;
    frameRotated = NULL;

    #ifdef Q_OS_LINUX
    sys1394 = NULL;
    #endif

    this->name = name;

    imageProcessing = new ImageProcessing(this, QSize(1,1), videoSourceId);

    frameWriter.setBaseName("/home/hornang/video/" + name + "/");
    cameraControls = new CameraControls(this);

    connect(this, SIGNAL(videoInfoChanged()), cameraControls, SLOT(updateVideoInfo()));
    connect(imageProcessing, SIGNAL(trackersChanged()), cameraControls, SLOT(updateTrackerList()));

}

/**
  * Sets the camera manager. Must be set properly such that this thread can tell when
  * each frame is finished processing.
  */
void Camera::setCameraManager(CameraManager * cameraManager)
{
    this->cameraManager = cameraManager;
}

/**
  * Rotate the camera picture 90 degress counter clockwise.
  */
void Camera::setRotatedCamera()
{
    rotatedCamera = true;
}

/**
  * Sets intrinsic parameters for compensating for lens distortion only.
  *
  * @param intrinsic OpenCV intrinsic matrix
  * @param distortion OpenCV distortion parameters.
  */
void Camera::setIntrinsic(CvMat * intrinsic, CvMat * distortion)
{
    if (intrinsic == NULL) {
        qDebug() << "intrinsic argument is NULL";
        return;
    } else {
        if (intrinsic->cols != 3 || intrinsic->rows != 3) {
            qDebug() << "intrinsic matrix does not have 3x3 dimension";
            return;
        }
    }

    cvCopy(distortion, distortionCoeff);
    cvCopy(intrinsic, intrinsicMatrix);
    intrinsicParamsSet = true;
}

/**
  * Tells this camera thread to aquire and process the next frame.
  */
void Camera::processFrame()
{
    mutex.lock();
    processNextFrame = true;
    mutex.unlock();
    condition.wakeOne();
}

/**
  * Tells this thread to return as soon as possible.
  */
void Camera::finishThread()
{
    mutex.lock();
    doFinishThread = true;
    mutex.unlock();
    condition.wakeOne();
    wait();

}

/**
  * Thread execution. Captures video and processes it.
  */
void Camera::run()
{
    forever {

        mutex.lock();

        /* If neither processNextFrame or doFinishThread, then just wait */
        if (!processNextFrame && !doFinishThread) {
            condition.wait(&mutex);
        }

        /* Check if we want to exit thread */
        if (doFinishThread) {
            doFinishThread = false;
            processNextFrame = false;
            mutex.unlock();
            qDebug() << "Camera thread " << name << " finished";
            return;
        }
        processNextFrame = false;
        mutex.unlock();

        if (!firstFrameCaptured) {
            if (!videoSource->captureFrame()) {
                qDebug() << "Could not grab frame";
                //videoSource->close();
                videoSource->init();
                videoSource->captureFrame();
            }
            firstFrameCaptured = false;
        }
        IplImage * newFrame = videoSource->getFrame();
        // frameWriter.writeFrame(newFrame);

        if (newFrame != NULL) {
            if (newFrame->nChannels == 3) {
                if (intrinsicParamsSet) {
                    cvRemap(newFrame, frame, undistortMapX, undistortMapY);
                } else {
                    cvCopy(newFrame, frame);
                }

            } else {
                // Convert to internal BGR format
                if (intrinsicParamsSet) {
                    cvRemap(newFrame, frameGrayUndistorted, undistortMapX, undistortMapY);
                    cvCvtColor(frameGrayUndistorted, frame, CV_GRAY2BGR);
                } else {
                    cvCvtColor(newFrame, frame, CV_GRAY2BGR);
                }
            }
        }

        imageProcessing->processFrame(frame);

        if (cameraManager != NULL) {
            cameraManager->frameProcessed(this->name);
        }
    }
}

/**
  * Initializes video source and allocates memory.
  *
  * @return True if successful opened video.
  */
bool Camera::initVideo()
{
    if (videoInitialized) {
        cvReleaseImage(&frame);
        videoSource->close();
        videoInitialized = false;
    }

    if (videoSourceType == OPENCV_CAMERA) {
        videoSource = new VideoCamera(videoSourceId);
    } else if (videoSourceType == VIDEO_FILE) {
        videoSource = new VideoFile(videoSourceId);
    } else if (videoSourceType == IMAGE) {
        videoSource = new Image(videoSourceId);
    }
    #ifdef Q_OS_LINUX
    else if (videoSourceType == DC1394) {
        if (sys1394 == NULL) {
            qDebug() << "IEEE1394 subsystem not set, cannot open camera";
            return false;
        }

        bool ok = false;
        qlonglong id = videoSourceId.toLongLong(&ok);
        if (!ok) {
            qDebug() << "Failed to convert videoSourceId to appropriate integer.";
            return false;
        }
        Video1394 * camera = sys1394->openCamera(id);
        if (camera == NULL) {
            return false;
        }
        if (rotatedCamera) {
            camera->setRotated();
        }
        videoSource = camera;
    }
    #endif
    else {
        return false;
    }

    if (videoSource->init() == false) {
        qDebug() << "FAILED initialize video";
        delete videoSource;
        return false;
    }


    QSize videoSize = videoSource->getVideoSize();
    frameWindow = QRect(0, 0, videoSize.width(), videoSize.height());

    frame = cvCreateImage(cvSize(videoSize.width(), videoSize.height()), IPL_DEPTH_8U, 3);
    frameGrayUndistorted = cvCreateImage(cvSize(videoSize.width(), videoSize.height()), IPL_DEPTH_8U, 1);
    cvSet(frame, cvScalar(255, 0, 0));

    if (intrinsicParamsSet) {
        undistortMapX = cvCreateImage(cvSize(videoSize.width(), videoSize.height()), IPL_DEPTH_32F, 1);
        undistortMapY = cvCreateImage(cvSize(videoSize.width(), videoSize.height()), IPL_DEPTH_32F, 1);
        qDebug() << "initing undistortmap with intrinsic matrix:";
        printMatrix(intrinsicMatrix);
        qDebug() << "and distortion coeff:";
        printMatrix(distortionCoeff);
        cvInitUndistortMap(intrinsicMatrix, distortionCoeff, undistortMapX, undistortMapY);
    }

    imageProcessing->setVideoSize(videoSize);
    imageProcessing->setCameraMatrix(cameraMatrix);
    videoInitialized = true;
    return true;
}

/**
  * @return The bounding rectangle for the frame.
  */
QRect Camera::getFrameWindow()
{
    return frameWindow;
}

/**
  * Sets the check image from a region in the current video frame.
  *
  * @param region The rectangle in the video to capture the check image.
  */
void Camera::setCheckImage(QRect region)
{
    if (checkImage != NULL)    {
        cvReleaseImage(&checkImage);
    }

    mutex.lock();
    cvSetImageROI(frame, cvRect(region.left(), region.top(), region.width(), region.height()));
    checkImage = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 3);
    cvCopy(frame, checkImage);
    cvNamedWindow("checkImage");
    cvShowImage("checkImage", checkImage);
    cvResetImageROI(frame);
    cvSaveImage("check-image.png", checkImage);
    mutex.unlock();
}

/**
  * Returns a pointer to the check image which is used for setting the
  * body origin appropriate with the use of template match.
  *
  * @return A pointer to the check image.
  */
IplImage * Camera::getCheckImage()
{
    return checkImage;
}

/**
  * Selects which video source to use.
  *
  * @param videoSourceType An enum defining the type of video source.
  * @param videoSourceId An id describing the video source. Filename for file capture,
  * Cam id for firewire capture etc.
  */
void Camera::setVideoSource(VideoSourceType videoSourceType, QString videoSourceId)
{
    this->videoSourceType = videoSourceType;
    this->videoSourceId = videoSourceId;

    QString typeStr("Video source: ");
    if (videoSourceType == OPENCV_CAMERA) {
        typeStr.append("OpenCV Capture id ");
        typeStr.append(QString("\"") + QString(videoSourceId) + QString("\""));
    } else if (videoSourceType == VIDEO_FILE) {
        typeStr.append("Video file ");
        typeStr.append(QString("\"") + QString(videoSourceId) + QString("\""));
    } else if (videoSourceType == IMAGE) {
        typeStr.append("Image ");
        typeStr.append(QString("\"") + QString(videoSourceId) + QString("\""));
    }
#ifdef Q_OS_LINUX
    else if (videoSourceType == DC1394) {
        typeStr.append("IEEE 1394 cam with GUID: ");
        typeStr.append(QString("\"") + QString(videoSourceId) + QString("\""));

    }
#endif
    else {
        return;
    }
    emit videoInfoChanged();
}

/**
  * @param A string describing the video source.
  */
QString Camera::getVideoSourceDescription()
{
    QString typeStr;

    if (videoSourceType == OPENCV_CAMERA) {
        typeStr.append("OpenCV Capture id ");
        typeStr.append(QString("\"") + QString(videoSourceId) + QString("\""));
    } else if (videoSourceType == VIDEO_FILE) {
        typeStr.append("Video file ");
        typeStr.append(QString("\"") + QString(videoSourceId) + QString("\""));
    } else if (videoSourceType == IMAGE) {
        typeStr.append("Image ");
        typeStr.append(QString("\"") + QString(videoSourceId) + QString("\""));
    }
#ifdef Q_OS_LINUX
    else if (videoSourceType == DC1394) {
        typeStr.append("DC1394 ");
        typeStr.append(QString("\"") + QString(videoSourceId) + QString("\""));
    }
#endif
    return typeStr;
}

/**
  * Called when a point in the video i selected
  *
  * @param point The selected point.
  */
void Camera::selectPoint(QPoint point)
{
    emit pointSelected(point);

    qDebug() << "point selected " << point;
    if (calibrationDialog != NULL) {
        calibrationDialog->pointSelected(point);
    }

}

Camera::~Camera()
{
    if (isRunning()) {
        finishThread();
    }
    cvReleaseMat(&cameraMatrix);
    if (videoInitialized) {
        videoSource->close();
        delete videoSource;
    }
    delete cameraControls;
    delete imageProcessing;
}

/**
  * @return The name of the camera.
  */
QString Camera::getName()
{
    return name;
}



/**
  * @return A pointer to the current video frame.
  */
IplImage * Camera::getFrame()
{
    return frame;
}

/**
  * Shows the calibration dialog.
  */
void Camera::showCalibrationDialog()
{
    if (calibrationDialog == NULL) {
        calibrationDialog = new CalibrationDialog();
    }

    calibrationDialog->show();
    connect(calibrationDialog, SIGNAL(finished(int)), this, SLOT(calibrationDialogClosed(int)));
}

/**
  * Returns a pointer to the image processing object.
  *
  * @return The pointer.
  */
ImageProcessing * Camera::getImageProcessing()
{
    return imageProcessing;
}

/**
  * Called when the calibration dialog is closed. Computes new camera matrix if
  * the dialog was accepted. Typically means that "OK" was pressed.
  *
  * @param result How the dialog was closed.
  */
void Camera::calibrationDialogClosed(int result)
{
    disconnect(calibrationDialog, SIGNAL(finished(int)), this, SLOT(calibrationDialogClosed(int)));

    if (result != QDialog::Accepted) {
        return;
    }

    QList <CalibrationPoint> * points;
    points = calibrationDialog->getCalibrationPoints();

    //delete calibrationDialog;
    calibrationDialog = NULL;

    int pointsCount = points->length();

    CvMat * L = cvCreateMat(pointsCount * 2, 11, CV_32FC1);
    CvMat * d = cvCreateMat(pointsCount * 2, 1, CV_32FC1);

    for(int i = 0; i < pointsCount; i++) {
        cvSetReal2D(L, i * 2, 0, points->at(i).x);
        cvSetReal2D(L, i * 2, 1, points->at(i).y);
        cvSetReal2D(L, i * 2, 2, points->at(i).z);
        cvSetReal2D(L, i * 2, 3, 1);
        cvSetReal2D(L, i * 2, 4, 0);
        cvSetReal2D(L, i * 2, 5, 0);
        cvSetReal2D(L, i * 2, 6, 0);
        cvSetReal2D(L, i * 2, 7, 0);
        cvSetReal2D(L, i * 2, 8, -points->at(i).u * points->at(i).x);
        cvSetReal2D(L, i * 2, 9, -points->at(i).u * points->at(i).y);
        cvSetReal2D(L, i * 2, 10, -points->at(i).u * points->at(i).z);
        cvSetReal2D(L, i * 2 + 1, 0, 0);
        cvSetReal2D(L, i * 2 + 1, 1, 0);
        cvSetReal2D(L, i * 2 + 1, 2, 0);
        cvSetReal2D(L, i * 2 + 1, 3, 0);
        cvSetReal2D(L, i * 2 + 1, 4, points->at(i).x);
        cvSetReal2D(L, i * 2 + 1, 5, points->at(i).y);
        cvSetReal2D(L, i * 2 + 1, 6, points->at(i).z);
        cvSetReal2D(L, i * 2 + 1, 7, 1);
        cvSetReal2D(L, i * 2 + 1, 8, -points->at(i).v * points->at(i).x);
        cvSetReal2D(L, i * 2 + 1, 9, -points->at(i).v * points->at(i).y);
        cvSetReal2D(L, i * 2 + 1, 10, -points->at(i).v * points->at(i).z);

        cvSetReal2D(d, i * 2, 0, points->at(i).u);
        cvSetReal2D(d, i * 2 + 1, 0, points->at(i).v);
    }

    CvMat * C = cvCreateMat(11, 1, CV_32FC1);
    if (!cvSolve(L, d, C)) {
        QMessageBox msgBox;
        msgBox.setText("Camera Matrix computation failed: X in least square is singular");
        msgBox.exec();
        return;
    }

    cvSetReal2D(cameraMatrix, 0, 0, cvGetReal2D(C, 0, 0));
    cvSetReal2D(cameraMatrix, 1, 0, cvGetReal2D(C, 4, 0));
    cvSetReal2D(cameraMatrix, 2, 0, cvGetReal2D(C, 8, 0));
    cvSetReal2D(cameraMatrix, 0, 1, cvGetReal2D(C, 1, 0));
    cvSetReal2D(cameraMatrix, 1, 1, cvGetReal2D(C, 5, 0));
    cvSetReal2D(cameraMatrix, 2, 1, cvGetReal2D(C, 9, 0));
    cvSetReal2D(cameraMatrix, 0, 2, cvGetReal2D(C, 2, 0));
    cvSetReal2D(cameraMatrix, 1, 2, cvGetReal2D(C, 6, 0));
    cvSetReal2D(cameraMatrix, 2, 2, cvGetReal2D(C, 10, 0));
    cvSetReal2D(cameraMatrix, 0, 3, cvGetReal2D(C, 3, 0));
    cvSetReal2D(cameraMatrix, 1, 3, cvGetReal2D(C, 7, 0));
    cvSetReal2D(cameraMatrix, 2, 3, 1);

    cvReleaseMat(&C);
    setCameraMatrix(cameraMatrix);
    printMatrix(cameraMatrix);

}

/**
  * Finds the pixel coordinates of a 3D point
  */
CvPoint Camera::mapToImage(CvPoint3D32f point)
{
    CvMat * world = cvCreateMat(4, 1, CV_32F);

    world->data.fl[0] = point.x;
    world->data.fl[1] = point.y;
    world->data.fl[2] = point.z;
    world->data.fl[3] = 1;

    CvMat * image = cvCreateMat(3, 1, CV_32F);
    cvGEMM(cameraMatrix, world, 1.0f, NULL, 1.0f, image, 0);

    float u = image->data.fl[0] / image->data.fl[2];
    float v = image->data.fl[1] / image->data.fl[2];

    CvPoint p = cvPoint(u, v);

    cvReleaseMat(&world);
    cvReleaseMat(&image);

    return p;
}

void Camera::setCameraMatrix(CvMat * newCameraMatrix)
{
    cvCopy(newCameraMatrix, cameraMatrix);
    if (imageProcessing != NULL) {
        imageProcessing->setCameraMatrix(cameraMatrix);
    } else {
        qDebug() << "image processing is NULL";
    }
    emit cameraMatrixChanged(cameraMatrix);
}

CvMat * Camera::getCameraMatrix()
{
    return cameraMatrix;
}

void Camera::printMatrix(CvMat * L)
{

    int rows = L->rows;
    int cols = L->cols;

    QString output;
    output.append("[");
    for(int i = 0; i < rows; i++) {
        for(int j = 0; j < cols; j++) {
            QString num;
            num.setNum(cvGetReal2D(L, i, j));
            output.append(num);
            output.append("\t");
        }
        output.append(";\n");
    }
    output.append("]");
    qDebug() << output;
}

void Camera::setCheckImageMode(bool mode)
{
    setCheckImageActive = mode;
    qDebug() << "setCheckImageMode set to " << mode;
}

void Camera::rectangleSelected(QRect rect)
{
    if (!setCheckImageActive) {
        imageProcessing->addTracker(rect);
    } else {
        setCheckImage(rect);
    }
}


void Camera::findTemplate(IplImage * templ)
{
    //QRect rect = imageProcessing->templateMatch(templ);
    //this->setTargetArea(rect);
}

void Camera::setMousePosition(QPoint pos)
{
    QString x,y;
    x.setNum(pos.x());
    y.setNum(pos.y());
    //m_ui->mousePosition->setText(QString("(") + x + QString(",") + y + QString(")"));

}

void Camera::vesselFound(QPolygon poly)
{
    for(int i = 0; i < poly.size(); i++) {
        QPoint point = poly.point(i);
        QRect targetRect(point.x() - 20, point.y() - 20, 40, 40);
        qDebug() << "targetRect is " << targetRect;
        imageProcessing->addTracker(targetRect);
    }

    //imageProcessing->setVesselSearch(false);
}

CameraControls * Camera::getCameraControls()
{
    return cameraControls;
}

#ifdef Q_OS_LINUX
void Camera::set1394sys(Sys1394 * sys1394)
{
    this->sys1394 = sys1394;
}
#endif

void Camera::loadCameraMatrix(QString fileName)
{
    try {
        CvMat * newCameraMatrix = (CvMat*) cvLoad(fileName.toAscii().data());
        if (newCameraMatrix == NULL) {
            int ret = QMessageBox::critical(cameraControls, tr("Shiptrack"), tr("Could not open file ") + fileName, QMessageBox::Ok);
            return;
        }
        setCameraMatrix(newCameraMatrix);
    } catch (cv::Exception e) {
        int ret = QMessageBox::critical(cameraControls, tr("Shiptrack"), QString(e.err.c_str()), QMessageBox::Ok);
    }
}

