#ifndef Camera_H
#define Camera_H

#include <QtGui/QWidgetItem>
#include <QGraphicsScene>
#include <QModelIndex>
#include <QListWidget>
#include <QString>
#include <QMutex>
#include <QWaitCondition>
#include <QThread>

#include <cv.h>

#include "videosource.h"
#include "framewriter.h"

class CalibrationDialog;
class VideoSource;
class FrameWriter;
class ImageProcessing;
class VideoCamera;
class VideoFile;
class Image;
class CameraControls;
#ifdef Q_OS_LINUX
class Sys1394;
#endif
class CameraManager;


class Camera : public QThread {
    Q_OBJECT
public:
    Camera(QString name);
    void run();
    ImageProcessing * getImageProcessing();
    ~Camera();
    bool doTracking;
    void setCameraMatrix(CvMat *);
    CvPoint mapToImage(CvPoint3D32f);
    void setVideoSource(VideoSourceType, QString);
    void processFrame();
    bool initVideo();
    CvMat * getCameraMatrix();
    IplImage * getFrame();
    QString getName();
    CameraControls * getCameraControls();
    QString getVideoSourceDescription();
    void finishThread();
    #ifdef Q_OS_LINUX
    void set1394sys(Sys1394 *);
    #endif
    QRect getFrameWindow();
    void setIntrinsic(CvMat * intrinsic, CvMat * distortion);
    void setCheckImage(QRect region);
    void setCheckImageMode(bool);
    IplImage * getCheckImage();
    void setRotatedCamera();
    void setCameraManager(CameraManager * cameraManager);
    void loadCameraMatrix(QString fileName);

protected:
    void changeEvent(QEvent *e);

private:
    bool init();
    void printMatrix(CvMat *);

    /* A pointer to the sys1394 subsystem */
    #ifdef Q_OS_LINUX
    Sys1394 * sys1394;
    #endif

    /* Video frame. Always 3 channel */
    IplImage * frame;

    /* Undistorted gray video frame */
    IplImage * frameGrayUndistorted;

    /* Wether corrent video sizes are allocated */
    bool videoInitialized;

    /* Video source object */
    VideoSource * videoSource;

    /* Video source info */
    VideoSourceType videoSourceType;
    QString videoSourceId;

    /* Wether first frame the video source has been captured yet */
    bool firstFrameCaptured;

    /* A pointer to the image processing object for this camera */
    ImageProcessing * imageProcessing;

    /* The camera calibration matrix */
    CvMat * cameraMatrix;

    /* OpenCV distortion coefficients. Not nessecearily set, used for compensating for lens distortion */
    CvMat * distortionCoeff;

    /* OpenCV Camera intrinsic matrix. Not nessecearily set, used for compensating for lens distortion */
    CvMat * intrinsicMatrix;

    /* undistortMaps */
    IplImage * undistortMapX;
    IplImage * undistortMapY;

    /* A rotated version of the camera frame */
    IplImage * frameRotated;

    /* Tells wether intrinsic parameters (lens distortion, and OpenCV instrinsic matri is set */
    bool intrinsicParamsSet;

    /* An unique name of the camera */
    QString name;

    /* Mutex to provide wait condition and exclusive access to control flags */
    QMutex mutex;

    /* Wait condition object to provide wakeOne() functionality */
    QWaitCondition condition;

    /* A pointer to the camera control widget */
    CameraControls * cameraControls;

    /* A pointer to the camera calibration dialog */
    CalibrationDialog * calibrationDialog;

    /* Controls weather the thread loop should pause or not */
    bool processNextFrame;

    /* Controls wether the thread should exit or not */
    bool doFinishThread;

    /* A frame writer object */
    FrameWriter frameWriter;

    /* Rectangle which whose size equals the video frame size */
    QRect frameWindow;

    /* Image used for setting the body origin */
    IplImage * checkImage;

    /* Determines wether the check image should be update when a rectangle in the video is selected */
    bool setCheckImageActive;

    /* Performs 90 degree counter clockwise rotation of each frame if true */
    bool rotatedCamera;

    /* A pointer to the camera manager object. Used for telling it when frame is finished processing */
    CameraManager * cameraManager;

private slots:
    void selectPoint(QPoint point);
    void calibrationDialogClosed(int);
    void findTemplate(IplImage *);
    void setMousePosition(QPoint);
    void vesselFound(QPolygon poly);
    void showCalibrationDialog();

public slots:
    void rectangleSelected(QRect);

signals:
    void cameraMatrixChanged(CvMat *);
    void templateTrackerAdded(IplImage *);
    void videoInfoChanged();
    void pointSelected(QPoint);
};

#endif // Camera_H
