#include "cameramanager.h"
#include "trackingmanager.h"
#include "positionestimator.h"
#include "waveestimator.h"

#include <QDebug>


CameraManager::CameraManager(QObject *parent) :
    QThread(parent)
{
    busyCameras = new QSemaphore(cameras.size());
    counter = 0;
    maxSampleRate = 16;
    numAvgSampleRate = 5;
    sumProcessingTime = 0;
    underRate = false;
    endThread = false;
    frameNo = 0;
    doExport = false;

}
void CameraManager::setPositionEstimator(PositionEstimator * positionEstimator)
{
    this->positionEstimator = positionEstimator;
}

void CameraManager::setTrackingManager(TrackingManager * trackingManager)
{
    this->trackingManager = trackingManager;
}

void CameraManager::setWaveEstimator(WaveEstimator * waveEstimator)
{
    this->waveEstimator = waveEstimator;
}


void CameraManager::run()
{
    qDebug() << "CameraManager thread started";
    if (cameras.size() == 0) {
        qDebug() << "No cameras to manage";
        return;
    } else {
        for(int i = 0; i < cameras.size(); i++) {
            cameras.at(i)->start();
        }
    }

    forever {

        mutex.lock();
        if (endThread == true) {
            endThread = false;
            qDebug() << "CameraManager thread finished";
            mutex.unlock();
            return;
        }


        for(int i = 0; i < cameras.size(); i++) {
            busyCameras->acquire(1);
            cameras.at(i)->processFrame();
        }

        busyCameras->acquire(cameras.size());

        int processingTime = time.elapsed();

        int sleepTime = 1000 / maxSampleRate - processingTime;
        if (sleepTime < 0) {
            sleepTime = 0;
            underRate = true;
        } else {
            underRate = false;
        }

        this->msleep(sleepTime);
        sumProcessingTime += time.elapsed();
        time.restart();

        if ((counter % numAvgSampleRate) == 0) {
            sampleRate = numAvgSampleRate * 1000 / sumProcessingTime;
            sumProcessingTime = 0;
        }

        counter++;

        busyCameras->release(cameras.size());

        mutex.unlock();

        trackingManager->checkTracking();
        positionEstimator->compute();
        waveEstimator->setPosition(positionEstimator->getPosition(), positionEstimator->getRotMatrix());

        if (doExport) {
            exportObj.addData(frameNo * 1.0f / 15.0f, positionEstimator->getPosition(), positionEstimator->getAttitude(), trackingManager->getTrackingState(), positionEstimator->numTrackerGroups());
            emit exportInfo(exportObj.getNumDataPoints(), exportObj.getDataSize());
        }

        frameNo++;
        emit allFramesProcessed();
    }
}


void CameraManager::addCamera(Camera * camera)
{
    mutex.lock();
    cameras.append(camera);    
    camera->processFrame();
    emit allFramesProcessed();
    camera->setCameraManager(this);
    delete busyCameras;
    busyCameras = new QSemaphore(cameras.size());
    if (isRunning()) {
        camera->start();
    }
    mutex.unlock();

}

void CameraManager::removeCamera(Camera * camera)
{
    for (int i = 0; i < cameras.size(); i++) {
        if (cameras.at(i) == camera) {
            mutex.lock();
            cameras.remove(i);
            delete busyCameras;
            busyCameras = new QSemaphore(cameras.size());
            mutex.unlock();
        }
    }
}

void CameraManager::frameProcessed(QString id)
{
    if (busyCameras != NULL) {
        busyCameras->release(1);
    }
}

float CameraManager::getSampleRate()
{
    return sampleRate;
}

bool CameraManager::getUnderRate()
{
    return underRate;
}

CameraManager::~CameraManager()
{
    finishThread();
    deleteCameras();

}

bool CameraManager::finishThread()
{
    if (isRunning()) {
        endThread = true;
        wait();

        for(int i = 0; i < cameras.size(); i++) {
            cameras.at(i)->finishThread();
        }
    }
    sampleRate = 0;
    return true;
}

bool CameraManager::deleteCameras()
{
    for(int i = 0; i < cameras.size(); i++) {
        delete cameras.at(i);
    }
}

void CameraManager::startExport()
{
    doExport = true;
    exportObj.clear();
    emit exportInfo(exportObj.getNumDataPoints(), exportObj.getDataSize());
}

void CameraManager::stopExport()
{
    doExport = false;
}

void CameraManager::saveLog(QString fileName)
{
    exportObj.writeToMat(fileName);
}

