#include "houghsvg.h"
#include "vesselsearch.h"

#include <QPainter>
#include <QFileDialog>
#include <QSvgGenerator>
#include <QMessageBox>
#include <math.h>

#include <QDebug>

HoughSVG::HoughSVG(QWidget *parent) :
    QWidget(parent)
{
    inited = false;
    showWorldSet = false;
}

void HoughSVG::paintEvent(QPaintEvent *event)
{
    QPainter painter;
    painter.begin(this);
    painter.setRenderHint(QPainter::Antialiasing);
    if (showWorldSet) {
        paintWorld(painter);
    } else {
        paint(painter);
    }
    painter.end();
}

void HoughSVG::showOriginal()
{
    showWorldSet = false;
    this->show();
}

void HoughSVG::showWorld()
{
    showWorldSet = true;
    show();
}

void HoughSVG::paint(QPainter &painter)
{
    if (inited == false) {
        return;
    }

    //painter.fillRect(QRect(0, 0, 200, 200), Qt::darkBlue);

    for(int i = 0; i < lines->rows; i++) {

        float rho = lines->data.fl[i * 2];
        float theta = lines->data.fl[i * 2 + 1];

        QLineF line = clipLineAgainstRect(QPointF(rho * cos(theta), rho * sin(theta)), QVector2D(cos(theta + M_PI/2), sin(theta + M_PI/2)), QRectF(0, 0, 1600, 1200), false);
        painter.drawLine(line.p1(), line.p2());
    }
}


void HoughSVG::paintWorld(QPainter &painter)
{
    if (inited == false) {
        return;
    }

    //painter.fillRect(QRect(0, 0, 200, 200), Qt::darkBlue);

    qDebug() << "painting";
    qDebug() << "lines->rows = " << lines->rows;
    QPoint center(1200 / 2 - 400, 1600 / 2);
    QPoint topLeft(-800, -800);
    QPoint bottomRight(1000, 800);

    for(int i = 0; i < lines->rows; i++) {

        //float rho = lines->data.fl[i * 2];
        //float theta = lines->data.fl[i * 2 + 1];

        float rho = worldLines[i].rho * 2000;
        float theta = worldLines[i].theta;

        QLineF line = clipLineAgainstRect(QPointF(rho * cos(theta), rho * sin(theta)), QVector2D(cos(theta + M_PI/2), sin(theta + M_PI/2)), QRectF(-center, bottomRight), false);
        painter.drawLine(line.p1() + center, line.p2() + center);
    }

    painter.setPen(QColor(255, 0, 0));
    painter.drawLine(center + QPointF(0, 0), center + QPointF(200, 0));
    painter.setPen(QColor(0, 255, 0));
    painter.drawLine(center + QPointF(0, 0), center + QPointF(0, 200));


}

void HoughSVG::setLines(CvMat * lines, Line * worldLines, int num)
{
    this->lines = lines;
    this->worldLines = worldLines;
    this->num = lines->rows;
    inited = true;
}

void HoughSVG::compute()
{
    qDebug() << "compute called";
    repaint();

}

void HoughSVG::saveSvg(QString filename, QSize videoSize)
{
    if (!inited) {
        qDebug() << "lines not set";
        return;
    }

    if (num == 0) {
        QMessageBox msgBox;
        msgBox.setText("No lines exist so output will be empty.");
        msgBox.exec();
    }

    QSvgGenerator generator;
    generator.setFileName(filename);
    generator.setSize(videoSize);
    generator.setViewBox(QRect(0, 0, videoSize.width(), videoSize.height()));

    QPainter painter;
    painter.begin(&generator);
    paint(painter);
    painter.end();
}

void HoughSVG::saveWorldSvg(QString filename, QSize videoSize)
{
    if (!inited) {
        qDebug() << "lines not set";
        return;
    }

    if (num == 0) {
        QMessageBox msgBox;
        msgBox.setText("No lines exist so output will be empty.");
        msgBox.exec();
    }

    QSvgGenerator generator;
    generator.setFileName(filename);
    generator.setSize(QSize(videoSize.height(), videoSize.width()));
    generator.setViewBox(QRect(0, 0, videoSize.height(), videoSize.width()));

    QPainter painter;
    painter.begin(&generator);
    paintWorld(painter);
    painter.end();
}

/**
  * TODO: Does not direct the line properly, Fix as done in the integer variant above.
  */
QLineF HoughSVG::clipLineAgainstRect(QPointF p, QVector2D dir, QRectF window, bool debug)
{
    debug = false;
    // Intersection left edge
    float leftY = -1;
    if (dir.x() != 0) {
        leftY = p.y() + (window.left() - p.x()) / dir.x() * dir.y();

        if  (leftY < window.top() || leftY > window.bottom()) {
            leftY = -1;
        }
    }

    // Intersection right edge
    float rightY = -1;
    if (dir.x() != 0) {
        rightY = p.y() + (window.right() - p.x()) / dir.x() * dir.y();


        if  (rightY < window.top() || rightY > window.bottom()) {
            rightY = -1;
        }
    }

    // Intersection top edge
    float topX = -1;
    if (dir.y() != 0) {
        topX = p.x() + (window.top() - p.y()) / dir.y() * dir.x();


        if (topX < window.left() || topX > window.right()) {
            topX = -1;
        }
    }

    // Intersection bottom edge
    float bottomX = -1;
    if (dir.y() != 0) {
        bottomX = p.x() + (window.bottom() - p.y()) / dir.y() * dir.x();


        if (bottomX < window.left() || bottomX > window.right()) {
            bottomX = -1;
        }
    }

    // Find two intersections
    int found = 0;

    QPointF pt1, pt2;

    if (leftY != -1) {
        pt1 = QPointF(window.left(), leftY);
        found++;
    }

    if (rightY != -1) {
        if (found == 0) {
            pt1 = QPointF(window.right(), rightY);
        } else {
            pt2 = QPointF(window.right(), rightY);
        }
        found++;
    }

    if (topX != -1) {
        if (found == 0) {
            pt1 = QPointF(topX, window.top());
        } else {
            pt2 = QPointF(topX, window.top());
        }
        found++;
    }

    if (bottomX != -1) {
        if (found == 0) {
            pt1 = QPointF(bottomX, window.bottom());
        } else {
            pt2 = QPointF(bottomX, window.bottom());
        }
        found++;
    }

    QVector2D vec(pt2 - pt1);


    QVector2D diff = vec.normalized() - dir.normalized();

    if (debug) {
        qDebug() << "vec: " << vec.normalized();
        qDebug() << "dir: " << dir.normalized();
        qDebug() << "x fabs: " << fabs(diff.x()) << " x fabs: " << fabs(diff.y());
    }

    QLineF ret;

    if (fabs(diff.x()) < 0.01 && fabs(diff.y()) < 0.01) {
        if (debug) {
            qDebug() << "returning pt1 -> pt2";
        }
        ret = QLineF(pt1, pt2);
    }
    if (debug) {
        qDebug() << "returning pt2 -> pt1";
    }
    ret = QLineF(pt2, pt1);

    if (debug) {
        qDebug() << "ret is " << ret;
    }
    return ret;

}
