#include <QDebug>
#include <highgui.h>
#include <QList>

#include "imageprocessing.h"
#include "tracker.h"
#include "vesselsearch.h"
#include "videosource.h"
#include "camera.h"
#include "util.h"

#include <QVector2D>
#include <QVector3D>
#include <QLine>


ImageProcessing::ImageProcessing(Camera * parent, QSize size, QString id)
{
    this->parent = parent;
    inited = false;
    doVesselSearch = false;
    nextTrackerId = 1;
    cameraMatrix = cvCreateMat(3, 4, CV_32F);
    vesselSearch = new VesselSearch(this, size, id.toInt());
    setVideoSize(size);
    matchLimit = 0.95;
    accelLimit = 10;

    // Make sure that vessel search is stopped when the vessel is found.
    connect(vesselSearch, SIGNAL(vesselEdgesFound(QLine,QLine)), this, SLOT(stopVesselSearch(QLine, QLine)));
}

/**
  * Sets parameters used in tracking
  *
  * @param accelLimit The maximuim allowed acceleration
  * @param matchLimit The minimum correlation match
  */
void ImageProcessing::setTrackingParams(float accelLimit, float matchLimit)
{
    this->accelLimit = accelLimit;
    this->matchLimit = matchLimit;
    qDebug() << "New tracking parameters set";
}

/**
  * Returns a pointer to the parent camera
  *
  * @return The pointer.
  */
Camera * ImageProcessing::getParent()
{
    return parent;
}

/**
  * Initializes video frame sizes
  *
  * @param videoSize The new video size.
  */
void ImageProcessing::setVideoSize(QSize videoSize)
{
    if (videoSize == this->videoSize) {
        return;
    }

    if (inited) {
        cvReleaseImage(&filtered);
    }

    filtered = cvCreateImage(cvSize(videoSize.width(), videoSize.height()), IPL_DEPTH_8U, 3);
    vesselSearch->setVideoSize(videoSize);

    this->videoSize = videoSize;
    inited = true;
}

/**
  * Destroys allocated data structures.
  */
ImageProcessing::~ImageProcessing()
{
    cvReleaseImage(&filtered);
    cvReleaseMat(&cameraMatrix);
    delete vesselSearch;
}

/**
  * Adds a tracker with the specified rectangle to
  * the current video frame.
  *
  * @param rect The area to add the tracker.
  */
int ImageProcessing::addTracker(QRect rect)
{

    QRect image(0, 0, frame->width, frame->height);
    if (!image.contains(rect)) {
        qDebug() << "Rect " << rect << " is not a valid tracker.";
        return 0;
    }

    Tracker * tracker = new Tracker(frame->width, frame->height, this);
    tracker->setFrame(frame);
    tracker->setTargetRect(rect);
    tracker->track();
    trackerData.lock();
    trackers[nextTrackerId] = tracker;

    trackerData.unlock();

    emit trackersChanged();
    emit trackerAdded(nextTrackerId);

    nextTrackerId++;
    return nextTrackerId - 1;
}

/**
  * Remove tracker with the specified id.
  *
  * @param id Tracker id.
  * @return Returns true if the tracker was found.
  */
bool ImageProcessing::removeTracker(int id)
{

    trackerData.lock();
    if (!trackers.contains(id)) {
        trackerData.unlock();
        return false;

    }
    delete trackers.value(id);
    trackers.remove(id);
    trackerData.unlock();
    emit trackersChanged();
    return true;
}

/**
  * Removes all trackers
  */
void ImageProcessing::removeAllTrackers()
{
    trackerData.lock();
    QList <int> keys = trackers.keys();
    for (int i = 0; i < keys.size(); i++) {
        delete trackers.value(keys.at(i));
        trackers.remove(keys.at(i));
    }
    trackerData.unlock();

    emit trackersChanged();
}

/**
  * Iterates through the trackers and calls the track method on them.
  */
void ImageProcessing::track()
{
    cvZero(filtered);

    trackerData.lock();
    QList <int> keys = trackers.keys();

    for(int i = 0; i < keys.size(); i++) {

        Tracker * currrentTracker = trackers.value(keys.at(i));
        currrentTracker->setFrame(frame);
        currrentTracker->track();

        float accel = QVector2D(currrentTracker->getAcceleration()).length();
        if (accel > accelLimit || currrentTracker->getMatch() < matchLimit) {
            if (accel > accelLimit) {
                qDebug() << "Tracker failed with to large acceleration: " << accel;
            } else {
                qDebug() << "Tracker failed with to bad match: " << currrentTracker->getMatch();
            }

            trackers.remove(keys.at(i));
            delete currrentTracker;
            emit trackersChanged();
            continue;
        }

        QRect roi = currrentTracker ->getRoi();
        cvSetImageROI(filtered, cvRect(roi.x(), roi.y(), roi.width(), roi.height()));
        cvCopy(currrentTracker ->getFilteredRoiImage(), filtered);
        cvResetImageROI(filtered);
        cvRectangle(filtered, cvPoint(roi.x(), roi.y()), cvPoint(roi.x() + roi.width(), roi.y() + roi.height()), cvScalar(0, 0, 255), 1);
        QRect templRect = currrentTracker ->getTemplateRect();
        cvRectangle(filtered, cvPoint(roi.x() + templRect.left(), roi.y() + templRect.top()), cvPoint(roi.x() + templRect.left() + templRect.width() , roi.y() + templRect.top() + templRect.height()), cvScalar(255, 0, 0), 1);

    }
    trackerData.unlock();
}

/**
  * Returns a image filled with the template of the specified tracker.
  * The image must be freed by the caller when appropriate.
  *
  * @return A pointer to the created image.
  */
IplImage * ImageProcessing::getTrackerTemplate(int id)
{
    trackerData.lock();
    Tracker * tracker = trackers.value(id, NULL);

    if (tracker == NULL) {
        trackerData.unlock();
        return NULL;
    }

    IplImage * trackerTempl = tracker->getTemplate();
    IplImage * templ = cvCreateImage(cvGetSize(trackerTempl), IPL_DEPTH_8U, 3);
    cvCopy(trackerTempl, templ);
    trackerData.unlock();
    return templ;
}

/**
  * Returns the ROI image of a specific tracker.
  *
  * @return The ROI image.
  */
IplImage * ImageProcessing::getTargetRoi(unsigned int)
{
    return NULL;
}

/**
  * Returns a pointer to the filtered image
  *
  * @return The pointer
  */
IplImage * ImageProcessing::getFilteredFrameCV()
{
    if (doVesselSearch) {
        return vesselSearch->getLineImage();
    } else {
        return filtered;
    }
}

/**
  * Sets the next frame to be processed.
  * Perform tracking and maybe vessel search.
  *
  * @param frame The next frame to be processed.
  */
void ImageProcessing::processFrame(IplImage * frame)
{
    this->frame = frame;
    frameWindow = QRect(0, 0, frame->width, frame->height);

    track();

    if (doVesselSearch) {
        vesselSearch->setFrame(frame);
    }
}

/**
  * Draws a 3D cube by drawing directly on the specified frame. For debugging camera calibration.
  *
  * @param frame The frame to draw in.
  */
void ImageProcessing::drawCube(IplImage * frame)
{
    cvCircle(frame, mapToImage(cvPoint3D32f(0, 0, 0)), 8, cvScalar(0, 255, 0), 8);

    cvLine(frame, mapToImage(cvPoint3D32f(0, 0, 0)), mapToImage(cvPoint3D32f(0.15, 0, 0)), cvScalar(0, 0, 255), 3);
    cvLine(frame, mapToImage(cvPoint3D32f(0.15, 0, 0)), mapToImage(cvPoint3D32f(0.15, 0.15, 0)), cvScalar(0, 0, 255), 3);
    cvLine(frame, mapToImage(cvPoint3D32f(0.15, 0.15, 0)), mapToImage(cvPoint3D32f(0, 0.15, 0)), cvScalar(0, 0, 255), 3);
    cvLine(frame, mapToImage(cvPoint3D32f(0, 0.15, 0)), mapToImage(cvPoint3D32f(0, 0, 0)), cvScalar(0, 0, 255), 3);
    
    cvLine(frame, mapToImage(cvPoint3D32f(0, 0, 0.15)), mapToImage(cvPoint3D32f(0.15, 0, 0.15)), cvScalar(0, 0, 255), 3);
    cvLine(frame, mapToImage(cvPoint3D32f(0.15, 0, 0.15)), mapToImage(cvPoint3D32f(0.15, 0.15, 0.15)), cvScalar(0, 0, 255), 3);
    cvLine(frame, mapToImage(cvPoint3D32f(0.15, 0.15, 0.15)), mapToImage(cvPoint3D32f(0, 0.15, 0.15)), cvScalar(0, 0, 255), 3);
    cvLine(frame, mapToImage(cvPoint3D32f(0, 0.15, 0.15)), mapToImage(cvPoint3D32f(0, 0, 0.15)), cvScalar(0, 0, 255), 3);
    
    cvLine(frame, mapToImage(cvPoint3D32f(0, 0, 0)), mapToImage(cvPoint3D32f(0, 0, 0.15)), cvScalar(0, 0, 255), 3);
    cvLine(frame, mapToImage(cvPoint3D32f(0.15, 0, 0)), mapToImage(cvPoint3D32f(0.15, 0, 0.15)), cvScalar(0, 0, 255), 3);
    cvLine(frame, mapToImage(cvPoint3D32f(0.15, 0.15, 0)), mapToImage(cvPoint3D32f(0.15, 0.15, 0.15)), cvScalar(0, 0, 255), 3);
    cvLine(frame, mapToImage(cvPoint3D32f(0, 0.15, 0)), mapToImage(cvPoint3D32f(0, 0.15, 0.15)), cvScalar(0, 0, 255), 3);
}

/**
  * @return Number of trackers.
  */
int ImageProcessing::getNumberOfTrackers()
{
    return trackers.size();
}

/**
  * Finds good features within polygon using cvGoodFeaturesToTrack()
  *
  * @param polygon Restrict the search within this polygon.
  * @param numPoints Maximum number of points to find.
  * @return The points which was found.
  */
QVector <QPoint> ImageProcessing::findGoodFeatures(QPolygon polygon, int numPoints)
{
    this->vesselPolygon = polygon;
    QRect rect = polygon.boundingRect();

    if (rect.left() < 0) {
        rect.setLeft(0);
    }

    if (rect.top() < 0) {
        rect.setTop(0);
    }

    if (rect.right() > frame->width - 1) {
        rect.setRight(frame->width - 1);
    }

    if (rect.bottom() > frame->height - 1) {
        rect.setBottom(frame->height - 1);
    }

    if (rect.width() < 1 || rect.height() < 1 ) {
        qDebug() << "Polygon is completely outside window so search area is empty.";
        QVector <QPoint> empty;
        return empty;
    }

    IplImage * mask = cvCreateImage(cvSize(rect.width(), rect.height()), IPL_DEPTH_8U, 1);
    cvZero(mask);

    CvPoint * points;
    points = new CvPoint[polygon.size()];

    QPolygon roiPolygon = polygon.translated(-rect.topLeft());

    for(int i = 0; i < polygon.size(); i++) {
        points[i].x = roiPolygon.point(i).x();
        points[i].y = roiPolygon.point(i).y();
    }

    cvFillConvexPoly(mask, points, polygon.size(), cvScalar(255, 0, 0));
    delete[] points;

    IplImage * eig = cvCreateImage(cvSize(rect.width(), rect.height()), IPL_DEPTH_32F, 1);
    IplImage * templ = cvCreateImage(cvSize(rect.width(), rect.height()), IPL_DEPTH_32F, 1);

    IplImage * frameRegion = cvCreateImage(cvSize(rect.width(), rect.height()), IPL_DEPTH_8U, 3);
    IplImage * frameGrayRegion = cvCreateImage(cvSize(rect.width(), rect.height()), IPL_DEPTH_8U, 1);

    cvSetImageROI(frame, cvRect(rect.x(), rect.y(), rect.width(), rect.height()));
    cvCopy(frame, frameRegion);
    cvResetImageROI(frame);

    cvCvtColor(frameRegion, frameGrayRegion, CV_BGR2GRAY);
    cvReleaseImage(&frameRegion);

    CvPoint2D32f * corners = new CvPoint2D32f[numPoints];

    cvGoodFeaturesToTrack(frameGrayRegion, eig, templ, corners, &numPoints, 0.98f, 60, mask, 10, 0);

    QVector <QPoint> goodPoints;
    for(int i = 0; i < numPoints; i++) {
        goodPoints.append(QPoint(rect.left() + corners[i].x, rect.top() + corners[i].y));
    }

    delete[] corners;
    cvReleaseImage(&mask);
    cvReleaseImage(&eig);
    cvReleaseImage(&templ);
    cvReleaseImage(&frameGrayRegion);

    return goodPoints;
}

/**
  * Returns the time when the tracker was created
  *
  * @param id Tracker id.
  * @return The start time of the tracker.
  */
QTime ImageProcessing::getTrackerStartTime(int id)
{
    trackerData.lock();
    Tracker * tracker = trackers.value(id, NULL);
    if (tracker == NULL) {
        trackerData.unlock();
        throw -1;
    }
    QTime time = tracker->getStartTime();
    trackerData.unlock();
    return time;
}

/**
  * Returns a hash of the tracker start times.
  *
  * @param id Tracker id.
  * @return The start times.
  */
QHash <int, QTime> ImageProcessing::getTrackerStartTimes()
{
    trackerData.lock();
    QHash <int, QTime> startTimes;

    QList <int> keys = trackers.keys();
    for(int i = 0; i < keys.size(); i++) {
        startTimes[keys.at(i)] = trackers.value(keys.at(i))->getStartTime();
    }

    trackerData.unlock();
    return startTimes;
}

/**
  * Returns a vector of the positions of the trackers.
  *
  * @return The positions of the trackers.
  */
QVector <QPointF> ImageProcessing::getTargetPositions()
{

    trackerData.lock();
    QVector <QPointF> positions;
    QList <int> keys = trackers.keys();
    for(int i = 0; i < keys.size(); i++) {
        QPointF target = trackers.value(keys.at(i))->getPos();
        positions.append(target);
    }
    trackerData.unlock();
    return positions;
}

/**
  * Returns a list of all the tracker ids.
  *
  * @return The tracker ids.
  */
QVector <int> ImageProcessing::getTrackerIds()
{
    QList <int> keys = trackers.keys();
    return keys.toVector();
}

/**
  * Returns the position of the specified tracker.
  *
  * @return The position.
  */
QPointF ImageProcessing::getTrackerPos(int id)
{
    trackerData.lock();
    Tracker * tracker = trackers.value(id, NULL);
    if (tracker == NULL) {
        trackerData.unlock();
        return QPointF(-1, -1);
    }
    QPointF pos = tracker->getPos();
    trackerData.unlock();
    return pos;
}

/**
  * Returns a hash of all the tracker positions.
  *
  * @return The tracker positions.
  */
QHash <int, QPointF> ImageProcessing::getTrackerPositions()
{
    trackerData.lock();
    QHash <int, QPointF> trackerPositions;

    QList <int> keys = trackers.keys();
    for(int i = 0; i < keys.size(); i++) {
        trackerPositions[keys.at(i)] = trackers.value(keys.at(i))->getPos();
    }

    trackerData.unlock();
    return trackerPositions;
}

/**
  * Sets the camera matrix.
  *
  * @param mat The Matrix to update from.
  */
void ImageProcessing::setCameraMatrix(CvMat * mat)
{
    cvCopy(mat, cameraMatrix);
    vesselSearch->setCameraMatrix(mat);
}

/**
  * Returns a pointer to the camera matrix
  *
  * @return The pointer.
  */
CvMat * ImageProcessing::getCameraMatrix()
{
    return cameraMatrix;
}

/**
  * Returns a polygon covering the vessel. For debugging purposes.
  *
  * @return The polygon.
  */
QPolygon ImageProcessing::getVesselPolygon()
{
    return vesselPolygon;
}

/**
  * Performs template match in the whole current video frame.
  *
  * @param The patch to search with
  * @param match Updated with the resulting correlation match
  * @return The resulting match area.
  */
QRect ImageProcessing::templateMatch(IplImage * patch, float & match)
{
    QPolygon poly;
    poly.append(frameWindow.topLeft());
    poly.append(frameWindow.topRight());
    poly.append(frameWindow.bottomRight());
    poly.append(frameWindow.bottomLeft());
    QRect matchRect = templateMatch(patch, poly, match);
    vesselPolygon.clear();
    vesselPolygon.append(matchRect.topLeft());
    vesselPolygon.append(matchRect.topRight());
    vesselPolygon.append(matchRect.bottomRight());
    vesselPolygon.append(matchRect.bottomLeft());
    return matchRect;
}

/**
  * Performs template match inside the specified polygon
  *
  * @param The patch to search with
  * @param polygon The search area. Must be convex.
  * @param match Updated with the resulting correlation match
  * @return The resulting match area.
  */
QRect ImageProcessing::templateMatch(IplImage * patch, QPolygon polygon, float & match)
{
    QRect rect = polygon.boundingRect();
    rect.setWidth(rect.width() + patch->width);
    rect.setHeight(rect.height() + patch->height);

    if (rect.bottom() > frame->height) {
        rect.setBottom(frame->height - 1);
    }

    if (rect.right() > frame->width) {
        rect.setRight(frame->width - 1);
    }

    if (rect.left() < 0) {
        rect.setLeft(0);
    }

    if (rect.top() < 0) {
        rect.setTop(0);
    }

    QSize corrResultSize(rect.width() - patch->width + 1, rect.height() - patch->height + 1);

    CvPoint * points;
    points = new CvPoint[polygon.size()];

    QPolygon roiPolygon = polygon.translated(-rect.topLeft());
    for(int i = 0; i < polygon.size(); i++) {
        points[i].x = roiPolygon.point(i).x();
        points[i].y = roiPolygon.point(i).y();
    }

    try {

        IplImage * mask = cvCreateImage(cvSize(corrResultSize.width(), corrResultSize.height()), IPL_DEPTH_8U, 1);
        cvZero(mask);
        cvFillConvexPoly(mask, points, polygon.size(), cvScalar(255));

        IplImage * frameGray = cvCreateImage(cvSize(frame->width, frame->height), IPL_DEPTH_8U, 1);
        cvCvtColor(frame, frameGray, CV_BGR2GRAY);

        IplImage * roi = cvCreateImage(cvSize(rect.width(), rect.height()), IPL_DEPTH_8U, 1);
        cvSetImageROI(frameGray, cvRect(rect.x(), rect.y(), rect.width(), rect.height()));
        cvCopy(frameGray, roi);

        IplImage * patchGray = cvCreateImage(cvSize(patch->width, patch->height), IPL_DEPTH_8U, 1);
        cvCvtColor(patch, patchGray, CV_BGR2GRAY);

        IplImage * corrResult = cvCreateImage(cvSize(corrResultSize.width(), corrResultSize.height()), IPL_DEPTH_32F, 1);
        cvZero(corrResult);

        cvMatchTemplate(roi, patchGray, corrResult, CV_TM_CCORR_NORMED);

        double min, max;
        CvPoint minPos, maxPos;
        cvMinMaxLoc(corrResult, &min, &max, &minPos, &maxPos, mask);

        match = max;

        QRect match(rect.left() + maxPos.x, rect.top() + maxPos.y, patch->width, patch->height);
        cvDrawRect(mask, cvPoint(maxPos.x, maxPos.y), cvPoint(maxPos.x + patch->width, maxPos.y + patch->height), cvScalar(255), 1);
        cvDrawRect(mask, cvPoint(maxPos.x, maxPos.y), cvPoint(maxPos.x + patch->width, maxPos.y + patch->height), cvScalar(0), 1);

        cvCircle(roi, cvPoint(maxPos.x, maxPos.y), 6, cvScalar(0), 4);
        cvCircle(roi, cvPoint(maxPos.x, maxPos.y), 3, cvScalar(255), 4);
        cvDrawRect(roi, cvPoint(maxPos.x, maxPos.y), cvPoint(maxPos.x + patch->width, maxPos.y + patch->height), cvScalar(255), 1);

        //CvPoint * points2[] = new CvPoint[polygon.size()];
        CvPoint ** points2 = (CvPoint**) malloc(sizeof(CvPoint*) * polygon.size());

        for(int i = 0; i < polygon.size(); i++) {
            CvPoint * point = new CvPoint;
            point->x = points[i].x;
            point->y = points[i].y;
            points2[i] = points;
        }
        int pointCount = polygon.size();
        cvPolyLine(roi, points2, &pointCount, 1, 1, cvScalar(255), 1);
        cvSetImageROI(roi, cvRect(maxPos.x, maxPos.y, patch->width, patch->height));
        cvCopy(patchGray, roi);
        cvResetImageROI(roi);

        //cvNamedWindow("roi");
        //cvShowImage("roi", roi);

        //cvNamedWindow("res");
        //cvShowImage("res", corrResult);

        free(points2);
        delete[] points;
        cvReleaseImage(&frameGray);
        cvReleaseImage(&corrResult);
        cvReleaseImage(&roi);
        cvReleaseImage(&patchGray);
        cvReleaseImage(&mask);
        return match;
    } catch (cv::Exception) {
        qDebug() << "Template match failed. This should not happen.";
        return QRect(0, 0, 0, 0);
    }
}

/**
  * Increase the successive retries parameter for a tracker.
  * Used by TrackingManager.
  *
  * @param The tracker id to increase successive retries
  */
void ImageProcessing::incrSuccessiveRetries(int id)
{
        trackerData.lock();
        Tracker * tracker = trackers.value(id, NULL);
        if (tracker != NULL) {
            tracker->incrSuccessiveTries();
        }
        trackerData.unlock();
}

/**
  * Get the successive retries for a tracker
  *
  * @param id Tracker id
  * @return The successive retries
  */
int ImageProcessing::getTrackerSuccesiveRetries(int id)
{
        trackerData.lock();
        Tracker * tracker = trackers.value(id, NULL);
        if (tracker == NULL) {
            trackerData.unlock();
            return 0;
        }
        int tries = tracker->getSuccessiveTries();
        trackerData.unlock();
        return tries;
}

/**
  * Performs template match along a band.
  *
  * @param patch The patch to search with.
  * @param p A point on the line.
  * @param dir Direction of the line.
  * @param match Updated with the resulting correlation match.
  * @param width The band with around the line
  * @return the matched rectangle area.
  */
QRect ImageProcessing::templateMatchAlongLine(IplImage * patch, QPointF p, QVector2D dir, int width, float &match)
{
    QRect roi(0, 0, frame->width, frame->height);
    return templateMatchAlongLine(patch, p, dir, width, roi, match);
}

/**
  * Creates strip shaped polygon along the line going through center, with direction dir.
  * The band is correctly clipped to window.
  *
  * @param center A center point on the band.
  * @param dir Direction of the band.
  * @param width Width of the band.
  * @param window The clipping window to restrict the search within.
  * @return The created polygon.
  */
QPolygon ImageProcessing::createPolygonFromParallelLines(QPoint center, QVector2D dir, int width, QRect window)
{
    QVector2D ortho(-dir.y(), dir.x());
    ortho.normalize();

    QPoint lineLeftP = center - (ortho * width / 2).toPoint();
    QPoint lineRightP = center + (ortho * width / 2).toPoint();

    QLine lineLeft = vesselSearch->clipLineAgainstRect(lineLeftP, dir, window, true);
    QLine lineRight = vesselSearch->clipLineAgainstRect(lineRightP, dir, window, true);

    QPolygon polygon;

    if ( (lineLeft.p1() == QPointF(0, 0) && lineLeft.p2() == QPointF(0, 0)) || (lineRight.p1() == QPointF(0, 0) && lineRight.p2() == QPointF(0, 0))) {
        qDebug() << "Line 1 and/or line 2 is outside clipping window. No resulting polygon created.";
        throw -1;
    }

    // The following is to make sure that the polygon gets convex
    polygon.append(lineLeft.p1());
    if (lineLeft.y1() == window.top() && lineRight.x1() == window.left()) {
        polygon.append(window.topLeft());
    } else if (lineLeft.x1() == window.left() && lineRight.y1() == window.bottom()) {
        polygon.append(window.bottomLeft());
    } else if (lineLeft.y1() == window.bottom() && lineRight.x1() == window.right()) {
        polygon.append(window.bottomRight());
    } else if (lineLeft.x1() == window.right() && lineRight.y1() == window.top() ) {
        polygon.append(window.topRight());
    }
    polygon.append(lineRight.p1());
    polygon.append(lineRight.p2());

    if (lineRight.y2() == window.bottom() && lineLeft.x2() == window.right()) {
        polygon.append(window.bottomRight());
    } else if (lineRight.x2() == window.right() && lineLeft.y2() == window.top()) {
        polygon.append(window.topRight());
    } else if (lineRight.y2() == window.top() && lineLeft.x2() == window.left()) {
        polygon.append(window.topLeft());
    } else if (lineRight.x2() == window.left() && lineLeft.y2() == window.bottom()) {
        polygon.append(window.bottomLeft());
    }
    polygon.append(lineLeft.p2());
    return polygon;
}
/**
  * Performs template match along a band.
  *
  * @param patch The patch to search with.
  * @param p A point on the line.
  * @param dir Direction of the line.
  * @param roi Restrict the search within this rectangle.
  * @param match Updated with the resulting correlation match.
  * @return the matched rectangle area.
  */
QRect ImageProcessing::templateMatchAlongLine(IplImage * patch, QPointF p, QVector2D dir, int width, QRect roi, float &match)
{
    // Let dir be pointing forward, then ortho is a vector pointing perpendicual to the left
    QVector2D ortho(-dir.y(), dir.x());
    ortho.normalize();
    CvSize size = cvGetSize(patch);
    roi.translate(QPoint(-size.width / 2, -size.height / 2));

    // Find line wich is parallel to dir and intersects upper left corner of patch
    QPoint center(patch->width / 2, patch->height / 2);
    QVector2D halfDiagonal(QPoint(0, 0) - center);

    QVector2D offset = (halfDiagonal * ortho) / pow(ortho.length(), 2) * ortho;

    if (offset.x() > 0 || offset.y() > 0) {
        if (offset.x() > 0 && offset.y() > 0) {
            offset = -offset;
        } else {
            qDebug() << "templateMatchALongLine geometry failed (maybe this should never happen)";
            return QRect(0, 0, 0, 0);
        }

    }

    QPoint topLeft = p.toPoint() + offset.toPoint();
    QPolygon poly = createPolygonFromParallelLines(topLeft, dir, width, roi);

    // For visualization
    roi.translate(QPoint(size.width / 2, size.height / 2));
    emit templateMatchAlongLineDone(createPolygonFromParallelLines(p.toPoint(), dir, width, roi), Util::clipLineAgainstRect(p.toPoint(), dir, frameWindow));

    return templateMatch(patch, poly, match);
}

/**
  * Sets vessel search to be performed on each frame. If it is
  * alredy running the search is reset.
  */
void ImageProcessing::setVesselSearch()
{
    if (doVesselSearch == false) {
        vesselSearch->reset();
        doVesselSearch = true;
    }
}

/**
  * Returns a pointer the vessel search object.
  */
VesselSearch * ImageProcessing::getVesselSearch()
{
    return vesselSearch;
}

/**
  * Sets vessel search not to be computed on each frame.
  */
void ImageProcessing::stopVesselSearch(QLine, QLine)
{
    doVesselSearch = false;  
}

/**
  * Finds the pixel coordinates of a 3D point by using the camera matrix
  *
  * @param point The 3D point.
  * @return The image point.
  */
CvPoint ImageProcessing::mapToImage(CvPoint3D32f point)
{
    CvMat * world = cvCreateMat(4, 1, CV_32F);

    world->data.fl[0] = point.x;
    world->data.fl[1] = point.y;
    world->data.fl[2] = point.z;
    world->data.fl[3] = 1;

    CvMat * image = cvCreateMat(3, 1, CV_32F);
    cvGEMM(cameraMatrix, world, 1.0f, NULL, 1.0f, image, 0);

    float u = image->data.fl[0] / image->data.fl[2];
    float v = image->data.fl[1] / image->data.fl[2];

    CvPoint p = cvPoint(u, v);

    cvReleaseMat(&world);
    cvReleaseMat(&image);

    return p;
}
