#ifndef IMAGEPROCESSING_H
#define IMAGEPROCESSING_H

#include <QObject>
#include <QImage>
#include <QMutex>
#include <QWaitCondition>
#include <QHash>
#include <QVector2D>

#include <cv.h>
#include <highgui.h>

class Tracker;
class VesselSearch;
class Camera;
class VideoSource;

class ImageProcessing : public QObject
{
Q_OBJECT;

public:
    ImageProcessing(Camera * camera, QSize, QString);
    void setVideoSize(QSize);
    ~ImageProcessing();
    void processFrame(IplImage *);
    IplImage * getFilteredFrameCV();
    IplImage * getTargetRoi(unsigned int);
    QVector <QPointF> getTargetPositions();
    bool removeTracker(int);
    void setCameraMatrix(CvMat *);
    CvMat * getCameraMatrix();
    QRect templateMatch(IplImage *, QPolygon poly, float &match);
    QRect templateMatch(IplImage * patch, float & match);
    QPolygon getVesselPolygon();
    VesselSearch * getVesselSearch();
    QVector <QPoint> findGoodFeatures(QPolygon polygon, int numPoints);
    QRect templateMatchAlongLine(IplImage * patch, QPointF p, QVector2D dir, int width, float &match);
    QRect templateMatchAlongLine(IplImage * patch, QPointF p, QVector2D dir, int width, QRect roi, float &match);
    int getNumberOfTrackers();
    QHash <int, QPointF> getTrackerPositions();
    QVector <int> getTrackerIds();
    QPointF getTrackerPos(int id);
    IplImage * getTrackerTemplate(int id);
    QTime getTrackerStartTime(int id);
    QHash <int, QTime> getTrackerStartTimes();
    Camera * getParent();
    void incrSuccessiveRetries(int id);
    int getTrackerSuccesiveRetries(int id);
    void setTrackingParams(float accelLimit, float matchLimit);

private:
    void track(void);
    QPolygon createPolygonFromParallelLines(QPoint center, QVector2D  dir, int width, QRect window);

    /* A pointer to the parent camera. */
    Camera * parent;

    /* Tells wether the video sizes have been inited. */
    bool inited;

    /* The camera calibration matrix. */
    CvMat * cameraMatrix;

    /* A pointer to current frame being processed. */
    IplImage * frame;

    /* A pointer to a filtered frame which visualizes some of the filtering. */
    IplImage * filtered;

    /* A window whose width/height is equal to the video frame. */
    QRect frameWindow;

    /* Associative list of trackers. */
    QHash <int, Tracker *> trackers;

    /* A pointer to the vesselSearch object. */
    VesselSearch * vesselSearch;

    /* Determines wether vesselSearch should be run for each frame. */
    bool doVesselSearch;

    /* Locked when tracker data is being accessed. */
    QMutex trackerData;

    /* Holds the size of the video frames. */
    QSize videoSize;

    /* Finds the image point of a 3D point using the camera matrix. */
    CvPoint mapToImage(CvPoint3D32f point);

    /* Draws a cube using mapToImage. */
    void drawCube(IplImage * frame);

    /* Holds the id of the next tracker to be created. */
    int nextTrackerId;

    /* Specifies the minimum correlation match of the trackers. */
    float matchLimit;

    /* Specifices the maximum allowed acceleration for a tracker. */
    float accelLimit;

    /* A polygon which are used only for debugging/visualization purposes. */
    QPolygon vesselPolygon;


public slots:
    void setVesselSearch();
    void removeAllTrackers();
    void stopVesselSearch(QLine, QLine);
    int addTracker(QRect);

signals:
    void trackersChanged();
    void processingFinished();
    void trackerAdded(int);
    void templateMatchAlongLineDone(QPolygon polygon, QLine line);
};

#endif // IMAGEPROCESSING_H
