#include "mainwindow.h"
#include "ui_mainwindow.h"

#include <QErrorMessage>
#include <QList>
#include <QMessageBox>
#include <QDebug>
#include <QSettings>
#include <QPolygon>
#include <QInputDialog>
#include <QVariant>

#include "camera.h"
#include "cameracontrols.h"

#ifdef Q_OS_LINUX
#include "sys1394.h"
#endif

#include "presentation.h"
#include "wavemap.h"
#include "waveestimator.h"
#include "vesselsearch.h"
#include "imageprocessing.h"
#include "videoframe.h"
#include "sourceselector.h"
#include "export.h"
#include "util.h"

MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent), ui(new Ui::MainWindow)
{
    patch = cvCreateImage(cvSize(50, 50),  IPL_DEPTH_8U, 3);
    cvSet(patch, cvScalar(127, 127, 127));

    ui->setupUi(this);

    plot = new Plot();
    parameters = new Parameters();

    ui->verticalLayout->setStretch(0, 4);
    ui->verticalLayout->setStretch(1, 2);

    timer = new QTimer(this);
    timer->setInterval(20);
    time.restart();

    positionEstimator = new PositionEstimator();
    positionEstimator->setAftToCenterDistance(0.21);
    waveestimator = new WaveEstimator();
    presentation = new Presentation();

    readSettings();

    cameraManager = new CameraManager();
    trackingManager = new TrackingManager();
    cameraManager->setTrackingManager(trackingManager);
    trackingManager->setPositionEstimator(positionEstimator);
    cameraManager->setPositionEstimator(positionEstimator);
    cameraManager->setWaveEstimator(waveestimator);

    showVideo = true;

    connect(cameraManager, SIGNAL(allFramesProcessed()), this, SLOT(scheduleUpdateUi()));
    connect(cameraManager, SIGNAL(exportInfo(int,int)), plot, SLOT(setExportInfo(int,int)));
    connect(plot, SIGNAL(startLogging()), cameraManager, SLOT(startExport()));
    connect(plot, SIGNAL(stopLogging()), cameraManager, SLOT(stopExport()));
    connect(plot, SIGNAL(saveToMatFile(QString)), cameraManager, SLOT(saveLog(QString)));

    connect(timer, SIGNAL(timeout()), this, SLOT(updateUi()));

    videoScene = new VideoScene(ui->tabWidget);
    videoScene->scale(0.1, 0.1);
    ui->tabWidget->addTab(videoScene, tr("Video analysis"));
    ui->tabWidget->addTab(plot, tr("Measurements"));
    ui->tabWidget->addTab(parameters, tr("Parameters"));

    #ifdef Q_OS_LINUX
    sys1394 = new Sys1394();
    #endif

    /* 576537786845372278
       576537786845372279
       576537786841918924 */

    CvMat * distortionLeft = (CvMat*) cvLoad("Distortion_left.xml");
    CvMat * intrinsicLeft = (CvMat*) cvLoad("Intrinsics_left.xml");
    CvMat * distortionRight = (CvMat*) cvLoad("Distortion_right.xml");
    CvMat * intrinsicRight = (CvMat*) cvLoad("Intrinsics_right.xml");

    sourceSelector = new SourceSelector(this);
    connect(this->sourceSelector, SIGNAL(finished(int)), this, SLOT(sourceSelectorClosed(int)));
    sourcesMenu = ui->menuBar->addMenu(QString("Video sources"));
    sourcesMenu->addAction(QString("New..."), sourceSelector, SLOT(show()));

    //Camera * camera = addCamera(QString("LEFT"));


    //addCameraFromSource(DC1394, QString("576537786841918924"), "LEFT");
    //addCameraFromSource(VIDEO_FILE, QString("C:\\Documents and Settings\\hornang\\Desktop\\left_6_meters.avi"), "LEFT");
    //addCameraFromSource(VIDEO_FILE, QString("C:\\Documents and Settings\\hornang\\Desktop\\right_6_meters.avi"), "RIGHT");
    addCameraFromSource(VIDEO_FILE, QString("/home/hornang/Videos/6_meter_waves/left_6_meters.avi"), "LEFT");
    addCameraFromSource(VIDEO_FILE, QString("/home/hornang/Videos/6_meter_waves/right_6_meters.avi"), "RIGHT");
    //addCameraFromSource(VIDEO_FILE, QString("/home/hornang/Videos/10_meter_waves/right_10_meters.avi"), "RIGHT");

    //camera->set1394sys(sys1394);
    //camera->setVideoSource(DC1394, QString("576537786845372278"));
    //camera->setIntrinsic(intrinsicLeft, distortionLeft);
    //camera->setVideoSource(VIDEO_FILE, QString("/home/hornang/Videos/10_meter_waves/left_10_meters.avi"));
    //camera->setVideoSource(VIDEO_FILE, QString("/home/hornang/Desktop/converted.avi"));
    //camera->setVideoSource(IMAGE, QString("frame.png"));

    //Camera * camera2 = addCamera(QString("RIGHT"));
    //camera2->set1394sys(sys1394);
    //camera2->setVideoSource(IMAGE, QString("frame.png"));
    //camera2->setVideoSource(DC1394, QString("576537786845372279"));
    //camera2->setIntrinsic(intrinsicRight, distortionRight);
    //camera2->setVideoSource(VIDEO_FILE, QString("/home/hornang/Videos/10_meter_waves/right_10_meters.avi"));
    //camera2->setRotatedCamera();

    /*Camera * camera3 = addCamera(QString("OVERVIEW"));
    camera3->set1394sys(sys1394);
    camera3->setVideoSource(DC1394, QString("576537786841918924"));
    camera3->setRotatedCamera(); */

    //initCameras();



    processVideo = false;
    framesInProcessing = 0;

    timer->start();
    //presentation->showMaximized();


    trackingViewInitied = false;
}

void MainWindow::initPresentation()
{
    for(int i = 0; i < cameras.size(); i++) {
        if (cameras.at(i)->getName() == QString("LEFT")) {
            presentation->getFrameLeft()->setCameraMatrix(cameras.at(i)->getCameraMatrix());
        } else if (cameras.at(i)->getName() == QString("RIGHT")) {
            presentation->getFrameRight()->setCameraMatrix(cameras.at(i)->getCameraMatrix());
        } else if (cameras.at(i)->getName() == QString("OVERVIEW")) {
            presentation->getFrameTop()->setCameraMatrix(cameras.at(i)->getCameraMatrix());
        }
    }

}

bool MainWindow::addCameraFromSource(VideoSourceType type, QString id, QString name)
{
    Camera * camera = new Camera(name);
    #ifdef Q_OS_LINUX
    camera->set1394sys(sys1394);
    #endif
    camera->setVideoSource(type, id);

    if (camera->initVideo() == false) {
        int ret = QMessageBox::critical(this, tr("Shiptrack"), tr("Failed to initialize the video source: ") + camera->getVideoSourceDescription(), QMessageBox::Ok);
        delete camera;
        return false;
    }
   

    cameraManager->addCamera(camera);
    cameras.append(camera);
    videoScene->addCamera(camera);
    trackingManager->addCamera(camera);
    ui->cameraLayout->addWidget(camera->getCameraControls());
    parameters->setCameras(this->cameras);
    updateMenuBar();
    
    
    QSettings settings("NTNU", "Shiptrack");
    settings.beginGroup("Camera");

    settings.beginGroup(name);

    QList <QVariant> list = settings.value("matrix").toList();

    if (list.size()) {

        CvMat * matrix = cvCreateMat(3, 4, CV_32FC1);
        cvZero(matrix);

        for(int i = 0; i < 3; i++) {
            for(int j = 0; j < 4; j++) {
                cvSetReal2D(matrix, i, j, list.at(i * 4 + j).toFloat());
            }
        }
        camera->setCameraMatrix(matrix);
        cvReleaseMat(&matrix);
    }

    settings.endGroup();
    settings.endGroup();
    
    initPresentation();
    return true;
}

void MainWindow::updateMenuBar()
{
    sourcesMenu->clear();
    sourcesMenu->addAction(QString("New..."), sourceSelector, SLOT(show()));
    for(int i = 0; i < cameras.size(); i++) {
        QAction * action = sourcesMenu->addAction(QString("Remove ") + QString(cameras.at(i)->getName()));
        QVariant data(cameras.at(i)->getName());
        action->setData(data);
        connect(action, SIGNAL(triggered()), this, SLOT(cameraRemoveRequest()));
    }
}

void MainWindow::cameraRemoveRequest()
{
    QAction * action = (QAction *) QObject::sender();
    for(int i = 0; i < cameras.size(); i++) {
        if (cameras.at(i)->getName() == action->data().toString()) {
            qDebug() << "remove cam " << cameras.at(i)->getName();
            removeCamera(cameras.at(i));
            updateMenuBar();
            return;
        }
    }
}

void MainWindow::removeCamera(Camera * camera)
{
    videoScene->removeCamera(camera);
    cameraManager->removeCamera(camera);
    trackingManager->removeCamera(camera);
    ui->cameraLayout->removeWidget(camera->getCameraControls());

    for(int i = 0; i < cameras.size(); i++) {
        if (cameras.at(i) == camera) {
            cameras.remove(i);
        }
    }

    parameters->setCameras(this->cameras);
    delete camera;
    if (cameras.size() == 0) {
        cameraManager->finishThread();
        ui->startCaptureButton->setDisabled(false);
        ui->stopCaptureButton->setDisabled(true);
    }
}


void MainWindow::scheduleUpdateUi()
{
    doUpdateUi = true;
}

void MainWindow::vesselPolygonsFound(QVector <QPolygonF> polygons)
{
}

Camera * MainWindow::addCamera(QString name)
{
    Camera * camera = new Camera(name);
    cameras.append(camera);

    QSettings settings("NTNU", "Shiptrack");
    settings.beginGroup("Camera");

    QString num;
    num.setNum(cameras.size() - 1);

    settings.beginGroup(name);

    QList <QVariant> list = settings.value("matrix").toList();

    if (list.size()) {

        CvMat * matrix = cvCreateMat(3, 4, CV_32FC1);
        cvZero(matrix);

        for(int i = 0; i < 3; i++) {
            for(int j = 0; j < 4; j++) {
                cvSetReal2D(matrix, i, j, list.at(i * 4 + j).toFloat());
            }
        }
        camera->setCameraMatrix(matrix);
        cvReleaseMat(&matrix);
    }

    settings.endGroup();
    settings.endGroup();

    return camera;
}


/**
  * Write all settings do persistent storage using nested groups with QSettings
  */
void MainWindow::writeSettings()
{
    qDebug() << "writing settings";
    QSettings settings("NTNU", "Shiptrack");

    settings.beginGroup("MainWindow");
    settings.setValue("geometry", saveGeometry());
    settings.setValue("windowState", saveState());
    settings.endGroup();
    settings.beginGroup("Camera");
    for(int i = 0; i < cameras.size(); i++) {
        QString num;
        num.setNum(i);
        settings.beginGroup(cameras.at(i)->getName());
        CvMat * matrix = cameras.at(i)->getCameraMatrix();
        printMatrix(matrix);

        QList <QVariant> list;
        for(int i = 0; i < 3; i++) {
            for(int j = 0; j < 4; j++) {
                list.append(QVariant(cvGetReal2D(matrix, i, j)));
            }
        }
        settings.setValue("matrix", QVariant(list));
        settings.endGroup();
    }
    settings.endGroup();
}

/**
  * Read settings regarding window sizes etc.
  */
 void MainWindow::readSettings()
 {
     QSettings settings("NTNU", "Shiptrack");
     qDebug() << "Stringlist:";
     qDebug() << settings.allKeys();
     qDebug() << "end";
     settings.beginGroup("MainWindow");
     restoreGeometry(settings.value("geometry").toByteArray());
     restoreState(settings.value("windowState").toByteArray());
     settings.endGroup();
 }

MainWindow::~MainWindow()
{
    delete ui;
    delete plot;
    delete cameraManager;
    delete videoScene;
    #ifdef Q_OS_LINUX
    delete sys1394;
    #endif
}

void MainWindow::closeEvent(QCloseEvent *)
{
    presentation->close();
    writeSettings();
}

void MainWindow::updateUi()
{
    if (!doUpdateUi) {
        return;
    }
    doUpdateUi = false;

    plot->setPositions(positionEstimator->getPositions());
    plot->showExportInfo();

    // if (positionEstimator->inTracking()) {
    if (trackingManager->getTrackingState() == IN_TRACKING) {
        if (!trackingViewInitied) {
            trackingViewInitied = true;
            time.restart();
            waveestimator->clear();
            plot->clear();
            presentation->clear();
        }

        videoScene->setDeckVisible(true);
        plot->setPos(time.elapsed() / 1000.0f, positionEstimator->getPosition(), positionEstimator->getAttitude());

        videoScene->setDeckPosition(positionEstimator->getPosition(), positionEstimator->getAttitude());
        videoScene->setDeckSize(waveestimator->getSize(), waveestimator->getSquareSize());
        presentation->setWavemap(time.elapsed() / 1000.0f, waveestimator->getZ(), waveestimator->getAmplitude(), waveestimator->getSize(), waveestimator->getSquareSize());
        presentation->setRollPitch(positionEstimator->getAttitude().x(), positionEstimator->getAttitude().y(), positionEstimator->getPosition().z());
        presentation->setDeckVisible(true);

    } else {
        videoScene->setDeckVisible(false);
        presentation->setDeckVisible(false);
        trackingViewInitied = false;
    }

    videoScene->setPositions(positionEstimator->getPositions());

    QString trackingInfo;
    if (trackingManager->getTrackingState() == SEARCHING) {
        trackingInfo = QString("Searching for vessel");
        ui->state->setText(QString("Initializing"));
    } else if (trackingManager->getTrackingState() == FINDING_AFT || trackingManager->getTrackingState() == ISOLATING_AFT) {
        trackingInfo = QString("Finding aftest tracking area");
    } else if (trackingManager->getTrackingState() == ADDING_TRACKERS) {
        QString num;
        num.setNum(positionEstimator->numTrackerGroups());
        trackingInfo = QString("Found ") + num + QString(" point(s)");
    } else if (trackingManager->getTrackingState() == IN_TRACKING) {
        QString num;
        num.setNum(positionEstimator->numTrackerGroups());
        QString sampleRate;
        sampleRate.setNum(cameraManager->getSampleRate(), 'f', 0);
        trackingInfo = QString("Tracking 6 DOF with ") + num + QString(" points @ ") + sampleRate + QString(" Hz");
        ui->state->setText(QString("Tracking"));
    }

    presentation->setTrackingInfo(trackingInfo);

    if (showVideo) {
        videoScene->updateFrames();
        updatePresentation();
    }

    ui->sampleRateDisp->display((int) cameraManager->getSampleRate());
}

void MainWindow::updatePresentation()
{
    if (cameras.size() >= 2) {
        presentation->getVideoFrame(0)->setFrameCV(cameras.at(0)->getFrame());
        presentation->getVideoFrame(1)->setFrameCV(cameras.at(1)->getFrame());
        if (cameras.size() >= 3) {
            presentation->getVideoFrame(2)->setFrameCV(cameras.at(2)->getFrame());
        }
    }

    presentation->setDeckPosition(positionEstimator->getPosition(), positionEstimator->getAttitude());
    QVector3D pos = positionEstimator->getPosition();
}

void MainWindow::pointSelected(QPoint point)
{
    qDebug() << "Point selected2";

    int nPoints = 8;
    CvPoint3D32f * wPoints = new CvPoint3D32f[nPoints];
    CvMat * points1 = cvCreateMat(2, nPoints, CV_32F);
    CvMat * points2 = cvCreateMat(2, nPoints, CV_32F);

    wPoints [0] = cvPoint3D32f(0, 0, 0);
    wPoints [1] = cvPoint3D32f(0.1, 0, 0);
    wPoints [2] = cvPoint3D32f(0.1, 0.1, 0);
    wPoints [3] = cvPoint3D32f(0, 0.1, 0);
    wPoints [4] = cvPoint3D32f(0, 0, 0.1);
    wPoints [5] = cvPoint3D32f(0.1, 0, 0.1);
    wPoints [6] = cvPoint3D32f(0.1, 0.1, 0.1);
    wPoints [7] = cvPoint3D32f(0, 0.1, 0.1);

    /*wPoints [0] = cvPoint3D32f(0, 0, -0.11);
    wPoints [1] = cvPoint3D32f(0.11, 0, -0.11);
    wPoints [2] = cvPoint3D32f(0.11, 0.11, -0.11);
    wPoints [3] = cvPoint3D32f(0, 0.11, -0.11);
    wPoints [4] = cvPoint3D32f(0, 0, 0);
    wPoints [5] = cvPoint3D32f(0.11, 0, 0);
    wPoints [6] = cvPoint3D32f(0.11, 0.11, 0);
    wPoints [7] = cvPoint3D32f(0, 0.11, 0); */

    for(int i = 0; i < nPoints; i++) {
        CvPoint point1 = cameras.at(0)->mapToImage(wPoints[i]);
        cvSetReal2D(points1, 0, i, point1.x);
        cvSetReal2D(points1, 1, i, point1.y);

        CvPoint point2 = cameras.at(1)->mapToImage(wPoints[i]);
        cvSetReal2D(points2, 0, i, point2.x);
        cvSetReal2D(points2, 1, i, point2.y);
    }

    printMatrix(points1);
    printMatrix(points2);


    CvMat * F = cvCreateMat(3, 3, CV_32F);
    cvFindFundamentalMat(points1, points2, F);

    printMatrix(F);
    cvReleaseMat(&points1);
    cvReleaseMat(&points2);

    delete[] wPoints;


    qDebug() << "Points: " << point;
    CvMat * points = cvCreateMat(2, 3, CV_32F);
    cvZero(points);
    cvSetReal2D(points, 0, 0, point.x());
    cvSetReal2D(points, 1, 0, point.y());
    CvMat * lines = cvCreateMat(3, 3, CV_32F);
    cvZero(lines);
    cvComputeCorrespondEpilines(points, 1, F, lines);

    float a = cvGetReal2D(lines, 0, 0);
    float b = cvGetReal2D(lines, 1, 0);
    float c = cvGetReal2D(lines, 2, 0);

    QPointF p(0, -c/b);
    QVector2D dir(b, -a);

    //QPoint p1(0, - (cvGetReal2D(lines, 2, 0) / cvGetReal2D(lines, 1, 0)));
    //QPoint p2(cameras.at(1)->getFrame()->width, -cameras.at(1)->getFrame()->width * (cvGetReal2D(lines, 0, 0) / cvGetReal2D(lines, 1, 0)) - (cvGetReal2D(lines, 2, 0) / cvGetReal2D(lines, 1, 0)));

    printMatrix(lines);

    cvReleaseMat(&points);
    cvReleaseMat(&lines);
    cvReleaseMat(&F);

}

void MainWindow::on_startCaptureButton_clicked()
{
    cameraManager->start();
    ui->startCaptureButton->setDisabled(true);
    ui->stopCaptureButton->setDisabled(false);
}

void MainWindow::on_stopCaptureButton_clicked()
{
    cameraManager->finishThread();
    ui->startCaptureButton->setDisabled(false);
    ui->stopCaptureButton->setDisabled(true);
}

void MainWindow::rememberTrackerId(int id)
{
    /*QPolygon poly;
    poly.append(QPoint(10, 10));
    poly.append(QPoint(100, 10));
    poly.append(QPoint(120, 100));
    poly.append(QPoint(20, 104));

    cameras.at(1)->getImageProcessing()->templateMatch(cameras.at(0)->getImageProcessing()->getTracker(id)->getTemplate(), poly);
    return;*/

    int nPoints = 8;
    CvPoint3D32f * wPoints = new CvPoint3D32f[nPoints];
    CvMat * points1 = cvCreateMat(2, nPoints, CV_32F);
    CvMat * points2 = cvCreateMat(2, nPoints, CV_32F);

    wPoints [0] = cvPoint3D32f(0, 0, 0);
    wPoints [1] = cvPoint3D32f(0.1, 0, 0);
    wPoints [2] = cvPoint3D32f(0.1, 0.1, 0);
    wPoints [3] = cvPoint3D32f(0, 0.1, 0);
    wPoints [4] = cvPoint3D32f(0, 0, 0.1);
    wPoints [5] = cvPoint3D32f(0.1, 0, 0.1);
    wPoints [6] = cvPoint3D32f(0.1, 0.1, 0.1);
    wPoints [7] = cvPoint3D32f(0, 0.1, 0.1);

    for(int i = 0; i < nPoints; i++) {
        CvPoint point1 = cameras.at(0)->mapToImage(wPoints[i]);
        cvSetReal2D(points1, 0, i, point1.x);
        cvSetReal2D(points1, 1, i, point1.y);

        CvPoint point2 = cameras.at(1)->mapToImage(wPoints[i]);
        cvSetReal2D(points2, 0, i, point2.x);
        cvSetReal2D(points2, 1, i, point2.y);
    }

    //printMatrix(points1);
    //printMatrix(points2);


    CvMat * F = cvCreateMat(3, 3, CV_32F);
    cvFindFundamentalMat(points1, points2, F);

    //printMatrix(F);
    cvReleaseMat(&points1);
    cvReleaseMat(&points2);

    delete[] wPoints;

    QPointF point = cameras.at(0)->getImageProcessing()->getTrackerPos(id);
    this->trackerId = id;

    CvMat * points = cvCreateMat(2, 3, CV_32F);
    cvZero(points);
    cvSetReal2D(points, 0, 0, point.x());
    cvSetReal2D(points, 1, 0, point.y());
    CvMat * lines = cvCreateMat(3, 3, CV_32F);
    cvZero(lines);
    cvComputeCorrespondEpilines(points, 1, F, lines);

    float a = cvGetReal2D(lines, 0, 0);
    float b = cvGetReal2D(lines, 1, 0);
    float c = cvGetReal2D(lines, 2, 0);

    QPointF p(0, -c/b);
    QVector2D dir(b, -a);

    //QPoint p1(0, - (cvGetReal2D(lines, 2, 0) / cvGetReal2D(lines, 1, 0)));
    //QPoint p2(cameras.at(1)->getFrame()->width, -cameras.at(1)->getFrame()->width * (cvGetReal2D(lines, 0, 0) / cvGetReal2D(lines, 1, 0)) - (cvGetReal2D(lines, 2, 0) / cvGetReal2D(lines, 1, 0)));

    //printMatrix(lines);

    IplImage * templ = cameras.at(0)->getImageProcessing()->getTrackerTemplate(id);
    if (templ != NULL) {
        float match;
        QRect rect = cameras.at(1)->getImageProcessing()->templateMatchAlongLine(templ, p, dir, 60, match);
        cameras.at(1)->getImageProcessing()->addTracker(rect);
        cvReleaseImage(&templ);
    }

    cvReleaseMat(&points);
    cvReleaseMat(&lines);
    cvReleaseMat(&F);

}

void MainWindow::addTrackerGroup(int id2)
{

    /*IplImage * roi = cameras.at(0)->getImageProcessing()->getTracker(id)->getTemplate();
    QRect rect = cameras.at(1)->getImageProcessing()->templateMatch(roi);
    int id2 = cameras.at(1)->getImageProcessing()->addTracker(rect);*/


    qDebug() << "addTrackerGroup";
    TrackerGroup group;
    TrackerRef trackerRef;

    trackerRef.imageProcessing = cameras.at(0)->getImageProcessing();
    trackerRef.trackerId = trackerId;
    group.trackerRef.append(trackerRef);

    trackerRef.imageProcessing = cameras.at(1)->getImageProcessing();
    trackerRef.trackerId = id2;
    group.trackerRef.append(trackerRef);

    positionEstimator->addTrackerGroup(group);
}

void MainWindow::storeMatrix(CvMat * L, QString filename)
{
    QFile file;
    file.setFileName(filename);
    file.open(QIODevice::WriteOnly);

    char num[100];

    int rows = L->rows;
    int cols = L->cols;

    sprintf(num, "%d\r\n", rows);
    file.write(QByteArray(num));

    sprintf(num, "%d\r\n", cols);
    file.write(QByteArray(num));

    for(int i = 0; i < rows; i++) {
        for(int j = 0; j < cols; j++) {
            sprintf(num, "%f\r\n", cvGetReal2D(L, i, j));
            file.write(QByteArray(num));
        }
    }

    file.close();
}

CvMat * MainWindow::retrieveMatrix(QString filename)
{
    QFile file;
    file.setFileName(filename);
    file.open(QIODevice::ReadOnly);

    if (!file.isReadable()) {
        return NULL;
    }

    int rows = file.readLine().simplified().toInt();
    int cols = file.readLine().simplified().toInt();

    if (rows < 0 || rows > 1023) {
        return NULL;
    }

    if (cols < 0 || cols > 1023) {
        return NULL;
    }

    CvMat * mat = cvCreateMat(rows, cols, CV_32FC1);

    for(int i = 0; i < rows; i++) {
        for(int j = 0; j < cols; j++) {
            cvSetReal2D(mat, i, j, file.readLine().simplified().toDouble());
        }
    }

    file.close();
    return mat;
}

void MainWindow::printMatrix(CvMat * L)
{
    int rows = L->rows;
    int cols = L->cols;

    QString output;
    output.append("[");
    for(int i = 0; i < rows; i++) {
        for(int j = 0; j < cols; j++) {
            QString num;
            num.setNum(cvGetReal2D(L, i, j));
            output.append(num);
            output.append("\t");
        }
        output.append(";\n");
    }
    output.append("]");
    qDebug() << output;
}

void MainWindow::on_actionNew_triggered()
{
    sourceSelector->show();
}

void MainWindow::sourceSelectorClosed(int result)
{
    sourceSelector->stopVideo();

    if (result == QDialog::Accepted) {
        bool ok;
        QString cameraName = QInputDialog::getText(this, tr("Camera name"), tr("name:"), QLineEdit::Normal,
                             QString::number(cameras.size() + 1), &ok);
        if (ok) {
            addCameraFromSource(sourceSelector->getVideoSourceType(), sourceSelector->getVideoSourceId(), cameraName );
        }
    }
}


void MainWindow::on_horizontalSlider_sliderMoved(int position)
{
    CvMat * bodyToNED = cvCreateMat(3, 3, CV_32F);
    Util::setZRotation(position / 100.0f * 2 * M_PI, bodyToNED);
    positionEstimator->setAttitude(bodyToNED);
}

void MainWindow::on_showVideoButton_toggled(bool checked)
{
    this->showVideo = checked;
}

void MainWindow::on_showPresentation_clicked()
{
    presentation->show();
}
