#include <QDebug>
#include <QString>

#include "positionestimator.h"
#include "imageprocessing.h"

PositionEstimator::PositionEstimator()
{

    /* Used for attitude estimation */
    v1 = cvCreateMat(3, 1, CV_32F);
    v2 = cvCreateMat(3, 1, CV_32F);
    v3 = cvCreateMat(3, 1, CV_32F);

    xVector = cvCreateMat(3, 1, CV_32F);
    yVector = cvCreateMat(3, 1, CV_32F);
    zVector = cvCreateMat(3, 1, CV_32F);

    initialRotMatrix = cvCreateMat(3, 3, CV_32F);
    cvSetIdentity(initialRotMatrix);

    bodyToNEDFiltered = cvCreateMat(3, 3, CV_32F);
    cvSetIdentity(bodyToNEDFiltered);

    newAttitude = cvCreateMat(3, 3, CV_32F);
    cvSetIdentity(newAttitude);

    NEDToBody = cvCreateMat(3, 3, CV_32F);
    cvSetIdentity(NEDToBody);

    bodyToNED = cvCreateMat(3, 3, CV_32F);
    cvSetIdentity(bodyToNED);

    cvZero(bodyToNED);
    cvSetReal2D(bodyToNED, 0, 0, 1);
    cvSetReal2D(bodyToNED, 1, 1, 1);
    cvSetReal2D(bodyToNED, 2, 2, 1);

    transMatrix = cvCreateMat(3, 1, CV_32F);
    cvZero(transMatrix);

    rollMean = pitchMean = yawMean = 0;

    visualization = cvCreateImage(cvSize(500, 500), IPL_DEPTH_8U, 3);
    //cvNamedWindow("3dpoints");yaw
    initialPointsSet = false;
    newSet = false;

    xRot = yRot = zRot = 0;
    aftToCenter = 0;
    maxError = 0.06;
    newForwardVec = false;
    newCenterLinePt = false;
    bodyOriginSet = false;
    newPositionAndAttitude = false;
}


QVector3D PositionEstimator::getPositionLowpass()
{
    return positionLowpass;
}

/**
  * Sets the distance from the aft of the ship to the desired origin
  */
void PositionEstimator::setAftToCenterDistance(float dist)
{
    aftToCenter = dist;
}

/**
  * Sets an arbitrary point where the center line of the ship intersects at this moment
  */
void PositionEstimator::setCenterLinePt(QPointF center)
{
    newCenterLinePt = true;
    centerLinePt = center;
}


/**
  * Find the position of a point giveen the body vector
  */
QVector3D PositionEstimator::findPosFromBodyOffset(QVector3D bodyOffset)
{
    CvMat * nedToBody = cvCreateMat(3, 3, CV_32F);
    cvTranspose(NEDToBody, nedToBody);
    QVector3D position = getPosition() + rotateVector(bodyOffset, nedToBody);
    cvReleaseMat(&nedToBody);
    return position;
}

/**
  * Find the position of a point relative to the body origin
  */
QVector3D PositionEstimator::findBodyVector(QVector3D point)
{
    CvMat * nedToBody = cvCreateMat(3, 3, CV_32F);
    cvTranspose(NEDToBody, nedToBody);
    QVector3D body = rotateVector(point - getPosition(), NEDToBody);
    cvReleaseMat(&nedToBody);
    return body;
}

/**
  * Find the tracking point which is closest to the aft of the ship according to the body coordinate system
  */
float PositionEstimator::findAftestTrackingPoint()
{
    float aftest = 100;
    int aftestIndex = -1;

    CvMat * nedToBody = cvCreateMat(3, 3, CV_32F);
    cvTranspose(NEDToBody, nedToBody);

    for(int i = 0; i < trackerGroups.size(); i++) {
        //QVector3D body = rotateVector(trackerGroups.at(i).position - getPosition(), nedToBody);
        QVector3D body = rotateVector(trackerGroups.at(i).position - getPosition(), NEDToBody);
        if (body.x() < aftest) {
            aftestIndex = i;
            aftest = body.x();
        }
    }
    cvReleaseMat(&nedToBody);

    if (aftestIndex != -1) {
        return aftest;
    }

    return 0;
}

/**
  * Sets forward direction of vessel in the XY plane
  */
void PositionEstimator::setForwardVec(QVector2D forward)
{
    this->forwardVec = forward;
    newForwardVec = true;
}

/**
  * Compute the initial rotation matrix from forward vec
  */
void PositionEstimator::computeInitialRotation()
{
    cvTranspose(newAttitude, initialRotMatrix);
    /*
    rollPitchYaw = QVector3D(0, 0, 0);
    rollMean = 0;
    pitchMean = 0;

    qDebug() << "Setting initial rotation matrix";
*/
    /*
    if (newForwardVec) {
        float yaw = -atan2(forwardVec.y(), forwardVec.x());
        setZRotation(yaw, initialRotMatrix);
        newForwardVec = false;
    }

    if (newCenterLinePt) {
        QVector3D centerPoint = rotateVector(QVector3D(centerLinePt.x(), centerLinePt.y(), 0), initialRotMatrix) - initialCentroid;
        bodyOffset = QVector3D(0, centerLinePt.y(), 0);
        newCenterLinePt = false;
    }*/

}

void PositionEstimator::draw3Dpoints()
{
    cvZero(visualization);
    CvPoint prevCenter = cvPoint(0, 0);
    CvPoint firstCenter = cvPoint(0, 0);
    bool first = true;
    for(int i = 0; i < trackerGroups.size(); i++) {
        QVector3D pos = trackerGroups.at(i).position;
        CvPoint center = cvPoint(pos.x() * 1000.0f + 250, -pos.y() * 1000.0f + 250);
        if (!first)  {
            cvLine(visualization, prevCenter, center, cvScalar(0, 0, 255), 1);
        } else {
            firstCenter = center;
            first = false;
        }
        cvCircle(visualization, center, 2, cvScalar(0, 0, 255), 2);
        prevCenter = center;
    }
    cvLine(visualization, prevCenter, firstCenter, cvScalar(0, 0, 255), 1);

    first = true;
    for(int i = 0; i < initial.size(); i++) {
        QVector3D pos = initial.at(i);
        CvPoint center = cvPoint(pos.x() * 1000.0f + 250, -pos.y() * 1000.0f + 250);
        if (!first)  {
            cvLine(visualization, prevCenter, center, cvScalar(0, 255, 0), 1);
        } else {
            firstCenter = center;
            first = false;
        }
        cvCircle(visualization, center, 2, cvScalar(0, 255, 0), 2);
        prevCenter = center;
    }
    cvLine(visualization, prevCenter, firstCenter, cvScalar(0, 255, 0), 1);

    QVector3D ang = getAttitude();
    cvLine(visualization, cvPoint(0, 50), cvPoint(cos(ang.x()) * 200, 50 + sin(ang.x()) * 200), cvScalar(70, 70, 255), 2);
    cvLine(visualization, cvPoint(0, 150), cvPoint(cos(ang.y()) * 200, 150 + sin(ang.y()) * 200), cvScalar(70, 255, 70), 2);
    cvLine(visualization, cvPoint(0, 250), cvPoint(cos(ang.z()) * 200, 250 + sin(ang.z()) * 200), cvScalar(255, 70, 70), 2);
    //cvShowImage("3dpoints", visualization);
}

void PositionEstimator::setCameraDisplays(QVector <Camera *> cameras)
{
    this->cameras = cameras;
}

int PositionEstimator::positions()
{
    return trackerGroups.size();
}

PositionEstimator::~PositionEstimator()
{
    cvReleaseMat(&v1);
    cvReleaseMat(&v2);
    cvReleaseMat(&v3);
    cvReleaseMat(&xVector);
    cvReleaseMat(&yVector);
    cvReleaseMat(&zVector);
    cvReleaseMat(&initialRotMatrix);
    cvReleaseMat(&NEDToBody);
    cvReleaseMat(&bodyToNED);
    cvReleaseMat(&transMatrix);
}

void PositionEstimator::addTrackerGroup(TrackerGroup group)
{
    dataMutex.lock();
    trackerGroups.append(group);
    numTrackerGroupsChanged = true;
    dataMutex.unlock();
}

QVector3D PositionEstimator::getPosition()
{
    QVector3D pos;
    pos.setX(cvGetReal2D(transMatrix, 0, 0));
    pos.setY(cvGetReal2D(transMatrix, 1, 0));
    pos.setZ(cvGetReal2D(transMatrix, 2, 0));
    return pos;
}


QVector3D PositionEstimator::getPosition(int num)
{
    dataMutex.lock();
    if (num < 0 || num > trackerGroups.size() - 1) {
        qDebug() << "no such tracker group";
        dataMutex.unlock();
        return QVector3D(0, 0, 0);
    }
    dataMutex.unlock();
    return trackerGroups.at(num).position;
}

QVector <QVector3D> PositionEstimator::getPositions()
{
    dataMutex.lock();
    QVector <QVector3D> positions;

    for(int i = 0; i < trackerGroups.size(); i++) {
        positions.append(trackerGroups.at(i).position);
    }
    dataMutex.unlock();
    return positions;
}

QVector3D PositionEstimator::getPosition2()
{
    return QVector3D(cvGetReal2D(transMatrix, 0, 0), cvGetReal2D(transMatrix, 1, 0), cvGetReal2D(transMatrix, 2, 0));
}


int PositionEstimator::numTrackerGroups()
{
    return this->trackerGroups.size();
}

QVector <TrackerGroup> PositionEstimator::getTrackerGroups()
{
    return trackerGroups;
}


void PositionEstimator::removeTrackerGroup(int num)
{
    if (num < 0) {
        qDebug() << "num < 0, cannot remove TrackerGroup";
        return;
    }
    if (num > trackerGroups.size() - 1) {
        qDebug() << "num > trackerGroups.size() - 1 cannot remove group";
        return;
    }
    for(int i = 0; i < trackerGroups.at(num).trackerRef.size(); i++) {
        trackerGroups.at(num).trackerRef.at(i).imageProcessing->removeTracker(trackerGroups.at(num).trackerRef.at(i).trackerId);
    }

    trackerGroups.remove(num);
    numTrackerGroupsChanged = true;
}


/**
  * Try to compute 6DOF if sufficient tracking is available
  */
void PositionEstimator::compute()
{
    dataMutex.lock();
    computePositions();
    computeOrientation();
    dataMutex.unlock();
}

void PositionEstimator::computePositions()
{
    QVector <QPointF> points;
    QVector <CvMat *> cameraMatrices;

    for(int i = 0; i < trackerGroups.size(); i++) {

        points.clear();
        cameraMatrices.clear();

        for(int j = 0; j < trackerGroups.at(i).trackerRef.size(); j++) {
            TrackerRef trackerRef = trackerGroups.at(i).trackerRef.at(j);
            QPointF point = trackerRef.imageProcessing->getTrackerPos(trackerRef.trackerId);
            if (point == QPointF(-1, -1)) {
                qDebug() << "Tracker does not exists anymore, removing from group";
                trackerGroups[i].trackerRef.remove(j);
                j--;
                continue;
            }

            points.append(point);
            cameraMatrices.append(trackerRef.imageProcessing->getCameraMatrix());
        }

        if (trackerGroups.at(i).trackerRef.size() == 0) {
            removeTrackerGroup(i);
            i--;
            continue;
        } else if (trackerGroups.at(i).trackerRef.size() < 2) {
            removeTrackerGroup(i);
            i--;
            continue;
        }

        float error;
        trackerGroups[i].position = estimatePosition(points, cameraMatrices, error);
        if (error > 35) {
            qDebug() << "trackerGroup failed with error " << error;
            removeTrackerGroup(i);
            i--;
        }
    }
    //draw3Dpoints();
}

void PositionEstimator::computeOrientation()
{
    QVector <QVector3D> current;

    for(int i = 0; i < trackerGroups.size(); i++) {
        current.append(trackerGroups.at(i).position);
    }

    if (inTracking()) {

        if (initialPointsSet && !numTrackerGroupsChanged && !newPositionAndAttitude){
            estimateOrientation(current);
        } else {

            if (numTrackerGroupsChanged) {
                setInitial(current);
                initialPointsSet = true;
                numTrackerGroupsChanged = false;
            } else if (newPositionAndAttitude) {
                cvTranspose(newAttitude, initialRotMatrix);
                bodyOffset = rotateVector(newPosition - currentCentroid, initialRotMatrix);
                initial = current;
                estimateOrientation(current);
                newPositionAndAttitude = false;
            }



            /*if (!initialPointsSet) {
                computeInitialRotation();
            }*/
            /*if (!bodyOriginSet) {
                float pos = findAftestTrackingPoint();
                bodyOffset.setX(bodyOffset.x() + pos + aftToCenter);
            }*/




        }

    } else {
        initialPointsSet = false;
        // bodyOriginSet = false;
    }
}



QVector3D PositionEstimator::estimatePosition(QVector <QPointF> points, QVector <CvMat *> cameraMatrices, float & error)
{

    if (points.size() != cameraMatrices.size() || points.size() < 2) {
        qDebug() << "Unequal QVector length or less than two cameras";
        return QVector3D(0, 0, 0);
    }

    CvMat * A = cvCreateMat(points.size() * 2, 3, CV_32F);
    CvMat * b = cvCreateMat(points.size() * 2, 1, CV_32F);
    CvMat * pos = cvCreateMat(3, 1, CV_32F);

    for(int i = 0; i < points.size(); i++) {
        CvMat * currentMatrix = cameraMatrices.at(i);

        cvSetReal2D(A, i * 2, 0, points.at(i).x() * cvGetReal2D(currentMatrix, 2, 0) - cvGetReal2D(currentMatrix, 0, 0));
        cvSetReal2D(A, i * 2, 1, points.at(i).x() * cvGetReal2D(currentMatrix, 2, 1) - cvGetReal2D(currentMatrix, 0, 1));
        cvSetReal2D(A, i * 2, 2, points.at(i).x() * cvGetReal2D(currentMatrix, 2, 2) - cvGetReal2D(currentMatrix, 0, 2));
        cvSetReal2D(A, i * 2 + 1, 0, points.at(i).y() * cvGetReal2D(currentMatrix, 2, 0) - cvGetReal2D(currentMatrix, 1, 0));
        cvSetReal2D(A, i * 2 + 1, 1, points.at(i).y() * cvGetReal2D(currentMatrix, 2, 1) - cvGetReal2D(currentMatrix, 1, 1));
        cvSetReal2D(A, i * 2 + 1, 2, points.at(i).y() * cvGetReal2D(currentMatrix, 2, 2) - cvGetReal2D(currentMatrix, 1, 2));

        cvSetReal1D(b, i * 2, cvGetReal2D(currentMatrix, 0, 3) - points.at(i).x() * cvGetReal2D(currentMatrix, 2, 3));
        cvSetReal1D(b, i * 2 + 1, cvGetReal2D(currentMatrix, 1, 3) - points.at(i).y()  * cvGetReal2D(currentMatrix, 2, 3));
    }

    if (!cvSolve(A, b, pos)) {
        qDebug() << "Inverting matrix is singular cannot estimate position.";
        cvReleaseMat(&A);
        cvReleaseMat(&b);
        cvReleaseMat(&pos);
        error = 0;
        return QVector3D(0, 0, 0);
    }

    // Convert b to -b
    for(int i = 0; i < b->rows; i++) {
        b->data.fl[i] *= -1;
    }

    CvMat * errorVec = cvCreateMat(points.size() * 2, 1, CV_32F);

    cvGEMM(A, pos, 1.0f, b, 1.0f, errorVec);

    float sumError = 0;
    for(int i = 0; i < b->rows; i++) {
        sumError += pow(cvGetReal2D(errorVec, i, 0), 2);
    }
    error = sqrt(sumError);

    float x = cvGetReal1D(pos, 0);
    float y = cvGetReal1D(pos, 1);
    float z = cvGetReal1D(pos, 2);

    cvReleaseMat(&A);
    cvReleaseMat(&b);
    cvReleaseMat(&pos);
    cvReleaseMat(&errorVec);

    return QVector3D(x, y, z);
}

QVector3D PositionEstimator::getInitialCentroid()
{
    return initialCentroid;
}

/**
  * Transforms a vector given a 3x3 rotation matrix. returns rot * vec.
  */
QVector3D PositionEstimator::rotateVector(QVector3D qVec, CvMat * rot)
{
    QVector3D res;
    res.setX( cvGetReal2D(rot, 0, 0) * qVec.x() + cvGetReal2D(rot, 0, 1) * qVec.y() + cvGetReal2D(rot, 0, 2) * qVec.z()) ;
    res.setY( cvGetReal2D(rot, 1, 0) * qVec.x() + cvGetReal2D(rot, 1, 1) * qVec.y() + cvGetReal2D(rot, 1, 2) * qVec.z());
    res.setZ( cvGetReal2D(rot, 2, 0) * qVec.x() + cvGetReal2D(rot, 2, 1) * qVec.y() + cvGetReal2D(rot, 2, 2) * qVec.z());
    return res;
}

CvMat * PositionEstimator::getRotMatrix()
{
    return bodyToNEDFiltered;
}

/**
  * Sets "rot" to be a rotation matrix with angle "angle" around the X-axis
  */
void PositionEstimator::setXRotation(float angle, CvMat * rot)
{
    float costheta = cos(angle);
    float sintheta = sin(angle);

    cvZero(rot);
    cvSetReal2D(rot, 0, 0, 1);
    cvSetReal2D(rot, 1, 1, costheta);
    cvSetReal2D(rot, 2, 2, costheta);
    cvSetReal2D(rot, 2, 1, sintheta);
    cvSetReal2D(rot, 1, 2, -sintheta);
}


/**
  * Sets "rot" to be a rotation matrix with angle "angle" around the X-axis
  */
void PositionEstimator::setYRotation(float angle, CvMat * rot)
{
    float costheta = cos(angle);
    float sintheta = sin(angle);

    cvZero(rot);
    cvSetReal2D(rot, 0, 0, costheta);
    cvSetReal2D(rot, 2, 2, costheta);
    cvSetReal2D(rot, 1, 1, 1);
    cvSetReal2D(rot, 2, 0, -sintheta);
    cvSetReal2D(rot, 0, 2, sintheta);
}

/**
  * Sets "rot" to be a rotation matrix with angle "angle" around the Z-axis
  */
void PositionEstimator::setZRotation(float angle, CvMat * rot)
{
    float costheta = cos(angle);
    float sintheta = sin(angle);
    cvZero(rot);
    cvSetReal2D(rot, 0, 0, costheta);
    cvSetReal2D(rot, 1, 1, costheta);
    cvSetReal2D(rot, 2, 2, 1);

    cvSetReal2D(rot, 1, 0, sintheta);
    cvSetReal2D(rot, 0, 1, -sintheta);
}

/**
  * Only used for button press in MainWindow
  */
void PositionEstimator::setInitialAttitude()
{
    QVector <QVector3D> current;

    for(int i = 0; i < trackerGroups.size(); i++) {
        current.append(trackerGroups.at(i).position);
    }
    setInitial(current);
}

void PositionEstimator::setInitial(QVector <QVector3D> newInitial)
{    

    QVector3D prevCurrentCentroid(0, 0, 0);
    for(int i = 0; i < prevCurrent.size(); i++) {
        prevCurrentCentroid += prevCurrent.at(i);
    }
    if (prevCurrent.size() > 0) {
        prevCurrentCentroid /= prevCurrent.size();
    }

    QVector3D newInitialCentroid(0, 0, 0);
    for(int i = 0; i < newInitial.size(); i++) {
        newInitialCentroid += newInitial.at(i);
    }
    if (newInitial.size() > 0) {
        newInitialCentroid /= newInitial.size();
    }

    QVector3D origin = prevCurrentCentroid + rotateVector(bodyOffset, bodyToNED);
    bodyOffset = rotateVector(origin - newInitialCentroid, NEDToBody);

    cvCopy(NEDToBody, initialRotMatrix);
    qDebug() << "initialRotMAtrix set in setInitial";
    initial = newInitial;   
}

void PositionEstimator::setBodyOrigin(QVector3D newOrigin)
{
    dataMutex.lock();

    setZRotation(0, initialRotMatrix);

    newOrigin = rotateVector(newOrigin, NEDToBody);
    newOrigin.setX(newOrigin.x());

    QVector3D origin = getPosition();
    bodyOffset += newOrigin - rotateVector(origin, NEDToBody);
    // bodyOriginSet = true;
    qDebug() << "## Body origin set ##";
    dataMutex.unlock();
}

void PositionEstimator::setPositionAndAttitude(QVector3D position, CvMat * attitude)
{
    dataMutex.lock();
    cvCopy(attitude, newAttitude);
    newPosition = position;
    newPositionAndAttitude = true;
    dataMutex.unlock();
}

void PositionEstimator::setAttitude(CvMat * attitude)
{
    dataMutex.lock();
    cvCopy(attitude, newAttitude);
    initialPointsSet = false;
    dataMutex.unlock();
}

QVector3D PositionEstimator::getCurrentCentroid()
{
    return currentCentroid;
}

void PositionEstimator::estimateOrientation(QVector <QVector3D> current)
{
    prevCurrent = current;

    QVector <QVector3D> initial;
    for(int i = 0; i < this->initial.size(); i++) {
        initial.append(rotateVector(this->initial.at(i) - initialCentroid, initialRotMatrix));
    }

    if (initial.size() != current.size()) {
        qDebug() << "Unequal number of points in initial and current, initial: " << initial.size() << " " << current.size();
        return;
    } else if (initial.size() < 3) {
        qDebug() << "Not enough points to estimate orientation";
        return;
    }

    initialCentroid = QVector3D(0, 0, 0);
    currentCentroid = QVector3D(0, 0, 0);

    // Find centroids of point sets
    for(int i = 0; i < initial.size(); i++) {
        initialCentroid += initial.at(i);
        currentCentroid += current.at(i);
    }

    initialCentroid /= initial.size();
    currentCentroid /= current.size();

    for(int i = 0; i < initial.size(); i++) {
        initial[i] -= initialCentroid;
        current[i] -= currentCentroid;
    }

    CvMat * H = cvCreateMat(3, 3, CV_32F);
    cvZero(H);

    CvMat * init = cvCreateMat(3, 1, CV_32F);
    CvMat * curr = cvCreateMat(3, 1, CV_32F);

    for(int i = 0; i < 3; i++) {
        cvSetReal2D(init, 0, 0, initial.at(i).x());
        cvSetReal2D(init, 1, 0, initial.at(i).y());
        cvSetReal2D(init, 2, 0, initial.at(i).z());

        cvSetReal2D(curr, 0, 0, current.at(i).x());
        cvSetReal2D(curr, 1, 0, current.at(i).y());
        cvSetReal2D(curr, 2, 0, current.at(i).z());

        cvGEMM(init, curr, 1.0f, H, 1.0f, H, CV_GEMM_B_T);
    }

    CvMat * W = cvCreateMat(3, 3, CV_32F);
    CvMat * U = cvCreateMat(3, 3, CV_32F);
    CvMat * V = cvCreateMat(3, 3, CV_32F);

    cvSVD(H, W, U, V, CV_SVD_V_T);

    CvMat * m = cvCreateMat(3, 3, CV_32F);
    CvMat * uvT = cvCreateMat(3, 3, CV_32F);
    CvMat * left = cvCreateMat(3, 3, CV_32F);
    cvZero(m);

    cvGEMM(U, V, 1.0f, NULL, 1.0f, uvT);

    cvSetReal2D(m, 0, 0, 1);
    cvSetReal2D(m, 1, 1, 1);
    cvSetReal2D(m, 2, 2, cvDet(uvT));


    cvGEMM(V, m, 1.0f, NULL, 1.0f, left, CV_GEMM_A_T);

    cvGEMM(left, U, 1.0f, NULL, 1.0f, bodyToNED, CV_GEMM_B_T);

    cvTranspose(bodyToNED, NEDToBody);


    QVector3D translate = currentCentroid + rotateVector(bodyOffset, bodyToNED);

    float x = cvGetReal2D(transMatrix, 0, 0);
    float y = cvGetReal2D(transMatrix, 1, 0);
    float z = cvGetReal2D(transMatrix, 2, 0);

    cvSetReal2D(transMatrix, 0, 0, translate.x());
    cvSetReal2D(transMatrix, 1, 0, translate.y());
    cvSetReal2D(transMatrix, 2, 0, translate.z());

    float rollMeasure = atan2(cvGetReal2D(bodyToNED, 2, 1), cvGetReal2D(bodyToNED, 2, 2));
    float pitchMeasure = -asin(cvGetReal2D(bodyToNED, 2, 0));
    float yawMeasure = atan2(cvGetReal2D(bodyToNED, 1, 0), cvGetReal2D(bodyToNED, 0, 0));

    lowpass(rollMeasure, rollMean, 0.002);
    lowpass(pitchMeasure, pitchMean, 0.002);
    lowpass(yawMeasure, yawMean, 0.002);
    
    float roll = rollPitchYaw.x();
    float pitch = rollPitchYaw.y();
    float yaw = rollPitchYaw.z();

    lowpass(rollMeasure, roll, 0.2);
    lowpass(pitchMeasure, pitch, 0.2);
    lowpass(yawMeasure, yaw, 0.2);

    rollPitchYaw = QVector3D(roll, pitch, yaw);
    //rollPitchYaw = QVector3D(rollMeasure, pitchMeasure, yawMeasure);
    //rollPitchYaw = QVector3D(roll, pitch, yaw);
    setRollPitchYaw(rollPitchYaw.x(), rollPitchYaw.y(), rollPitchYaw.z(), bodyToNEDFiltered);

    //printMatrix(bodyToNEDFiltered);

    float xLowpass = positionLowpass.x();
    float yLowpass = positionLowpass.y();
    float zLowpass = positionLowpass.z();

    lowpass(x, xLowpass, 0.1);
    lowpass(y, yLowpass, 0.1);
    lowpass(z, zLowpass, 0.1);

    positionLowpass = QVector3D(xLowpass, yLowpass, zLowpass);
    //positionLowpass = QVector3D(x, y, z);

    // Check orientation error
    for(int i = 0; i < initial.size(); i++) {
        cvSetReal2D(init, 0, 0, initial.at(i).x());
        cvSetReal2D(init, 1, 0, initial.at(i).y());
        cvSetReal2D(init, 2, 0, initial.at(i).z());

        cvGEMM(NEDToBody, init, 1.0f, NULL, 1.0f, curr);
        float error = sqrt( pow(cvGetReal2D(init, 0, 0) - cvGetReal2D(curr, 0, 0), 2) +
                            pow(cvGetReal2D(init, 1, 0) - cvGetReal2D(curr, 1, 0), 2) +
                            pow(cvGetReal2D(init, 2, 0) - cvGetReal2D(curr, 2, 0), 2) );
        if (error > maxError) {
            qDebug() << "Tracker group " << i << " has to large orientation error ("  << error << "). Removing group...";
            removeTrackerGroup(i);
            break;
        }
    }

    cvReleaseMat(&left);
    cvReleaseMat(&uvT);
    cvReleaseMat(&m);
    cvReleaseMat(&W);
    cvReleaseMat(&U);
    cvReleaseMat(&V);

    cvReleaseMat(&init);
    cvReleaseMat(&curr);
    cvReleaseMat(&H);
}

/**
  * Creates a rotation matrix based on roll, pitch and yaw angles
  */
void PositionEstimator::setRollPitchYaw(float roll, float pitch, float yaw, CvMat * output)
{
    CvMat * tmp = cvCreateMat(3, 3, CV_32F);
    setXRotation(roll, output);
    setYRotation(pitch, tmp);
    cvGEMM(tmp, output, 1.0f, NULL, 1.0f, output);
    setYRotation(yaw, tmp);
    cvGEMM(tmp, output, 1.0f, NULL, 1.0f, output);
    cvReleaseMat(&tmp);
}


/**
  * Indicates wether the vessel is tracked or not
  */
bool PositionEstimator::inTracking()
{
    return (numTrackerGroups() >= 4);
}

/**
  * Updates "output" according to input. When t -> infinity =>  output = input
  */
void PositionEstimator::lowpass(float input, float &output, float timeconstant)
{
    output += (input - output) * timeconstant;
}

QVector3D PositionEstimator::getNormalVector()
{
    return normalVector;
}

/**
  * Compute the angle between two vectors
  */
float PositionEstimator::vectorAngle(QVector3D v1, QVector3D v2)
{
    return acos(QVector3D::dotProduct(v1, v2) / (v1.length() * v2.length()));
}

/**
  * Find a plane whose angles best match the point set.
  * It is a assumed that the plane is close to horizontal.
  */
QPair <float, float> PositionEstimator::estimatePlaneAngles(QVector <QVector3D> points)
{
    if (points.size() < 3) {
        throw -1;
    }

    QVector3D centroid;
    for(int i = 0; i < points.size(); i++) {
        centroid += points.at(i);
    }
    centroid /= points.size();

    for(int i = 0; i < points.size(); i++) {
        points[i] -= centroid;
    }

    int rows = points.size();

    CvMat * A = cvCreateMat(rows, 2, CV_32F);
    CvMat * B = cvCreateMat(rows, 1, CV_32F);
    cvZero(A);
    cvZero(B);

    for(int i = 0; i < points.size(); i++) {
        cvSetReal2D(A, i, 0, points.at(i).x());
        cvSetReal2D(A, i, 1, points.at(i).y());
        cvSetReal2D(B, i, 0, points.at(i).z());
    }

    CvMat * x = cvCreateMat(2, 1, CV_32F);
    cvSolve(A, B, x);

    float a = cvGetReal2D(x, 0, 0);
    float b = cvGetReal2D(x, 1, 0);

    cvReleaseMat(&A);
    cvReleaseMat(&B);
    cvReleaseMat(&x);

    QVector3D v1(1, 0, a);
    QVector3D v2(0, 1, b);
    QVector3D v3 = QVector3D::crossProduct(v1, v2);
    normalVector = v3.normalized();

    float yAngle = atan(a);
    float xAngle = atan(b);

    return QPair <float, float> (xAngle, yAngle);
}

QVector3D PositionEstimator::getAttitude()
{
    return rollPitchYaw - QVector3D(rollMean, pitchMean, 0);
}


void PositionEstimator::estimateOrientationThreePt(CvPoint3D32f p1, CvPoint3D32f p2, CvPoint3D32f p3)
{

    CvMat * v1 = cvCreateMat(3, 1, CV_32F);
    CvMat * v2 = cvCreateMat(3, 1, CV_32F);
    CvMat * v3 = cvCreateMat(3, 1, CV_32F);

    CvMat * xVector = cvCreateMat(3, 1, CV_32F);
    CvMat * yVector = cvCreateMat(3, 1, CV_32F);
    CvMat * zVector = cvCreateMat(3, 1, CV_32F);

    cvSetReal1D(v1, 0, p2.x - p1.x);
    cvSetReal1D(v1, 1, p2.y - p1.y);
    cvSetReal1D(v1, 2, p2.z - p1.z);

    cvSetReal1D(v2, 0, p3.x - p1.x);
    cvSetReal1D(v2, 1, p3.y - p1.y);
    cvSetReal1D(v2, 2, p3.z - p1.z);

    cvCrossProduct(v1, v2, v3);
    double v3Length = sqrt(pow(cvGetReal1D(v3, 0), 2) + pow(cvGetReal1D(v3, 1), 2) + pow(cvGetReal1D(v3, 2), 2));

    cvSetReal1D(zVector, 0, cvGetReal1D(v3, 0) / v3Length);
    cvSetReal1D(zVector, 1, cvGetReal1D(v3, 1) / v3Length);
    cvSetReal1D(zVector, 2, cvGetReal1D(v3, 2) / v3Length);

    double v1Length = sqrt(pow(cvGetReal1D(v1, 0), 2) + pow(cvGetReal1D(v1, 1), 2) + pow(cvGetReal1D(v1, 2), 2));

    cvSetReal1D(xVector, 0, cvGetReal1D(v1, 0) / v1Length);
    cvSetReal1D(xVector, 1, cvGetReal1D(v1, 1) / v1Length);
    cvSetReal1D(xVector, 2, cvGetReal1D(v1, 2) / v1Length);

    cvCrossProduct(zVector, xVector, yVector);
    
    cvSetReal2D(bodyToNED, 0, 0, cvGetReal1D(xVector, 0));
    cvSetReal2D(bodyToNED, 1, 0, cvGetReal1D(xVector, 1));
    cvSetReal2D(bodyToNED, 2, 0, cvGetReal1D(xVector, 2));

    cvSetReal2D(bodyToNED, 0, 1, cvGetReal1D(yVector, 0));
    cvSetReal2D(bodyToNED, 1, 1, cvGetReal1D(yVector, 1));
    cvSetReal2D(bodyToNED, 2, 1, cvGetReal1D(yVector, 2));

    cvSetReal2D(bodyToNED, 0, 2, cvGetReal1D(zVector, 0));
    cvSetReal2D(bodyToNED, 1, 2, cvGetReal1D(zVector, 1));
    cvSetReal2D(bodyToNED, 2, 2, cvGetReal1D(zVector, 2));

}


void PositionEstimator::printMatrix(CvMat * L)
{
    int rows = L->rows;
    int cols = L->cols;

    QString output;
    output.append("[");
    for(int i = 0; i < rows; i++) {
        for(int j = 0; j < cols; j++) {
            QString num;
            num.setNum(cvGetReal2D(L, i, j));
            output.append(num);
            output.append("\t");
        }
        output.append(";\n");
    }
    output.append("]");
    qDebug() << output;
}
