#ifndef POSITIONESTIMATOR_H
#define POSITIONESTIMATOR_H


#include <QVector3D>
#include <QVector2D>
#include <QPair>
#include <QMutex>

#include <cv.h>

#include "camera.h"

struct TrackerRef {
    ImageProcessing * imageProcessing;
    int trackerId;
};

struct TrackerGroup {
    QVector <TrackerRef> trackerRef;
    QVector3D position;
    float error;
};

class PositionEstimator : public QObject
{
    Q_OBJECT

public:
    PositionEstimator();
    ~PositionEstimator();
    void setCameraDisplays(QVector <Camera *>);

    void compute();
    void setInitial(QVector <QVector3D> initial);

    void estimateOrientationThreePt(CvPoint3D32f p1, CvPoint3D32f p2, CvPoint3D32f p3);

    QVector3D getPosition(int);
    QVector3D getPosition();
    QVector3D getPositionLowpass();
    QVector <QVector3D> getPositions();

    int numTrackerGroups();

    QVector3D getPosition2();
    QVector3D getAttitude();
    CvMat * getRotMatrix();
    void printMatrix(CvMat *);
    void addTrackerGroup(TrackerGroup group);
    int positions();

    QVector3D getInitialCentroid();
    QVector <TrackerGroup> getTrackerGroups();
    void setForwardVec(QVector2D);
    void setCenterLinePt(QPointF);
    QVector3D getNormalVector();
    void setAftToCenterDistance(float dist);
    float xRot, yRot, zRot;
    void setInitialAttitude();
    bool inTracking();
    float findAftestTrackingPoint();
    void computeInitialRotation();
    void setBodyOrigin(QVector3D pos);
    void setAttitude(CvMat * attitude);
    void setPositionAndAttitude(QVector3D origin, CvMat * attitude);

    QVector3D findBodyVector(QVector3D point);
    QVector3D findPosFromBodyOffset(QVector3D bodyOffset);
    QVector3D getCurrentCentroid();

private:
    void removeTrackerGroup(int num);
    void draw3Dpoints();
    QVector3D estimatePosition(QVector <QPointF>, QVector <CvMat *>, float &);
    void computePositions();
    void computeOrientation();
    void estimateOrientation(QVector <QVector3D> current);
    QVector3D rotateVector(QVector3D qVec, CvMat * rot);
    QVector3D rollPitchYaw;
    QVector <QVector3D> prevCurrent;
    float yawMean, pitchMean, rollMean;

    void lowpass(float input, float &output, float timeconstant);
    QPair <float, float> estimatePlaneAngles(QVector <QVector3D> points);
    void setXRotation(float angle, CvMat * rot);
    void setYRotation(float angle, CvMat * rot);
    void setZRotation(float angle, CvMat * rot);
    void setRollPitchYaw(float roll, float pitch, float yaw, CvMat * output);
    float vectorAngle(QVector3D v1, QVector3D v2);
    bool bodyOriginSet;
    float aftToCenter;
    QVector <Camera *> cameras;

    CvMat * v1;
    CvMat * v2;
    CvMat * v3;

    CvMat * xVector;
    CvMat * yVector;
    CvMat * zVector;

    CvMat * initialRotMatrix;
    CvMat * NEDToBody;
    CvMat * bodyToNED;
    CvMat * bodyToNEDFiltered;
    CvMat * transMatrix;
    CvMat * newAttitude;
    QVector3D newPosition;

    QPointF centerLinePt;
    bool newCenterLinePt;
    bool newForwardVec;
    QVector <QVector3D> initial;
    bool newInitial;
    QVector3D bodyOffset;
    QVector3D bodyOffsetTarget;
    QVector2D forwardVec;
    QVector <TrackerGroup> trackerGroups;
    QVector3D initialCentroid;
    QVector3D currentCentroid;
    QVector3D normalVector;
    QVector3D positionLowpass;
    bool initialPointsSet;
    IplImage * visualization;
    bool newSet;
    bool numTrackerGroupsChanged;
    float maxError;
    QMutex dataMutex;
    bool newPositionAndAttitude;





signals:
    void allTrackerGroupsFailed();

};

#endif // POSITIONESTIMATOR_H
