#include <QDebug>

#include "videographicsitem.h"
#include <QGLWidget>

#include "presentation.h"
#include "ui_presentation.h"
#include "videoscene.h"
#include "videoframe.h"
#include "wavemap.h"
#include "mapselector.h"

#include <qwt_plot.h>
#include <qwt_plot_curve.h>

Presentation::Presentation(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::Presentation)
{
    ui->setupUi(this);

    //QGLWidget * glWidget = new QGLWidget(this);

    ui->cameraLeft->setViewport(new QGLWidget());
    ui->cameraRight->setViewport(new QGLWidget());
    ui->cameraTop->setViewport(new QGLWidget());


    ui->vesselView->setRenderHints(QPainter::Antialiasing | QPainter::SmoothPixmapTransform | QPainter::HighQualityAntialiasing);

    topScene = new QGraphicsScene(this);
    leftScene = new QGraphicsScene(this);
    rightScene = new QGraphicsScene(this);
    vesselScene = new QGraphicsScene(this);
    logoScene = new QGraphicsScene(this);

    ui->cameraLeft->setScene(leftScene);
    ui->cameraRight->setScene(rightScene);
    ui->cameraTop->setScene(topScene);
    ui->logoView->setScene(logoScene);

    topFrame = new VideoFrame();
    topFrame->videoItem->setGizmoVisible(false);
    leftFrame = new VideoFrame();
    leftFrame->videoItem->setGizmoVisible(false);
    rightFrame = new VideoFrame();
    rightFrame->videoItem->setGizmoVisible(false);

    topScene->addItem(topFrame);
    leftScene->addItem(leftFrame);
    rightScene->addItem(rightFrame);

    topFrame->setOSD(QString("CAMERA OVERVIEW"));
    leftFrame->setOSD(QString("CAMERA LEFT"));
    rightFrame->setOSD(QString("CAMERA RIGHT"));

    wavemap = new Wavemap();
    QGraphicsScene * waveMapScene = new QGraphicsScene(this);
    ui->waveMap->setScene(waveMapScene);
    waveMapScene->addItem(wavemap);

    plot = new QwtPlot();
    QwtText title;
    title.setText(QString("Selected heave measurement"));
    plot->setTitle(title);



    title.setText(QString("time [s]"));
    plot->setAxisTitle(QwtPlot::xBottom, title);

    heaveCurve = new QwtPlotCurve("Heave");
    heaveCurve->setPen(QPen(QColor(200, 0, 0)));
    heaveCurve->setStyle(QwtPlotCurve::Steps);
    heaveCurve->attach(plot);
    ui->plotLayout->addWidget(plot);
    ui->plotLayout->addItem(ui->plotLayout->takeAt(0));
    ui->plotLayout->addItem(ui->plotLayout->takeAt(0));
    mapselector = new Mapselector();
    ui->mapSelectorLayout->addWidget(mapselector);

    timeInterval = 40;
    scaleFactor = -100;
    prevSquareHeight = 0;
    derivative = 0;
    mapSize = QSize(1,1);
    connect(mapselector, SIGNAL(squareSelected(QPoint)), this, SLOT(changeSelectedSquare(QPoint)));

    QPixmap aftPixmap;
    aftPixmap.load(QString("supplyvessel-aft.jpg"));
    aftView.setPixmap(aftPixmap);
    aftView.setTransformOriginPoint(aftView.pixmap().width() / 2, aftView.pixmap().height() * 3 / 4);

    QPixmap sidePixmap;
    sidePixmap.load(QString("supplyvessel-side.jpg"));
    sideView.setPixmap(sidePixmap);
    sideView.setTransformOriginPoint(sideView.pixmap().width() / 4, sideView.pixmap().height() / 2);

    QPixmap logoPixmap;
    logoPixmap.load(QString("logo.png"));
    logoItem.setPixmap(logoPixmap);

    // aftView.scale(0.22, 0.22);
    // sideView.scale(0.22, 0.22);
    sideView.setPos(145, 0);

    logoScene->addItem(&logoItem);
    vesselScene->addItem(&aftView);
    vesselScene->addItem(&sideView);
    ui->vesselView->setScene(vesselScene);

    setScale(QString("m"), -100, 1);
    unitsInFeet = false;

    topFrame->videoItem->setHighlightedSquare(QPoint(0, 0), true);
    leftFrame->videoItem->setHighlightedSquare(QPoint(0, 0), true);
    rightFrame->videoItem->setHighlightedSquare(QPoint(0, 0), true);
}

void Presentation::clear()
{
    timeData.clear();
    for(int i = 0; i < heaveData.size(); i++) {
        heaveData[i].clear();
    }
}

void Presentation::setScale(QString unit, float scale, float plotDeltaTick)
{
    this->plotDeltaTick = plotDeltaTick;
    this->scaleFactor = scale;
    this->unit = unit;
    wavemap->setScaleFactor(scale, QString(" ") + unit);
    QString title("height [");
    plot->setAxisTitle(QwtPlot::yLeft, title + unit + QString("]"));
    plot->replot();
}



void Presentation::setRollPitch(float roll, float pitch, float z)
{
    aftView.setRotation(roll * 180.0f / M_PI);
    aftView.setY(z * 500);
    sideView.setRotation(- pitch * 180.0f / M_PI);
    sideView.setY(z * 500);
}

void Presentation::changeSelectedSquare(QPoint square)
{
    selectedSquare = square;
    topFrame->videoItem->setHighlightedSquare(square, true);
    leftFrame->videoItem->setHighlightedSquare(square, true);
    rightFrame->videoItem->setHighlightedSquare(square, true);
}

Wavemap * Presentation::getWavemap()
{
    return wavemap;
}

Presentation::~Presentation()
{
    delete ui;
}

VideoFrame * Presentation::getFrameLeft()
{
    return leftFrame;
}

VideoFrame * Presentation::getFrameRight()
{
    return rightFrame;
}

VideoFrame * Presentation::getFrameTop()
{
    return topFrame;
}

void Presentation::resizeEvent(QResizeEvent * even)
{
    //resizeItem(leftFrame, ui->cameraLeft->size());
    //resizeItem(rightFrame, ui->cameraRight->size());
    //resizeItem(topFrame, ui->cameraTop->size());

    /*QSize viewSize = ui->cameraLeft->size();
    QRect videoSize = leftFrame->boundingRect().toRect();
    float scaleX = (float) viewSize.width()/ (float) videoSize.width() ;
    float scaleY = (float) viewSize.height()/ (float) videoSize.height() ;

    float scale;
    if (scaleX < scaleY) {
        scale = scaleX;
    } else {
        scale = scaleY;
    }
    ui->cameraLeft->resetTransform();
    ui->cameraLeft->scale(scale, scale);

    resizeItem(wavemap, ui->waveMap->size());*/

    scaleView(leftFrame->boundingRect().toRect(), ui->cameraLeft);
    scaleView(rightFrame->boundingRect().toRect(), ui->cameraRight);
    scaleView(topFrame->boundingRect().toRect(), ui->cameraTop);
}

void Presentation::setDeckVisible(bool visible)
{
    leftFrame->videoItem->setDeckVisible(visible);
    rightFrame->videoItem->setDeckVisible(visible);
    topFrame->videoItem->setDeckVisible(visible);
}


void Presentation::setDeckPosition(QVector3D pos, QVector3D ang)
{
    leftFrame->videoItem->setDeck(pos, ang);
    rightFrame->videoItem->setDeck(pos, ang);
    topFrame->videoItem->setDeck(pos, ang);
}

void Presentation::scaleView(QRect itemSize, QGraphicsView * view)
{
    float scaleX = (float) view->width() / (float) itemSize.width();
    float scaleY = (float) view->height()/ (float) itemSize.height();

    float scale;
    if (scaleX < scaleY) {
        scale = scaleX;
    } else {
        scale = scaleY;
    }
    view->resetTransform();
    view->scale(scale * 0.98, scale * 0.98);
}

void Presentation::setTrackingInfo(QString text)
{
    leftFrame->setOSD(text);
    rightFrame->setOSD(text);
}

void Presentation::setWavemap(float time, float * zPos, float * amplitude, QSize squares, float squareSize)
{

    wavemap->setWavemap(amplitude, squares.width(), squares.height());
    leftFrame->videoItem->setDeckSize(squares, squareSize);
    rightFrame->videoItem->setDeckSize(squares, squareSize);
    topFrame->videoItem->setDeckSize(squares, squareSize);
    mapselector->setSize(squares);

    if (squares != mapSize) {
        heaveData.clear();
        timeData.clear();
        mapSize = squares;
        for(int i = 0; i < squares.width() * squares.height(); i++) {
            QVector <double> data;
            heaveData.append(data);
        }

    }

    QString num;
    num.setNum(zPos[selectedSquare.x() + selectedSquare.y() * mapSize.width()] * scaleFactor, 'f', 1);
    num.append(QString(" ") + unit);
    ui->heightMeasure->setText(num);

    num.setNum(fabs(amplitude[selectedSquare.x() + selectedSquare.y() * mapSize.width()] * scaleFactor), 'f', 1);
    num.append(QString(" ") + unit);
    ui->heightAmplitude->setText(num);

    // Append new data
    timeData.append(time);
    for(int i = 0; i < mapSize.width() * mapSize.height(); i++) {
        heaveData[i].append(zPos[i]);
    }

    // Remove old data
    if (timeData.size() > 0) {

        while (timeData.last() - timeData.first() > timeInterval) {
            if (timeData.size() == 0) {
                break;
            }

            timeData.remove(0);
            for(int i = 0; i < mapSize.width() * mapSize.height(); i++) {
                heaveData[i].remove(0);
            }
        }

    }

    // Find proper axis scale
    float min, max;
    bool first = true;
    for(int i = 0; i < mapSize.width() * mapSize.height(); i++) {
        for(int j = 0; j < heaveData.at(i).size(); j++) {
            float height = heaveData.at(i).at(j);
            if (first) {
                first = false;
                min = height;
                max = height;
            } else {
                if (height < min) {
                    min = height;
                }
                if (height > max) {
                    max = height;
                }
            }
        }
    }

    // Find derivative of square height
    float selectedSquarePos = zPos[selectedSquare.x() + selectedSquare.y() * mapSize.width()];
    float derivativeCurrent = selectedSquarePos - prevSquareHeight;
    lowpass(derivativeCurrent, derivative, 0.05);

    int red = 0;
    int green = 0;

    if (derivative > 0.00001) {
        directionUp = false;
    } else if (derivative < 0.0003) {
        directionUp = true;
    }

    if (directionUp) {
        red = 255;
    } else {
        green = 255;
    }

    leftFrame->videoItem->setHighlightedSquareColor(QColor(red, green, 0));
    rightFrame->videoItem->setHighlightedSquareColor(QColor(red, green, 0));

    prevSquareHeight = selectedSquarePos;

    QVector <double> selectedHeaveData = heaveData.at(selectedSquare.x() + selectedSquare.y() * mapSize.width());
    QVector <double> converted;
    for (int i = 0; i < selectedHeaveData.size(); i++) {
        converted.append(selectedHeaveData.at(i) * scaleFactor);
    }

    float minConverted = min  * scaleFactor;
    float maxConverted = max * scaleFactor;
    if (scaleFactor < 0) {
        float tmp = ceil(minConverted);
        minConverted = floor(maxConverted);
        maxConverted = tmp;
    } else {
        minConverted = floor(minConverted);
        maxConverted = floor(maxConverted);
    }
    heaveCurve->setData(timeData, converted);
    plot->setAxisScale(QwtPlot::xBottom, timeData.first(), timeData.last(), 5);
    plot->setAxisScale(QwtPlot::yLeft, minConverted, maxConverted, plotDeltaTick);
    plot->replot();

}

/**
  * Updates "output" according to input. When t -> infinity =>  output = input
  */
void Presentation::lowpass(float input, float &output, float timeconstant)
{
    output += (input - output) * timeconstant;
}

void Presentation::resizeItem(QGraphicsItem * item, QSize viewSize)
{
    QRect videoSize = item->boundingRect().toRect();
    float scaleX = (float) viewSize.width()/ (float) videoSize.width() ;
    float scaleY = (float) viewSize.height()/ (float) videoSize.height() ;

    float scale;
    if (scaleX < scaleY) {
        scale = scaleX;
    } else {
        scale = scaleY;
    }
    item->setScale(scale);
}

void Presentation::changeEvent(QEvent *e)
{
    QWidget::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

VideoFrame * Presentation::getVideoFrame(int num)
{
    if (num == 0) {
        return leftFrame;
    } else if (num == 1) {
        return rightFrame;
    } else if (num == 2) {
        return topFrame;
    }

    return NULL;
}

void Presentation::on_toggleUnit_clicked()
{
    if (unitsInFeet) {
        unitsInFeet = false;
        setScale(QString("m"), -100, 1);
        ui->toggleUnit->setText(QString("Show in feet"));
    } else {
        unitsInFeet = true;
        setScale(QString("ft"), -100 * 3.2808, 3);
        ui->toggleUnit->setText(QString("Show in meter"));
    }
}
