#include "sourceselector.h"
#include "ui_sourceselector.h"
#include "videofile.h"
#include <QFileDialog>
#include <QMessageBox>
#include <QDebug>
#include <QGLWidget>
#include <QInputDialog>

#include "videosource.h"
#include "videoframe.h"
#include "videocamera.h"

SourceSelector::SourceSelector(QWidget *parent) :
    QDialog(parent),
    ui(new Ui::SourceSelector)
{
    ui->setupUi(this);
    ui->graphicsView->setScene(&graphicsScene);
    ui->graphicsView->setViewport(new QGLWidget());
    connect(&(this->timer), SIGNAL(timeout()), this, SLOT(updateFrame()));
    videoRunning = false;

}

SourceSelector::~SourceSelector()
{
    stopVideo();
    if (videoRunning) {
        stopVideo();
    }
    delete ui;

}

void SourceSelector::changeEvent(QEvent *e)
{
    QDialog::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

void SourceSelector::resizeEvent(QResizeEvent * event)
{
    if (videoRunning) {
        scaleView(videoFrame->boundingRect().toRect(), ui->graphicsView);
    }
}

void SourceSelector::scaleView(QRect itemSize, QGraphicsView * view)
{
    float scaleX = (float) view->width() / (float) itemSize.width();
    float scaleY = (float) view->height()/ (float) itemSize.height();

    float scale;
    if (scaleX < scaleY) {
        scale = scaleX;
    } else {
        scale = scaleY;
    }
    view->resetTransform();
    view->scale(scale * 0.98, scale * 0.98);
}

void SourceSelector::updateFrame()
{
    if (!videoRunning) {
        timer.stop();
        qDebug() << "No video source, cannot display.";
        return;
    }
    if (!videoSource->captureFrame()) {
        stopVideo();
        return;
    }
    videoFrame->setFrameCV(videoSource->getFrame());
}

void SourceSelector::on_videoFileButton_clicked()
{
    QString filename = QFileDialog::getOpenFileName(this, tr("Open Video"), ".", tr("Video Files(*.avi *.mpg *.mpeg *.wmv *.mov *.ogm)"));
    if (filename.isNull()) {
        return;
    }

    startVideo(VIDEO_FILE, filename);
}

void SourceSelector::on_openCVButton_clicked()
{
    bool ok;
    int captureId = QInputDialog::getInt(this, tr("Select source"), tr("OpenCV capture ID:"), 0, 0, 400, 1, &ok);
    if (ok == false) {
        return;
    }

    startVideo(OPENCV_CAMERA, QString::number(captureId));
}

void SourceSelector::startVideo(VideoSourceType videoSourceType, QString id)
{
    if (videoRunning) {
        stopVideo();
    }

    switch(videoSourceType) {
    case OPENCV_CAMERA:
        videoSource = new VideoCamera(id);
        break;
    case VIDEO_FILE:
        videoSource = new VideoFile(id);
        break;
    }

    if (!videoSource->init()) {
        int ret = QMessageBox::critical(this, tr("Shiptrack"), tr("Failed to initialize the video source."), QMessageBox::Ok);
        delete videoSource;
        return;
    }

    this->videoSourceType = videoSourceType;
    this->videoSourceId = id;

    videoFrame = new VideoFrame();
    graphicsScene.addItem(videoFrame);
    videoRunning = true;
    updateFrame();
    scaleView(videoFrame->boundingRect().toRect(), ui->graphicsView);
    timer.start(50);
}

void SourceSelector::stopVideo()
{
    if (videoRunning) {
        timer.stop();
        videoSource->close();
        graphicsScene.removeItem(videoFrame);
        delete videoFrame;
        delete videoSource;
        videoRunning = false;
    }
}

VideoSourceType SourceSelector::getVideoSourceType()
{
    return videoSourceType;
}

QString SourceSelector::getVideoSourceId()
{
    return videoSourceId;
}

void SourceSelector::on_buttonBox_accepted()
{
    accept();
}

void SourceSelector::on_buttonBox_rejected()
{    
    reject();
}

void SourceSelector::closeEvent ( QCloseEvent * event )
{
    stopVideo();
}
