
#include <QDebug>

#include <cv.h>
#include <math.h>
#include <vector>
#include <highgui.h>

#include "imageprocessing.h"
#include "tracker.h"


Tracker::Tracker(unsigned int width, unsigned int height, ImageProcessing * parent)
{
    int roiWidth = 200;
    int roiHeight = 200;

    templ = cvCreateImage(cvSize(1, 1), IPL_DEPTH_8U, 3);
    corrResult = cvCreateImage(cvSize(1, 1), IPL_DEPTH_32F, 1);

    firstPos = true;
    subPixelScale = 1;

    match = 1.0;
    roi = QRect(0, 0, roiWidth, roiHeight);
    roiImage = cvCreateImage(cvSize(roi.width(), roi.height()), IPL_DEPTH_8U, 3);
    filteredRoiImage = cvCreateImage(cvSize(roi.width(), roi.height()), IPL_DEPTH_8U, 3);
    startTime = QTime::currentTime();
    successiveTries = 0;
}


int Tracker::getSuccessiveTries()
{
    return successiveTries;
}

void Tracker::incrSuccessiveTries()
{
    successiveTries++;
}

Tracker::~Tracker()
{
    cvReleaseImage(&roiImage);
    cvReleaseImage(&filteredRoiImage);
    cvReleaseImage(&templ);
    cvReleaseImage(&corrResult);
}

QTime Tracker::getStartTime()
{
    return startTime;
}

void Tracker::setFrame(IplImage * frame)
{
    this->frame = frame;
}

QRect Tracker::getRoi()
{
    return this->roi;
}

IplImage * Tracker::getRoiImage()
{
    return roiImage;
}

IplImage * Tracker::getFilteredRoiImage()
{
    return filteredRoiImage;
}

QPointF Tracker::getPos()
{
    return pos;
}

QPointF Tracker::getPrevPos()
{
    return prevPos;
}

QPointF Tracker::getAcceleration()
{
    return accel;
}

QPointF Tracker::track()
{
    IplImage * frame = cvCreateImageHeader(cvGetSize(this->frame), IPL_DEPTH_8U, 3);
    frame->imageData = this->frame->imageData;

    cvSetImageROI(frame, cvRect(roi.x(), roi.y(), roi.width(), roi.height()));
    cvCopy(frame, roiImage);
    cvReleaseImage(&frame);

    cvMatchTemplate(roiImage, templ, corrResult, CV_TM_CCORR_NORMED);


    cvZero(filteredRoiImage);

    for(int x = 0; x < corrResult->width; x++) {
        for(int y = 0; y < corrResult->height; y++) {
            unsigned char value = 255 * ((float *) (corrResult->imageData))[x + y * corrResult->width];
            int index = x * filteredRoiImage->nChannels + y * filteredRoiImage->widthStep;
            filteredRoiImage->imageData[index] = value;
            filteredRoiImage->imageData[index + 1] = value;
            filteredRoiImage->imageData[index + 2] = value;
        }
    }

    float max = 0;
    int maxPos = 0;

    for(int i = 0; i < corrResult->width * corrResult->height; i++) {
        float val = ((float *) corrResult->imageData)[i];
        if (val > max) {
            max = ((float *)corrResult->imageData)[i];
            maxPos = i;
        }
    }

    QPoint intPos(maxPos % corrResult->width, maxPos / corrResult->height);

    /* cvNamedWindow("roi");
    cvShowImage("roi", roiImage);

    cvNamedWindow("roi");
    cvShowImage("corrresult", corrResult);
    sleep(1); */

    templRect = QRect(intPos.x(), intPos.y(), templ->width, templ->height);

    // Subpixel refinement
    cvSetImageROI(roiImage, cvRect(intPos.x() - 1, intPos.y() - 1, templ->width + 2, templ->height + 2));
    cvResize(roiImage, roiImageSub);
    cvResetImageROI(roiImage);

    cvMatchTemplate(roiImageSub, templSub, corrResultSub, CV_TM_CCORR_NORMED);

    max = 0;
    maxPos = 0;
    for(int i = 0; i < corrResultSub->width * corrResultSub->height; i++) {
        float val = ((float *) corrResultSub->imageData)[i];
        if (val > max) {
            max = ((float *)corrResultSub->imageData)[i];
            maxPos = i;
        }
    }

    match = max;

    QPointF subPos( (maxPos % corrResultSub->width) / (float) subPixelScale - 1.0f, (maxPos / corrResultSub->height) / (float) subPixelScale - 1.0f);
    prevPos = pos;
    pos = roi.topLeft() + intPos + subPos + QPoint(templRect.width() / 2, templRect.height() / 2);

    if (!firstPos) {
        prevVelo = velo;
        velo = pos - prevPos;
        accel = velo - prevVelo;
    }
    firstPos = false;
    //qDebug() << "pos: " << pos.toPoint() << "\t velo: " << velo.toPoint() << "\t accel: " << accel.toPoint();


    QPoint offset(templRect.left() + templRect.width() / 2 - roi.width() / 2, templRect.top() + templRect.height() / 2 - roi.height() / 2);
    //if (abs(offset.x()) > 10 || abs(offset.y()) > 10) {
        roi.moveTopLeft(roi.topLeft() + offset);
        templRect.moveTopLeft(templRect.topLeft() - offset);
        correctRoiPos();
    //}


    return pos;
}

float Tracker::getMatch()
{
    return match;
}

QRect Tracker::getTemplateRect()
{
    return templRect;
}


void Tracker::correctRoiPos()
{
    if (roi.x() + roi.width() > frame->width) {
        int offsetX = roi.x() + roi.width() - frame->width;
        roi.moveLeft(roi.left() - offsetX);
        templRect.moveLeft(templRect.left() + offsetX);
    } else if (roi.x() < 0) {
        int offsetX = -roi.x();
        roi.moveLeft(roi.left() + offsetX);
        templRect.moveLeft(templRect.left() - offsetX);
    }

    if (roi.y() + roi.height() > frame->height) {
        int offsetY = roi.y() + roi.height() - frame->height;
        roi.moveTop(roi.top() - offsetY);
        templRect.moveTop(templRect.top() + offsetY);
    } else if (roi.y() < 0) {
        int offsetY = -roi.top();
        roi.moveTop(roi.top() + offsetY);
        templRect.moveTop(offsetY);
    }
}

void Tracker::setTargetRect(QRect rect)
{

    cvReleaseImage(&roiImage);
    cvReleaseImage(&filteredRoiImage);
    roi = QRect(0, 0, rect.width() + 50, rect.height() + 50);
    roiImage = cvCreateImage(cvSize(roi.width(), roi.height()), IPL_DEPTH_8U, 3);
    filteredRoiImage = cvCreateImage(cvSize(roi.width(), roi.height()), IPL_DEPTH_8U, 3);

    roi.moveLeft(rect.x() + rect.width() / 2 - roi.width() / 2);
    roi.moveTop(rect.y() + rect.height() / 2 - roi.height() / 2);
    correctRoiPos();

    cvReleaseImage(&corrResult);

    if (roi.width() - rect.width() + 1 < 1) {
        rect.setWidth(roi.width());
    }

    if (roi.height() - rect.height() + 1 < 1) {
        rect.setHeight(roi.height());
    }


    cvReleaseImage(&templ);
    templ = cvCreateImage(cvSize(rect.width(), rect.height()), IPL_DEPTH_8U, 3);
    corrResult = cvCreateImage(cvSize(roi.width() - rect.width() + 1, roi.height() - rect.height() + 1), IPL_DEPTH_32F, 1);

    roiImageSub = cvCreateImage(cvSize( (templ->width + 2) * subPixelScale, (templ->height + 2) * subPixelScale), IPL_DEPTH_8U, 3);
    templSub = cvCreateImage(cvSize(subPixelScale * templ->width, subPixelScale * templ->height), IPL_DEPTH_8U, 3);
    corrResultSub = cvCreateImage(cvSize(roiImageSub->width - templSub->width + 1, roiImageSub->height - templSub->height + 1), IPL_DEPTH_32F, 1);

    cvSetImageROI(frame, cvRect(rect.x(), rect.y(), rect.width(), rect.height()));

    cvCopy(frame, templ);

    cvResetImageROI(frame);

    cvResize(templ, templSub);

}

IplImage * Tracker::getTemplate()
{
    return templ;
}

