#include <QDebug>

#include "trackingmanager.h"
#include "positionestimator.h"
#include "camera.h"
#include "vesselsearch.h"
#include "imageprocessing.h"
#include "util.h"

TrackingManager::TrackingManager()
{
    timeSinceTrackingChange.restart();
    primaryCamera = 0;
    secondaryCamera = 1;
    addTrackerToPrimary = false;
    camerasOk = false;
    findCorrespondence = false;
    findInitialPos = false;
    aftestPointBodyOffsetSet = false;
    minMatch = 0.97;
    qDebug() << "changing state to SEARCHING";
    trackingState = SEARCHING;
    aftestTrackerId = -1;
    adjustWithCheckImageTimer = 0;
    maxLockOffset = 0.01;
    aftToBodyOrigin = 0.15;
    startVesselSearch = false;
    vesselHasBeenFound = false;   
    doAdjustWithCheckImage = false;
}


void TrackingManager::checkTracking()
{
    if (!camerasOk) {
        return;
    }

    stableTrackers = findStableTrackers(cameras.at(primaryCamera), 5, findTrackersInPositionGroup(cameras.at(primaryCamera)));
    if (stableTrackers.size() > 0) {
        findCorrespondence = true;
        cameras.at(primaryCamera)->getImageProcessing()->setVesselSearch();
    }

    int trackingChangeDelay = 0;

    if (trackingState == SEARCHING) {
        trackingChangeDelay = 0;
        if (cameras.at(primaryCamera)->getImageProcessing()->getNumberOfTrackers() > 0) {
            qDebug() << "changing state to FINDING_AFT";
            trackingState = FINDING_AFT;
            return;
        }
        addTrackerToPrimary = true;
    } else if (trackingState == FINDING_AFT) {
        if (cameras.at(primaryCamera)->getImageProcessing()->getNumberOfTrackers() == 0) {
            qDebug() << "changing state to SEARCHING";
            trackingState = SEARCHING;
            return;
        }
        trackingChangeDelay = 7000;
        addTrackerToPrimary = true;

    } else if (trackingState == ISOLATING_AFT) {
        if (cameras.at(primaryCamera)->getImageProcessing()->getNumberOfTrackers() == 0) {
            qDebug() << "changing state to SEARCHING";
            trackingState = SEARCHING;
            return;
        }
        
        findCorrespondence = stableTrackers.size() > 0;
        addTrackerToPrimary = true;
        trackingChangeDelay = 1000;

    } else if (trackingState == ADDING_TRACKERS) {
        if (cameras.at(primaryCamera)->getImageProcessing()->getNumberOfTrackers() == 0) {
            qDebug() << "changing state to SEARCHING";
            trackingState = SEARCHING;
            return;
        }

        QPointF aftestTrackerPos = cameras.at(primaryCamera)->getImageProcessing()->getTrackerPos(aftestTrackerId);
        if (aftestTrackerPos == QPointF(-1, -1)) {
            qDebug() << "Aftest tracker disappeared so reverting to SEARCHING";
            trackingState = SEARCHING;
            cameras.at(primaryCamera)->getImageProcessing()->removeAllTrackers();
            return;
        }

        if (positionEstimator->inTracking() == true) {

            if (doAdjustWithCheckImage) {
                adjustWithCheckImageTimer++;
                if (adjustWithCheckImageTimer > 6) {
                    adjustWithCheckImageTimer = 0;
                    if (adjustWithCheckImage()) {
                        qDebug() << "changing state to IN_TRACKING";
                        trackingState = IN_TRACKING;
                        return;
                    }
                }
            } else {
                findInitialPos = true;
            }
        }

        if (positionEstimator->numTrackerGroups() < 6) {
            addTrackerToPrimary = true;
        }
        aftestPointBodyOffsetSet = false;
        trackingChangeDelay = 300;
        findCorrespondence = stableTrackers.size() > 0;

    } else if (trackingState == IN_TRACKING) {
        if (cameras.at(primaryCamera)->getImageProcessing()->getNumberOfTrackers() < 4) {
            qDebug() << "changing state to ADDING_TRACKERS";
            trackingState = ADDING_TRACKERS;
            return;
        }
        if (!positionEstimator->inTracking()) {
            qDebug() << "changing state to ADDING_TRACKERS";
            trackingState = ADDING_TRACKERS;
            return;
        }

        if (cameras.at(primaryCamera)->getImageProcessing()->getNumberOfTrackers() < 6) {
            addTrackerToPrimary = true;
        }
        trackingChangeDelay = 1000;
        findCorrespondence = stableTrackers.size() > 0;

        /* Update the body vector which points to the point which originates from the aftest tracker.
           By doing this, this point can be used instead when the aftest tracker fails. */
        if (!aftestPointBodyOffsetSet) {
            int aftestTrackerGroup = findTrackerGroupByTracker(cameras.at(primaryCamera), aftestTrackerId);
            if (aftestTrackerGroup != -1) {
                aftestPointBodyOffset = positionEstimator->findBodyVector(positionEstimator->getPosition(aftestTrackerGroup));
                aftestPointBodyOffsetSet = true;
            }
        }
    }

    if ( timeSinceTrackingChange.elapsed() > trackingChangeDelay) {
        if (addTrackerToPrimary || findCorrespondence || findInitialPos) {
            if (!startVesselSearch) {
                qDebug() << "Start vessel search because: addTrackerToPrimary " << addTrackerToPrimary << " || findCorrespondence " << findCorrespondence << " || findInitialPos " << findInitialPos;
                startVesselSearch = true;
            }
        }
    }

    if (startVesselSearch) {
        cameras.at(primaryCamera)->getImageProcessing()->setVesselSearch();
    }

    if (vesselHasBeenFound) {
        runPendingVesselFound();
    }
}

bool TrackingManager::adjustWithCheckImage()
{
    QVector3D currentCentroid = positionEstimator->getCurrentCentroid();
    if (fabs(currentCentroid.z()) > maxLockOffset) {
        qDebug() << "Z-position is not within lock height";
        return false;
    }

    QPolygon vesselPolygon = cameras.at(primaryCamera)->getImageProcessing()->getVesselSearch()->getVesselPolygon();
    IplImage * checkImage = cameras.at(primaryCamera)->getCheckImage();
    if (checkImage == NULL) {
        qDebug() << "Check image is not set";
        return true;
    }
    float match;
    QRect rect = cameras.at(primaryCamera)->getImageProcessing()->templateMatch(checkImage,vesselPolygon, match);
    qDebug() << "findCheckImage: " << rect << " match: " << match;
    if (match < 0.99) {
        qDebug() << "check match was not good enough (" << match << ")";
        return false;
    }
    QPointF xyPos = cameras.at(primaryCamera)->getImageProcessing()->getVesselSearch()->mapToXY(rect.center());
    //CvMat * eye = cvCreateMat(3, 3, CV_32F);
    //cvSetIdentity(eye);
    //positionEstimator->setPositionAndAttitude(QVector3D(xyPos.x(), xyPos.y(), currentCentroid.z()), eye);
    //cvReleaseMat(&eye);
    positionEstimator->setBodyOrigin(QVector3D(xyPos.x() + 0.12, xyPos.y(), currentCentroid.z()));

    return true;
}


/**
  * Finds the position id group for a tracker
  */
int TrackingManager::findTrackerGroupByTracker(Camera * camera, int trackerId)
{
    QVector <TrackerGroup> trackerGroups = positionEstimator->getTrackerGroups();
    for(int i = 0; i < trackerGroups.size(); i++) {
        QVector <TrackerRef> trackerRef = trackerGroups.at(i).trackerRef;
        for(int j = 0; j < trackerRef.size(); j++) {
            if (trackerRef.at(j).imageProcessing->getParent() == camera && trackerRef.at(j).trackerId == trackerId) {
                return i;
            }
        }
    }
    return -1;
}

/**
  * Finds the trackers in camera which belong to a position group
  */
QVector <int> TrackingManager::findTrackersInPositionGroup(Camera * camera)
{
    QVector <int> trackersInGroup;

    QVector <TrackerGroup> trackerGroups = positionEstimator->getTrackerGroups();
    for(int i = 0; i < trackerGroups.size(); i++) {
        QVector <TrackerRef> trackerRef = trackerGroups.at(i).trackerRef;
        for(int j = 0; j < trackerRef.size(); j++) {
            if (trackerRef.at(j).imageProcessing->getParent() == camera) {
                trackersInGroup.append(trackerRef.at(j).trackerId);
            }
        }
    }
    return trackersInGroup;
}

void TrackingManager::setPositionEstimator(PositionEstimator * positionEstimator)
{
    this->positionEstimator = positionEstimator;
}

void TrackingManager::addCamera(Camera * camera)
{
    cameras.append(camera);
    connect(camera->getImageProcessing()->getVesselSearch(), SIGNAL(vesselEdgesFound(QLine,QLine)), this, SLOT(vesselFound(QLine,QLine)));
    camerasOk = (primaryCamera < cameras.size() && secondaryCamera < cameras.size());    
    if (camerasOk) {
        if (cameras.at(primaryCamera)->getCheckImage() != NULL) {
            doAdjustWithCheckImage = true;
        }
    }
}

void  TrackingManager::removeCamera(Camera * camera)
{
    for(int i = 0; i < cameras.size(); i++) {
        if (cameras.at(i) == camera) {
            cameras.remove(i);
            connect(camera->getImageProcessing()->getVesselSearch(), SIGNAL(vesselEdgesFound(QLine,QLine)), this, SLOT(vesselFound(QLine,QLine)));
            camerasOk = (primaryCamera < cameras.size() && secondaryCamera < cameras.size());
            return;
        }
    }
}

/**
  * Returns a list of the indices of the polygons which _do_ contain a tracker
  */
QVector <int> TrackingManager::findTrackedPolygons(Camera * camera, QVector <QPolygon> polygonsImage)
{
    QVector <int> untracked = findUntrackedPolygons(camera, polygonsImage);
    bool * polygonHasTracker = new bool[polygonsImage.size()];

    for(int i = 0; i < polygonsImage.size(); i++) {
        polygonHasTracker[i] = true;
    }

    for(int i = 0; i < untracked.size(); i++) {
        polygonHasTracker[untracked.at(i)] = false;
    }

    QVector <int> indicesWithTracker;
    for(int i = 0; i < polygonsImage.size(); i++) {
        if (polygonHasTracker[i] == true) {
            indicesWithTracker.append(i);
        }
    }
    delete[] polygonHasTracker;
    return indicesWithTracker;

}

/**
  * Returns a list of the indices of the polygons which _do not_ contain a tracker
  */
QVector <int> TrackingManager::findUntrackedPolygons(Camera * camera, QVector <QPolygon> polygonsImage)
{
    bool * polygonHasTracker = new bool[polygonsImage.size()];
    for(int i = 0; i < polygonsImage.size(); i++) {
        polygonHasTracker[i] = false;
    }

    QHash <int, QPointF> trackerPositions = camera->getImageProcessing()->getTrackerPositions();
    QList <int> keys = trackerPositions.keys();

    for(int i = 0; i < keys.size(); i++) {
        QPointF position = trackerPositions.value(keys.at(i));
        for(int j = 0; j < polygonsImage.size(); j++) {
            if (polygonsImage.at(j).containsPoint(position.toPoint(), Qt::OddEvenFill)) {
                //qDebug() << "polygon no. " << j << " does not have tracker";
                polygonHasTracker[j] = true;
            }
        }
    }

    QVector <int> indicesWithoutTracker;
    for(int i = 0; i < polygonsImage.size(); i++) {
        if (polygonHasTracker[i] == false) {
            indicesWithoutTracker.append(i);
        }
    }

    delete[] polygonHasTracker;
    return indicesWithoutTracker;
}

int TrackingManager::findPolygonFromPosition(QVector <QPolygon> polygons, QPointF position)
{
    for(int j = 0; j < polygons.size(); j++) {
        if (polygons.at(j).containsPoint(position.toPoint(), Qt::OddEvenFill)) {
            return j;
        }
    }
    return -1;
}

QVector <QPoint> TrackingManager::findGoodFeaturesInPolygons(Camera * camera, QVector <QPolygon> polygons, int maxPoints)
{
    QVector <QPoint> goodPoints;
    for(int i = 0; i < polygons.size(); i++) {
        if (goodPoints.size() >= maxPoints) {
            break;
        }
        goodPoints += camera->getImageProcessing()->findGoodFeatures(polygons.at(i), maxPoints - goodPoints.size());
    }
    return goodPoints;
}

/**
  * Add trackers to areas which do not have a tracker. This is a vessel dependent function.
  */
int TrackingManager::addTrackersToUntrackedAreas(Camera * camera, QVector <QPolygon> polygons, int aftestIdx)
{
    // Vectors of untracked polygons in the desired areas.
    QVector <QPolygon> aftLeft, aftRight, frontLeft, frontRight;
    int aftLeftTracked, aftRightTracked, frontLeftTracked, frontRightTracked;
    aftLeftTracked = aftRightTracked = frontLeftTracked = frontRightTracked = 0;

    QVector <int> trackedPolygonsIdx = findTrackedPolygons(camera, polygons);

    int aftStop = 7;
    int frontStart = 13;
    int frontEnd = 25;
    int maxTrackers = 10;

    for(int i = aftestIdx; i < polygons.size(); i++) {

        if (i < aftStop + aftestIdx) {
            if (!trackedPolygonsIdx.contains(i)) {
                if (i % 2 == 0) {
                    aftLeft.append(polygons.at(i));
                } else {
                    aftRight.append(polygons.at(i));
                }
            } else {
                if (i % 2 == 0) {
                    aftLeftTracked++;
                } else {
                    aftRightTracked++;
                }
            }


        } else if (i > frontStart + aftestIdx && i < frontEnd + aftestIdx){

            if (!trackedPolygonsIdx.contains(i)) {
                if (i % 2 == 0) {
                    frontLeft.append(polygons.at(i));
                } else {
                    frontRight.append(polygons.at(i));
                }
            } else {
                if (i % 2 == 0) {
                    frontLeftTracked++;
                } else {
                    frontRightTracked++;
                }
            }

        }
    }

    // qDebug() << "aftLeft " << aftLeftTracked << " " << aftLeft;
    // qDebug() << "aftRight " << aftRightTracked << " " << aftRight;
    // qDebug() << "frontLeft " << frontLeftTracked << " " << frontLeft;
    // qDebug() << "frontRight " << frontRightTracked << " " << frontRight;

    QVector <QPoint> pointsToAdd;
    QVector <QPoint> goodPointsAftLeft, goodPointsAftRight, goodPointsFrontLeft, goodPointsFrontRight;

    if (aftLeftTracked < 2) {
        goodPointsAftLeft = findGoodFeaturesInPolygons(camera, aftLeft, 3);
    }

    if (aftRightTracked < 2) {
        goodPointsAftRight = findGoodFeaturesInPolygons(camera, aftRight, 3);
    }

    if (frontLeftTracked < 2) {
        goodPointsFrontLeft = findGoodFeaturesInPolygons(camera, frontLeft, 3);
    }

    if (frontRightTracked < 2) {
        goodPointsFrontRight = findGoodFeaturesInPolygons(camera, frontRight, 3);
    }

    qDebug() << "Tracked areas: aft port : " << aftLeftTracked << ", aft starboard : " << aftRightTracked << ", front port : " << frontLeftTracked << ", front starboard: " << frontRightTracked;

    // Find region with least trackers
    int numTrackers = 2;
    int region = -1;

    if (aftLeftTracked < numTrackers) {
        numTrackers = aftLeftTracked;
        region = 0;
    }
    if (aftRightTracked < numTrackers) {
        numTrackers = aftRightTracked;
        region = 1;
    }
    if (frontLeftTracked < numTrackers) {
        numTrackers = frontLeftTracked;
        region = 2;
    }
    if (frontRightTracked < numTrackers) {
        numTrackers = frontRightTracked;
        region = 3;
    }

    switch(region) {

    case 0:
        pointsToAdd = goodPointsAftLeft;
        qDebug() << "aft port has the least trackers";
        break;
    case 1:
        pointsToAdd = goodPointsAftRight;
        qDebug() << "aft starboard has the least trackers";
        break;
    case 2:
        pointsToAdd = goodPointsFrontLeft;
        qDebug() << "front port has the least trackers";
        break;
    case 3:
        pointsToAdd = goodPointsFrontRight;
        qDebug() << "front starboard has the least trackers";
        break;
    }

    int numberOfTrackers = camera->getImageProcessing()->getNumberOfTrackers();

    for(int i = 0; i < pointsToAdd.size(); i++) {
        if ( (i + numberOfTrackers + 1) > maxTrackers) {
            break;
        }

        QPoint goodPoint = pointsToAdd.at(i);
        if (distanceToClosestTracker(camera, goodPoint) < 50) {
            continue;
        }
        QRect trackerRect(goodPoint.x() - 25, goodPoint.y() - 25, 50, 50);

        int firstTrackerId = camera->getImageProcessing()->addTracker(trackerRect);
    }

    return 0;
}

/**
  * Returns the current tracking state
  */
TrackingState TrackingManager::getTrackingState()
{
    return trackingState;
}


/**
  * Add trackers according to specified polygons
  */
void TrackingManager::addTrackers(Camera * camera, QVector <QPolygon> polygons)
{

    if (trackingState == SEARCHING) {
        int startPolygon = 2;
        int numPolygons = 7;
        QVector <QPolygon> searchPolygons;

        for(int i = startPolygon; i < startPolygon + numPolygons; i++) {
            if (i > polygons.size() - 1) {
                break;
            }
            searchPolygons.append(polygons.at(i));
        }
        QVector <QPoint> goodPoints = findGoodFeaturesInPolygons(camera, searchPolygons, 5);

        for(int i = 0; i < goodPoints.size(); i++) {
            QPoint goodPoint = goodPoints.at(i);
            QRect trackerRect(goodPoint.x() - 25, goodPoint.y() - 25, 50, 50);
            int firstTrackerId = camera->getImageProcessing()->addTracker(trackerRect);
        }
    } else if (trackingState == FINDING_AFT) {

        qDebug() << "Setting trackingState to ISOLATING_AFT";
        trackingState = ISOLATING_AFT;
        successiveAftNotInPolygon = 0;


    } else if (trackingState == ISOLATING_AFT) {

        QHash <int, QPointF> trackerPos = camera->getImageProcessing()->getTrackerPositions();

        if (trackerPos.size() > 1) {

            int lowestPoly = polygons.size() - 1;

            QList <int> keys = trackerPos.keys();
            int keepTrackerId = -1;
            for(int i = 0; i < keys.size(); i++) {
                int polyNum = findPolygonFromPosition(polygons, trackerPos.value(keys.at(i)));
                if (polyNum < lowestPoly) {
                    lowestPoly = polyNum;
                    keepTrackerId = keys.at(i);
                }
            }

            if (lowestPoly == -1) {
                qDebug() << "Aftest tracking point not in polygon";
                successiveAftNotInPolygon++;
                if (successiveAftNotInPolygon > 6) {
                    trackingState = SEARCHING;
                    cameras.at(primaryCamera)->getImageProcessing()->removeTracker(aftestTrackerId);
                    qDebug() << "changing state to SEARCHING";
                }

            } else {
                qDebug() << "Aftest tracking point is inside poly " << lowestPoly << " which is tracker id " << keepTrackerId;

                for(int i = 0; i < keys.size(); i++) {
                    if (keys.at(i) != keepTrackerId) {
                        camera->getImageProcessing()->removeTracker(keys.at(i));
                    }
                }
                aftestTrackerId = keepTrackerId;
            }

        } else if (trackerPos.size() == 1) {

            QPointF aftestTrackerPos = cameras.at(primaryCamera)->getImageProcessing()->getTrackerPos(aftestTrackerId);
            if (aftestTrackerPos == QPointF(-1, -1)) {
                qDebug() << "Aftest tracker disappeared so reverting to SEARCHING";
                trackingState == SEARCHING;
                cameras.at(primaryCamera)->getImageProcessing()->removeAllTrackers();
                return;
            }

            QVector <int> trackersInGroup = findTrackersInPositionGroup(camera);
            for(int i = 0; i < trackersInGroup.size(); i++) {
                if (trackersInGroup.at(i) == aftestTrackerId) {
                    trackingState = ADDING_TRACKERS;
                    qDebug() << "Aftest tracker has correspondence so changing state to ADDING_TRACKERS";
                }
            }
        }

    } else if (trackingState == ADDING_TRACKERS) {

        QPointF pos = camera->getImageProcessing()->getTrackerPos(aftestTrackerId);

        if (pos == QPointF(-1, -1)) {
            qDebug() << "changing state to SEARCHING";
            trackingState = SEARCHING;
            return;
        }

        int aftestPolyNum = findPolygonFromPosition(polygons, pos);

        if (aftestPolyNum != -1) {
            addTrackersToUntrackedAreas(camera, polygons, aftestPolyNum);
        }

    } else if (trackingState == IN_TRACKING) {

        if (aftestPointBodyOffsetSet) {

            QVector3D pos = positionEstimator->findPosFromBodyOffset(aftestPointBodyOffset);
            CvPoint imagePoint = camera->mapToImage(cvPoint3D32f(pos.x(), pos.y(), pos.z()));
            QPointF imagePos(imagePoint.x, imagePoint.y);

            int aftestPolyNum = findPolygonFromPosition(polygons, imagePos);

            if (aftestPolyNum != -1) {
                addTrackersToUntrackedAreas(camera, polygons, aftestPolyNum);
            }

        } else {
            qDebug() << "Missing aftestPointBodyOffset!";
        }
    }
}

/**
  * Find the euclidean distance in the image plane to the closest tracker
  * given a camera and a point. Returns -1 if no tracker exists.
  */
int TrackingManager::distanceToClosestTracker(Camera * camera, QPoint point)
{
    QHash <int, QPointF> points = camera->getImageProcessing()->getTrackerPositions();

    bool first = true;
    int closestDistance = -1;

    QList <int> keys = points.keys();
    for(int i = 0; i < keys.size(); i++) {
        QPointF current = points.value(keys.at(i));
        int distance = sqrt(pow(point.x() - current.x(), 2) + pow(point.y() - current.y(), 2));
        if (distance < closestDistance || first) {
            closestDistance = distance;
            first = false;
        }
    }

    return closestDistance;

}

/**
  * Called when vessel is found
  */
void TrackingManager::vesselFound(QLine port, QLine starboard)
{
    startVesselSearch = false;
    vesselHasBeenFound = true;
    vesselSearch = (VesselSearch *) QObject::sender();
    timeSinceTrackingChange.restart();
}


void TrackingManager::runPendingVesselFound()
{
    vesselHasBeenFound = false;

    if (!camerasOk) {
        return;
    }

    QVector <QPolygon> polygonsImage = vesselSearch->getVesselPolygonsImage();
    QVector <QPolygonF> polygonsWorld = vesselSearch->getVesselPolygons();
    emit newVesselPolygons(polygonsWorld);

    Camera * camera = vesselSearch->getParent()->getParent();

    if (addTrackerToPrimary) {
        addTrackers(camera, polygonsImage);
        addTrackerToPrimary = false;
    }

    if (findCorrespondence) {
        findCorrespondence = false;
        doFindCorrespondence(camera, polygonsImage, polygonsWorld);
    }

    if (findInitialPos) {
        setInitialPos(vesselSearch->getForwardVec(), vesselSearch->getCenterLinePt());
        findInitialPos = false;
    }

}

void TrackingManager::doFindCorrespondence(Camera * camera, QVector <QPolygon> polygonsImage, QVector <QPolygonF> polygonsWorld)
{
    int maxAddCorrespondence = 2;
    int added = 0;

    // Go through stable trackers and add correspondence
    for(int i = 0; i < stableTrackers.size(); i++) {
        QPointF point = cameras.at(primaryCamera)->getImageProcessing()->getTrackerPos(stableTrackers.at(i));
        if (point == QPointF(-1, -1)) {
            continue;
        }

        /* int polygonIdx = findPolygonFromPosition(polygonsImage, point);

        if (polygonIdx == -1) {
            qDebug() << "tracker is not inside a polygon";
            camera->getImageProcessing()->incrSuccessiveRetries(stableTrackers.at(i));
            if (camera->getImageProcessing()->getTrackerSuccesiveRetries(stableTrackers.at(i)) > 5) {
                qDebug() << "tracker is not inside a polygon too many times removing it...";
                camera->getImageProcessing()->removeTracker(stableTrackers.at(i));
            }
            continue;
        } */


        QPointF xyPos = Util::mapToXY(point, cameras.at(primaryCamera)->getCameraMatrix());
        QPolygonF square;
        square.append(xyPos - QPointF(-0.03, -0.03));
        square.append(xyPos - QPointF(0.03, -0.03));
        square.append(xyPos - QPointF(0.03, 0.03));
        square.append(xyPos - QPointF(-0.03, 0.03));

        int secondTrackerId = addCorrespondingTracker(cameras.at(primaryCamera), stableTrackers.at(i), cameras.at(secondaryCamera), square);

        if (secondTrackerId < 0) {
            continue;
        }

        TrackerGroup group;
        TrackerRef trackerRef;

        trackerRef.imageProcessing = cameras.at(primaryCamera)->getImageProcessing();
        trackerRef.trackerId = stableTrackers.at(i);
        group.trackerRef.append(trackerRef);

        trackerRef.imageProcessing = cameras.at(secondaryCamera)->getImageProcessing();
        trackerRef.trackerId = secondTrackerId;
        group.trackerRef.append(trackerRef);

        positionEstimator->addTrackerGroup(group);
        added++;

        if (added >= (maxAddCorrespondence - 1)) {
            break;
        }
    }
}


/**
  * Add corresponding tracker in "second" camera and return tracker id if sucessful
  */
int TrackingManager::addCorrespondingTracker(Camera * firstCam, int trackerId, Camera * secondCam, QPolygonF worldPolygon)
{
    QPointF point = firstCam->getImageProcessing()->getTrackerPos(trackerId);
    if (point == QPointF(-1, -1)) {
        return -1;
    }
    IplImage * templ = firstCam->getImageProcessing()->getTrackerTemplate(trackerId);
    if (templ == NULL) {
        return -1;
    }

    QLineF epiLine = computeEpipolarLine(firstCam, point, secondCam);

    QPolygon polygonImage;
    for(int p = 0; p < worldPolygon.size(); p++) {
        QPoint imageP = secondCam->getImageProcessing()->getVesselSearch()->mapToImage(worldPolygon.at(p));
        polygonImage.append(imageP);
    }

    try {
        QRect window = polygonImage.boundingRect();
        //window.setWidth(window.width() * 1.5f);
        //window.setHeight(window.height() * 1.5f);
        float match;
        QRect rect = secondCam->getImageProcessing()->templateMatchAlongLine(templ, epiLine.p1(), QVector2D(epiLine.p2() - epiLine.p1()), 70, window, match);
        //QRect rect = secondCam->getImageProcessing()->templateMatchAlongLine(templ, epiLine.p1(), QVector2D(epiLine.p2() - epiLine.p1()), 70, QRect(0,0, 1600, 1200), match);
        cvReleaseImage(&templ);

        if (match < minMatch) {
            qDebug() << "Match was not high enought (" << match << ")";
            throw -1;
        }
        int trackerIdSecondCam = secondCam->getImageProcessing()->addTracker(rect);
        if (trackerIdSecondCam == 0) {
            qDebug() << "Could not add tracker in second cam";
            throw -1;
        }

        return trackerIdSecondCam;

    } catch (int e) {
        firstCam->getImageProcessing()->incrSuccessiveRetries(trackerId);
        int num = firstCam->getImageProcessing()->getTrackerSuccesiveRetries(trackerId);
        if (num > 4) {
            firstCam->getImageProcessing()->removeTracker(trackerId);
            qDebug() << "Could not find correspondong region in other frame, incr successivetries to " << num;
            return -1;
        }
    }

    return -1;
}


/**
  *
  */
void TrackingManager::setInitialPos(QVector2D forward, QPointF center)
{
    float yaw = atan2(forward.y(), forward.x());
    CvMat * bodyToNED = cvCreateMat(3, 3, CV_32F);
    Util::setZRotation(yaw, bodyToNED);

    float aftest = 100;
    int aftestIndex = -1;

    CvMat * NEDToBody = cvCreateMat(3, 3, CV_32F);
    cvTranspose(bodyToNED, NEDToBody);

    QVector <TrackerGroup> trackerGroups = positionEstimator->getTrackerGroups();

    for(int i = 0; i < trackerGroups.size(); i++) {
        QVector3D body = Util::rotateVector(trackerGroups.at(i).position, NEDToBody);
        if (body.x() < aftest) {
            aftestIndex = i;
            aftest = body.x();
        }
    }


    QVector3D aftestPosition;

    if (aftestIndex == -1) {
        qDebug() << "could not set initial pos, unable to find aftest tracker";
    } else {
        aftestPosition = trackerGroups.at(aftestIndex).position;
        qDebug() << "aftest position is " << trackerGroups.at(aftestIndex).position;
    }

    QVector3D starboard = Util::rotateVector(QVector3D(0, 1, 0), bodyToNED);
    QPointF aftCenter = Util::lineIntersection(aftestPosition.toPointF(), starboard.toVector2D(), center, forward);
    qDebug() << "aft intersection is " << aftCenter;

    QVector3D bodyOrigin = QVector3D(aftCenter) + Util::rotateVector(QVector3D(aftToBodyOrigin, 0, 0), bodyToNED);
    bodyOrigin.setZ(positionEstimator->getCurrentCentroid().z());
    positionEstimator->setPositionAndAttitude(bodyOrigin, bodyToNED);


    trackingState = IN_TRACKING;

    cvReleaseMat(&NEDToBody);
    cvReleaseMat(&bodyToNED);

/*    QVector3D centerPoint = rotateVector(QVector3D(centerLinePt.x(), centerLinePt.y(), 0), initialRotMatrix) - initialCentroid;
    bodyOffset = QVector3D(0, centerLinePt.y(), 0); */



}

/**
  * Return a trackers id list for "camera" whose trackers have been present for "secondsOld" or more seconds.
  * Disregard trackers which have their id in "except"
  */
QVector <int> TrackingManager::findStableTrackers(Camera * camera, int secondsOld, QVector <int> except)
{

    QHash <int, QTime> startTimes = camera->getImageProcessing()->getTrackerStartTimes();
    QList <int> keys = startTimes.keys();

    QVector <int> readyTrackers;

    for(int i = 0; i < keys.size(); i++) {
        QTime trackerTime = startTimes.value(keys.at(i));
        trackerTime = trackerTime.addSecs(secondsOld);

        if (trackerTime <  QTime::currentTime()) {
            bool foundExcept = false;
            for(int j = 0; j < except.size(); j++) {
                if (except.at(j) == keys.at(i)) {
                    foundExcept = true;
                }
            }
            if (foundExcept) {
                continue;
            }
            readyTrackers.append(keys.at(i));
        }
    }
    return readyTrackers;
}

/**
  * Computes the epipolar line in camera2 given a point in camera1
  */
QLineF TrackingManager::computeEpipolarLine(Camera * camera1, QPointF point, Camera * camera2)
{
    int nPoints = 8;
    CvPoint3D32f * wPoints = new CvPoint3D32f[nPoints];
    CvMat * points1 = cvCreateMat(2, nPoints, CV_32F);
    CvMat * points2 = cvCreateMat(2, nPoints, CV_32F);


   /* wPoints [0] = cvPoint3D32f(0, 0, 0);
    wPoints [1] = cvPoint3D32f(0.1, 0, 0);
    wPoints [2] = cvPoint3D32f(0.1, 0.1, 0);
    wPoints [3] = cvPoint3D32f(0, 0.1, 0);
    wPoints [4] = cvPoint3D32f(0, 0, 0.1);
    wPoints [5] = cvPoint3D32f(0.1, 0, 0.1);
    wPoints [6] = cvPoint3D32f(0.1, 0.1, 0.1);
    wPoints [7] = cvPoint3D32f(0, 0.1, 0.1); */

    wPoints [0] = cvPoint3D32f(0, 0, 0);
    wPoints [1] = cvPoint3D32f(0.1, 0, 0);
    wPoints [2] = cvPoint3D32f(0.1, 0.1, 0);
    wPoints [3] = cvPoint3D32f(0, 0.1, 0);
    wPoints [4] = cvPoint3D32f(0, 0, 0.1);
    wPoints [5] = cvPoint3D32f(0.1, 0, 0.1);
    wPoints [6] = cvPoint3D32f(0.1, 0.1, 0.1);
    wPoints [7] = cvPoint3D32f(0, 0.1, 0.1);

    /* wPoints [0] = cvPoint3D32f(0, 0, 0);
    wPoints [1] = cvPoint3D32f(1, 0, 0);
    wPoints [2] = cvPoint3D32f(1, 1, 0);
    wPoints [3] = cvPoint3D32f(0, 1, 0);
    wPoints [4] = cvPoint3D32f(0, 0, -1);
    wPoints [5] = cvPoint3D32f(1, 0, -1);
    wPoints [6] = cvPoint3D32f(1, 1, -1);
    wPoints [7] = cvPoint3D32f(0, 1, -1); */

    for(int i = 0; i < nPoints; i++) {
        CvPoint point1 = camera1->mapToImage(wPoints[i]);
        cvSetReal2D(points1, 0, i, point1.x);
        cvSetReal2D(points1, 1, i, point1.y);

        CvPoint point2 = camera2->mapToImage(wPoints[i]);
        cvSetReal2D(points2, 0, i, point2.x);
        cvSetReal2D(points2, 1, i, point2.y);
    }

    //printMatrix(points1);
    //printMatrix(points2);

    CvMat * F = cvCreateMat(3, 3, CV_32F);
    cvFindFundamentalMat(points1, points2, F);

    //printMatrix(F);
    cvReleaseMat(&points1);
    cvReleaseMat(&points2);

    delete[] wPoints;

    CvMat * points = cvCreateMat(2, 3, CV_32F);
    cvZero(points);
    cvSetReal2D(points, 0, 0, point.x());
    cvSetReal2D(points, 1, 0, point.y());
    CvMat * lines = cvCreateMat(3, 3, CV_32F);
    cvZero(lines);
    cvComputeCorrespondEpilines(points, 1, F, lines);

    double a = cvGetReal2D(lines, 0, 0);
    double b = cvGetReal2D(lines, 1, 0);
    double c = cvGetReal2D(lines, 2, 0);

    QPointF p(0, -c/b);
    QVector2D dir(b, -a);


    QLineF line = Util::clipLineAgainstRect(p.toPoint(), dir, QRect(0, 0, camera2->getFrame()->width, camera2->getFrame()->height));

    // printMatrix(lines);

    cvReleaseMat(&points);
    cvReleaseMat(&lines);
    cvReleaseMat(&F);
    return line;
}
