#include "util.h"
#include <QDebug>

namespace Util {

    /**
      * Sets "rot" to be a rotation matrix with angle "angle" around the X-axis
      */
    void setXRotation(float angle, CvMat * rot)
    {
        float costheta = cos(angle);
        float sintheta = sin(angle);

        cvZero(rot);
        cvSetReal2D(rot, 0, 0, 1);
        cvSetReal2D(rot, 1, 1, costheta);
        cvSetReal2D(rot, 2, 2, costheta);
        cvSetReal2D(rot, 2, 1, sintheta);
        cvSetReal2D(rot, 1, 2, -sintheta);
    }


    /**
      * Sets "rot" to be a rotation matrix with angle "angle" around the X-axis
      */
    void setYRotation(float angle, CvMat * rot)
    {
        float costheta = cos(angle);
        float sintheta = sin(angle);

        cvZero(rot);
        cvSetReal2D(rot, 0, 0, costheta);
        cvSetReal2D(rot, 2, 2, costheta);
        cvSetReal2D(rot, 1, 1, 1);
        cvSetReal2D(rot, 2, 0, -sintheta);
        cvSetReal2D(rot, 0, 2, sintheta);
    }

    /**
      * Sets "rot" to be a rotation matrix with angle "angle" around the Z-axis
      */
    void setZRotation(float angle, CvMat * rot)
    {
        float costheta = cos(angle);
        float sintheta = sin(angle);
        cvZero(rot);
        cvSetReal2D(rot, 0, 0, costheta);
        cvSetReal2D(rot, 1, 1, costheta);
        cvSetReal2D(rot, 2, 2, 1);

        cvSetReal2D(rot, 1, 0, sintheta);
        cvSetReal2D(rot, 0, 1, -sintheta);
    }

    QPointF lineIntersection(QPointF p1, QVector2D v1, QPointF p2, QVector2D v2)
    {

        float num = v1.x() * ( p2.y() - p1.y() ) - v1.y() * ( p2.x() - p1.x() );
        float den = v1.y() * v2.x() - v1.x() * v2.y();
        if (den != 0) {
            return p2 + num/den * v2.toPointF();
        }

        return QPointF(0, 0);
    }


    /**
      * Computes the x and y (z = 0) position in the WCS given u and v in the picture
      */
    QPointF mapToXY(QPointF pixel, CvMat * C)
    {
        double u = pixel.x();
        double v = pixel.y();

        double x_w = u * (cvGetReal2D(C, 2, 3) * cvGetReal2D(C, 1, 1) - cvGetReal2D(C, 1, 3) * cvGetReal2D(C, 2, 1)) +
                     v * (cvGetReal2D(C, 0, 3) * cvGetReal2D(C, 2, 1) - cvGetReal2D(C, 2, 3) * cvGetReal2D(C, 0, 1)) +
                          cvGetReal2D(C, 1, 3) * cvGetReal2D(C, 0, 1) - cvGetReal2D(C, 0, 3) * cvGetReal2D(C, 1, 1);

        double y_w = u * (cvGetReal2D(C, 2, 0) * cvGetReal2D(C, 1, 3) - cvGetReal2D(C, 1, 0) * cvGetReal2D(C, 2, 3)) +
                     v * (cvGetReal2D(C, 0, 0) * cvGetReal2D(C, 2, 3) - cvGetReal2D(C, 2, 0) * cvGetReal2D(C, 0, 3)) +
                          cvGetReal2D(C, 1, 0) * cvGetReal2D(C, 0, 3) - cvGetReal2D(C, 0, 0) * cvGetReal2D(C, 1, 3);

        double det = u * (cvGetReal2D(C, 1, 0) * cvGetReal2D(C, 2, 1) - cvGetReal2D(C, 2, 0) * cvGetReal2D(C, 1, 1)) +
                     v * (cvGetReal2D(C, 2, 0) * cvGetReal2D(C, 0, 1) - cvGetReal2D(C, 0, 0) * cvGetReal2D(C, 2, 1)) +
                          cvGetReal2D(C, 0, 0) * cvGetReal2D(C, 1, 1) - cvGetReal2D(C, 1, 0) * cvGetReal2D(C, 0, 1);

        return QPointF(x_w / det, y_w / det);
    }

    QVector3D rotateVector(QVector3D qVec, CvMat * rot)
    {
        QVector3D res;
        res.setX( cvGetReal2D(rot, 0, 0) * qVec.x() + cvGetReal2D(rot, 0, 1) * qVec.y() + cvGetReal2D(rot, 0, 2) * qVec.z()) ;
        res.setY( cvGetReal2D(rot, 1, 0) * qVec.x() + cvGetReal2D(rot, 1, 1) * qVec.y() + cvGetReal2D(rot, 1, 2) * qVec.z());
        res.setZ( cvGetReal2D(rot, 2, 0) * qVec.x() + cvGetReal2D(rot, 2, 1) * qVec.y() + cvGetReal2D(rot, 2, 2) * qVec.z());
        return res;
    }

    QLine clipLineAgainstRect(QPoint p, QVector2D dir, QRect window)
    {
        return Util::clipLineAgainstRect(p, dir, window, false);
    }

    QLine clipLineAgainstRect(QPoint p, QVector2D dir, QRect window, bool debug)
    {
        debug = false;

        // Intersection left edge
        int leftY = -1;
        if (dir.x() != 0) {
            leftY = p.y() + (window.left() - p.x()) / dir.x() * dir.y();
            if  (leftY < window.top() || leftY > window.bottom()) {
                leftY = -1;
            }
        }

        // Intersection right edge
        int  rightY = -1;
        if (dir.x() != 0) {
            rightY = p.y() + (window.right() - p.x()) / dir.x() * dir.y();
            if  (rightY < window.top() || rightY > window.bottom()) {
                rightY = -1;
            }
        }

        // Intersection top edge
        int  topX = -1;
        if (dir.y() != 0) {
            topX = p.x() + (window.top() - p.y()) / dir.y() * dir.x();
            if (topX < window.left() || topX > window.right()) {
                topX = -1;
            }
        }

        // Intersection bottom edge
        int  bottomX = -1;
        if (dir.y() != 0) {
            bottomX = p.x() + (window.bottom() - p.y()) / dir.y() * dir.x();
            if (bottomX < window.left() || bottomX > window.right()) {
                bottomX = -1;
            }
        }

        // Find two intersections
        int found = 0;

        QPoint pt1, pt2;

        if (leftY != -1) {
            pt1 = QPoint(window.left(), leftY);
            found++;
        }

        if (rightY != -1) {
            if (found == 0) {
                pt1 = QPoint(window.right(), rightY);
            } else {
                pt2 = QPoint(window.right(), rightY);
            }
            found++;
        }

        if (topX != -1) {
            if (found == 0) {
                pt1 = QPoint(topX, window.top());
            } else {
                pt2 = QPoint(topX, window.top());
            }
            found++;
        }

        if (bottomX != -1) {
            if (found == 0) {
                pt1 = QPoint(bottomX, window.bottom());
            } else {
                pt2 = QPoint(bottomX, window.bottom());
            }
            found++;
        }

        QVector2D vec(pt2 - pt1);

        if (debug) {
            qDebug() << "direction is " << dir << "\t resulting vec is: " << vec;
        }

        if (QVector2D::dotProduct(vec.normalized(), dir.normalized()) > 0.8) {
            return QLine(pt1, pt2);
        }

        return QLine(pt2, pt1);
    }
}
