#include <QDebug>

#include "vesselsearch.h"
#include "imageprocessing.h"

#include <QWidget>

#define MAX_HOUGH_LINES 100

VesselSearch::VesselSearch(ImageProcessing * parent, QSize size, int id)
{
    this->parent = parent;
    frame = NULL;
    frameGray = NULL;
    frameEdges = NULL;
    frameEdgesRGB = NULL;

    inited = false;

    setVideoSize(size);

    lines = cvCreateMat(MAX_HOUGH_LINES, 1, CV_32FC2);
    cvZero(lines);
    lines->rows = 0;
    cameraMatrix = cvCreateMat(3, 4, CV_32FC1);
    cvZero(cameraMatrix);
    worldLines = new Line[MAX_HOUGH_LINES];
    cannyParam1 = 280;
    cannyParam2 = 345;
    houghThres = 100;

    houghWidth = 400;
    houghHeight = 400;
    houghWorldDisplay = cvCreateImage(cvSize(houghWidth, houghHeight), IPL_DEPTH_8U, 3);
    cvZero(houghWorldDisplay);

    this->id = id;
    QString num;
    num.setNum(id);
    worldDisplayText = QString("houghWorldDisplay") + num;
    houghLinesText = QString("houghLines") + num;

    //cvNamedWindow(worldDisplayText.toAscii().data());
    //cvNamedWindow(houghLinesText.toAscii().data());

    maxThetaDiff = 0.2;
    maxRhoDiff = 0.08;

    minTheta = 10.0f * M_PI / 180.0f;
    maxTheta = 180.0f * M_PI / 180.0f;

    minRho = -0.20;
    maxRho = 0.30;

    vesselWidth = 0.21;
    deviation = 0.03;
    step = 0.01;
    spacing = vesselWidth - 4 * deviation;
    searchStart = -0.30;
    searchEnd = 0.30;

    templLength = (deviation * 4 + spacing + deviation * 4) / step + 2;
    templ = new float[templLength];

    float mean1 = deviation * 2;
    float mean2 = deviation * 4 + spacing + deviation * 2;

    for(int i = 0; i < templLength; i++) {
        float templPos = step * i;
        float norm1 = 1 / ( sqrt(2 * M_PI) * deviation ) * exp(- pow((templPos - mean1), 2)/ (2 * pow(deviation, 2)));
        float norm2 = 1 / ( sqrt(2 * M_PI) * deviation ) * exp(- pow((templPos - mean2), 2)/ (2 * pow(deviation, 2)));
        float templValue = (norm1 + norm2) / 2.0f;
        templ[i] = templValue;
    }

    successiveMatches = 0;
    linesVisible = true;
}

CvMat * VesselSearch::getLines()
{
    return lines;
}

Line * VesselSearch::getWorldLines()
{
    return worldLines;
}

ImageProcessing * VesselSearch::getParent()
{
    return parent;
}

void VesselSearch::setVideoSize(QSize videoSize)
{

    if (videoSize == this->videoSize) {
        return;
    }

    if (inited) {
        cvReleaseImage(&frameGray);
        cvReleaseImage(&frameEdges);
        cvReleaseImage(&frameEdgesRGB );
    }

    frameGray = cvCreateImage(cvSize(videoSize.width(), videoSize.height()), IPL_DEPTH_8U, 1);
    frameEdges = cvCreateImage(cvSize(videoSize.width(), videoSize.height()), IPL_DEPTH_8U, 1);
    frameEdgesRGB = cvCreateImage(cvSize(videoSize.width(), videoSize.height()), IPL_DEPTH_8U, 3);

    this->videoSize = videoSize;
    inited = true;
}

VesselSearch::~VesselSearch()
{
    if (inited) {
        cvReleaseImage(&frameGray);
        cvReleaseImage(&frameEdges);
        cvReleaseImage(&frameEdgesRGB);
    }
    cvReleaseImage(&houghWorldDisplay);
    lines->rows = MAX_HOUGH_LINES;
    cvReleaseMat(&lines);
    delete[] templ;
    delete[] worldLines;
}

QPolygon VesselSearch::getVesselPolygon()
{
    dataMutex.lock();
    QPolygon copyOfVesselPolygon = vesselPolygon;
    dataMutex.unlock();
    return copyOfVesselPolygon;
}

/**
  * Returns a list of alternating port,starboard polygons which are located on different parts of the vessel in WCS XY coords
  */
QVector <QPolygonF> VesselSearch::getVesselPolygons()
{
    dataMutex.lock();
    QVector <QPolygonF> copyOfVesselPolygons = vesselPolygons;
    dataMutex.unlock();
    return copyOfVesselPolygons;
}

/**
  * Returns a list of non-overlapping polygons which are located on different parts of the vessel in image coords
  */
QVector <QPolygon> VesselSearch::getVesselPolygonsImage()
{
    dataMutex.lock();
    QVector <QPolygon> polygonsImage;

    for(int i = 0; i < vesselPolygons.size(); i++) {
        QPolygonF poly = vesselPolygons.at(i);
        QPolygon polyImage;
        for(int j = 0; j < poly.size(); j++) {
            polyImage.append(mapToImage(poly.at(j)));
        }
        polygonsImage.append(polyImage);
    }
    dataMutex.unlock();

    return polygonsImage;
}

QPointF VesselSearch::getCenterLinePt()
{
    return centerLinePt;
}

QVector2D VesselSearch::getForwardVec()
{
    return forwardVec;
}

/**
  * A requirement for this function is that the origin of WCS is in the cameras view
  */
void VesselSearch::createVesselPolygon(LineGroup lineGroup)
{

    float avgTheta = lineGroup.avgTheta;
    float portRho = lineGroup.matchRho + 2 * deviation;
    float starboardRho = portRho + vesselWidth;

    QLine fromPortTowardsStarboard = mapToImage(QLineF(QPointF(0, 0), QPointF(cos(avgTheta), sin(avgTheta))));
    QVector2D fromPortTowardsStarboardVec(fromPortTowardsStarboard.p2() - fromPortTowardsStarboard.p1());

    QVector2D forward(sin(avgTheta), -cos(avgTheta));
    QLine forwardInImage = mapToImage(QLineF(QPointF(0, 0), forward.toPointF()));
    QVector2D forwardImageVec(forwardInImage.p2() - forwardInImage.p1());

    // Ensure that forward direction is away from camera (has negative y-value in the image plane)
    if (forwardImageVec.y() > 0) {
        forward *= -1;
    }

    // Setting the forward vec to be fetchable later
    forwardVec = forward;

    /*QLine starboardImage = mapToImage(QLineF(QPointF(0, 0), QPointF(cos(avgTheta), sin(avgTheta))));
    QVector2D starboardImageVec(starboardImage.p2() - starboardImage.p1());*/

    QVector2D toStarboard = QVector2D(cos(avgTheta), sin(avgTheta));

    // Ensuring that port is always left of starboard in frame
    if (fromPortTowardsStarboardVec.x() <= 0) {
        toStarboard *= -1;
        float tmp = portRho;
        portRho = -starboardRho;
        starboardRho = -tmp;
        avgTheta += M_PI;
    }

    float centerRho = (starboardRho + portRho) / 2;

    QPointF port(portRho * cos(avgTheta), portRho * sin(avgTheta));
    QPointF center(centerRho * cos(avgTheta), centerRho * sin(avgTheta));
    centerLinePt = center;
    QPointF starboard(starboardRho * cos(avgTheta), starboardRho * sin(avgTheta));

    QPoint portP = mapToImage(port);
    QPoint starboardP = mapToImage(starboard);

    QLine linePort = mapToImageAndClip(port, forward, frameWindow);    
    if (linesVisible)
        cvLine(frameEdgesRGB, cvPoint(linePort.p1().x(), linePort.p1().y()),  cvPoint(linePort.p2().x(), linePort.p2().y()), cvScalar(0, 0, 255), 10);

    QLine lineStarboard = mapToImageAndClip(starboard, forward, frameWindow);
    if (linesVisible)
        cvLine(frameEdgesRGB, cvPoint(lineStarboard.p1().x(), lineStarboard .p1().y()),  cvPoint(lineStarboard.p2().x(), lineStarboard.p2().y()), cvScalar(0, 255, 0), 10);

    QLine lineCenter = mapToImageAndClip(center, forward, frameWindow);
    if (linesVisible)
        cvLine(frameEdgesRGB, cvPoint(lineCenter.p1().x(), lineCenter .p1().y()),  cvPoint(lineCenter.p2().x(), lineCenter.p2().y()), cvScalar(255, 255, 255), 15);

    if (linesVisible) {
        cvCircle(frameEdgesRGB, cvPoint(linePort.p1().x(), linePort.p1().y()), 30, cvScalar(0, 0, 255), 5);
        cvCircle(frameEdgesRGB, cvPoint(lineStarboard.p1().x(), lineStarboard.p1().y()), 30, cvScalar(0, 255, 0), 5);
    }

    // Create main vessel polygon
    vesselPolygon.clear();
    vesselPolygon.append(QPoint(linePort.p1()));
    vesselPolygon.append(QPoint(linePort.p2()));
    vesselPolygon.append(QPoint(lineStarboard.p2()));
    vesselPolygon.append(QPoint(lineStarboard.p1()));

    float polyWidth = 0.06;
    float polyHeight = 0.035;
    float forwardSpacing = 0.00;
    float edgeSpacing = 0.03;

    QLine portEdge = mapToImageAndClip(port + toStarboard.toPointF() * edgeSpacing, forward, frameWindow, false);
    QPointF portCorner = mapToXY(portEdge.p1());

    // Increase window size because of numeric inaccuracy in mapToImage
    QRect frameWindowPad = QRect(-5, -5, this->frameWindow.width() + 10, this->frameWindow.height() + 10);

    QVector <QPolygonF> polygons;
    int i;
    for(i = 0; i < 20; i++) {

        QPointF portFrontCorner = portCorner + forward.toPointF() * polyHeight;
        QPointF starboardCorner = portCorner + toStarboard.toPointF() * polyWidth;
        QPointF starboardFrontCorner = starboardCorner + forward.toPointF() * polyHeight;

        QPoint l = mapToImage(portFrontCorner);
        QPoint r = mapToImage(starboardFrontCorner);

        /*qDebug() << "portCorner " << mapToImage(portCorner);
        qDebug() << "starboardCorner " << mapToImage(starboardCorner);
        qDebug() << "portFrontCorner " << mapToImage(portFrontCorner);
        qDebug() << "starboardFrontCorner " << mapToImage(starboardFrontCorner);*/

        if (!frameWindowPad.contains(l) && !frameWindowPad.contains(r)) {
            break;
        }

        QPolygonF polyPort;
        polyPort.append(portCorner);
        polyPort.append(starboardCorner);
        polyPort.append(starboardFrontCorner);
        polyPort.append(portFrontCorner);

        polygons.append(polyPort);
        polygons.append(polyPort.translated(toStarboard.toPointF() * (vesselWidth - polyWidth - 2 * edgeSpacing)));

        portCorner += forward.toPointF() * (polyHeight + forwardSpacing);
    }

    //qDebug() << "added " << i << " polygons";
    vesselPolygons = polygons;
    emit vesselEdgesFound(linePort, lineStarboard);

}

QPointF VesselSearch::lineIntersection(QPointF p1, QVector2D v1, QPointF p2, QVector2D v2)
{

    float num = v1.x() * ( p2.y() - p1.y() ) - v1.y() * ( p2.x() - p1.x() );
    float den = v1.y() * v2.x() - v1.x() * v2.y();
    if (den != 0) {
        return p2 + num/den * v2.toPointF();
    }

    return QPointF(0, 0);
}

void VesselSearch::setFrame(IplImage * frame)
{
    dataMutex.lock();
    this->frame = frame;
    frameSize.setWidth(frame->width);
    frameSize.setHeight(frame->height);
    frameWindow.setRect(0, 0, frame->width + 1, frame->height + 1);
    houghLines();

    transformHoughLines();
    groupLinesByAngle();

    int matchGroup;
    bool success = false;
    float match = findBestLineGroupMatch(lineGroups, matchGroup, success);

    if (success) {
        //qDebug() << "Best match is group no: " << matchGroup << lineGroups.at(matchGroup).matchRho << " avgTheta: " << lineGroups.at(matchGroup).avgTheta * 180 / M_PI;

        if (checkMatchConsistency(lineGroups.at(matchGroup))) {
            successiveMatches++;

        } else {
            successiveMatches = 0;
        }


        for(int i = 0; i < lineGroups.at(matchGroup).lines.size(); i++) {

            float worldTheta = lineGroups.at(matchGroup).lines.at(i)->theta;
            float worldRho = lineGroups.at(matchGroup).lines.at(i)->rho;

            QVector2D middle(worldRho * cos(worldTheta), worldRho * sin(worldTheta));
            QVector2D vec2(worldRho * sin(worldTheta), -worldRho * cos(worldTheta));

            cvLine(houghWorldDisplay, convertToHough(middle - vec2 * 100), convertToHough(middle + vec2 * 100), cvScalar(80, 80, 255), 1);
        }

        float worldTheta = lineGroups.at(matchGroup).avgTheta;
        float worldRho = match;



        QVector2D middle = QVector2D(worldRho * cos(worldTheta), worldRho * sin(worldTheta));
        QVector2D vec2 = QVector2D(worldRho * sin(worldTheta), -worldRho * cos(worldTheta)).normalized();

        CvScalar matchColor;

        if (successiveMatches > 5) {
            matchColor = cvScalar(0, 255, 0, 0);
            // qDebug() << "match: theta: " << worldTheta * 180 / M_PI << " rho: " << worldRho;
            createVesselPolygon(lineGroups.at(matchGroup));

        } else {
            matchColor = cvScalar(200, 200, 200, 0);
        }

        cvLine(houghWorldDisplay, convertToHough(middle - vec2), convertToHough(middle + vec2), matchColor, 2);

        worldRho += deviation * 2 + spacing;

        middle = QVector2D(worldRho * cos(worldTheta), worldRho * sin(worldTheta));
        vec2 = QVector2D(worldRho * sin(worldTheta), -worldRho * cos(worldTheta)).normalized();

        cvLine(houghWorldDisplay, convertToHough(middle - vec2), convertToHough(middle + vec2), matchColor, 2);
        if (linesVisible)
            drawMatchLines(frameEdgesRGB, lineGroups.at(matchGroup).matchRho, worldTheta, matchColor);


    } else {
        if (successiveMatches > 0) {
            //successiveMatches--;
        }
        //qDebug() << "no group matches";
    }
    //qDebug() << "successiveMatches = " << successiveMatches;

   // cvShowImage(worldDisplayText.toAscii().data(), houghWorldDisplay);
    //cvShowImage(houghLinesText.toAscii().data(), frameEdgesRGB);
    //cvWaitKey(1);
    dataMutex.unlock();
}


void VesselSearch::drawMatchLines(IplImage * frame, float rho, float theta, CvScalar color)
{

    float side1Rho = rho + deviation * 2;
    float side2Rho = side1Rho  + vesselWidth;
    QPointF middle1 = QPointF(side1Rho * cos(theta), side1Rho * sin(theta));
    QPointF  middle2 = QPointF (side2Rho * cos(theta), side2Rho * sin(theta));
    QVector2D direction = QVector2D(sin(theta), -cos(theta)).normalized();

    QLine line = mapToImageAndClip(middle1, direction, QRect(0, 0, frame->width, frame->height));
    cvLine(frame, cvPoint(line.p1().x(), line.p1().y()),  cvPoint(line.p2().x(), line.p2().y()), color, 10);

    line = mapToImageAndClip(middle2, direction, QRect(0, 0, frame->width, frame->height));
    cvLine(frame, cvPoint(line.p1().x(), line.p1().y()),  cvPoint(line.p2().x(), line.p2().y()), color, 10);
}

IplImage * VesselSearch::getLineImage()
{
    return frameEdgesRGB;
}

/**
  * Checks if match is stable
  */
bool VesselSearch::checkMatchConsistency(LineGroup lineGroup)
{

    if (successiveMatches == 0) {

        prevMatchGroup = lineGroup;
        return true;

    } else {

        //qDebug() << "matchGroup theta: " << prevMatchGroup.avgTheta << " lineGroup theta: " << lineGroup.avgTheta;
        //qDebug() << "matchGroup rho: " << prevMatchGroup.matchRho << " lineGroup rho: " << lineGroup.matchRho;

        float thetaDiff = lineGroup.avgTheta - prevMatchGroup.avgTheta;
        float rhoDiff = lineGroup.matchRho - prevMatchGroup.matchRho;

        if (thetaDiff < 0) {
            thetaDiff = -thetaDiff;
        }

        if (rhoDiff < 0) {
            rhoDiff = -rhoDiff;
        }



        if ((thetaDiff < maxThetaDiff || thetaDiff > M_PI - maxThetaDiff) && rhoDiff < maxRhoDiff) {
            prevMatchGroup = lineGroup;
            return true;
        } else {
            successiveMatches = 0;
            prevMatchGroup = lineGroup;
            return false;
        }
    }


    return false;
}


void VesselSearch::setLinesVisible(bool visible)
{
    this->linesVisible = visible;
}

/**
  * Performs Hough transform on the edges of the frame
  */
void VesselSearch::houghLines()
{
    lines->rows = MAX_HOUGH_LINES - 1;
    cvCvtColor(frame, frameGray, CV_RGB2GRAY);
    cvCanny(frameGray, frameEdges, cannyParam1, cannyParam2, 3);
    cvHoughLines2(frameEdges, lines, CV_HOUGH_STANDARD, 4, CV_PI/50, houghThres);


    cvCvtColor(frameEdges, frameEdgesRGB, CV_GRAY2BGR);


    for(int i = 0; i < lines->rows; i++) {
        float rho = lines->data.fl[i * 2];
        float theta = lines->data.fl[i * 2 + 1];
        QLineF line = clipLineAgainstRect(QPointF(rho * cos(theta), rho * sin(theta)), QVector2D(cos(theta + M_PI/2), sin(theta + M_PI/2)), QRectF(0, 0, frame->width, frame->height));
        if (linesVisible)
            cvLine(frameEdgesRGB, cvPoint(line.p1().x(), line.p1().y()), cvPoint(line.p2().x(), line.p2().y()), cvScalar(100, 50, 50), 1);
    }
}

/**
  * Transform Hough lines in the image to the XY WCS (assuming z = 0)
  */
void VesselSearch::transformHoughLines(void)
{
    int i;

    cvZero(houghWorldDisplay);

    for(i = 0; i < lines->rows; i++) {

        double rho = lines->data.fl[i * 2];
        double theta = lines->data.fl[i * 2 + 1];

        QPointF p1Cam(cos(theta) * rho, sin(theta) * rho);
        QPointF p2Cam(p1Cam + QPointF(sin(theta), -cos(theta)) * 100);

        QPointF p1(mapToXY(p1Cam));
        QPointF p2(mapToXY(p2Cam));

        QPointF p3(0, 0);
        QPointF p2p1Diff(p2.x() - p1.x(), p2.y() - p1.y());
        double u = ( ( p3.x() - p1.x() ) * ( p2.x() - p1.x() ) + ( p3.y() - p1.y() ) * ( p2.y() - p1.y() ) ) / (pow(p2p1Diff.x(), 2) + pow(p2p1Diff.y(), 2));
        QPointF p4 = QPointF(p1.x() + u * (p2.x() - p1.x()), p1.y() + u * (p2.y() - p1.y()));
        QPointF vec(p2 - p1);


        double worldRho =  sqrt(pow(p4.x(), 2) + pow(p4.y(), 2));
        double worldTheta = atan2(p4.y(), p4.x());

        if (worldTheta > M_PI) {
            worldTheta -= M_PI;
            worldRho = -worldRho ;
        } else if(worldTheta < 0) {
            worldTheta += M_PI;
            worldRho = -worldRho;
        }

        worldLines[i].rho = worldRho;
        worldLines[i].theta = worldTheta;
        worldLines[i].checked = false;
        worldLines[i].srcLineNum = i;

        QPointF middle = QPointF(worldRho * cos(worldTheta), worldRho * sin(worldTheta));
        QVector2D vec2 = QVector2D(worldRho * sin(worldTheta), -worldRho * cos(worldTheta)).normalized();

        QLineF line = clipLineAgainstRect(middle, vec2, QRectF(-0.4, -0.4, 0.7, 0.7));
        cvLine(houghWorldDisplay, convertToHough(QVector2D(line.p1())), convertToHough(QVector2D(line.p2())), cvScalar(100, 50, 50), 1);
    }

    cvLine(houghWorldDisplay, convertToHough(QVector2D(0, 0)), convertToHough(QVector2D(0.18, 0.0)), cvScalar(255, 0, 0), 1);
    cvLine(houghWorldDisplay, convertToHough(QVector2D(0.18, 0)), convertToHough(QVector2D(0.18, 0.18)), cvScalar(255, 0, 0), 1);
    cvLine(houghWorldDisplay, convertToHough(QVector2D(0.18, 0.18)), convertToHough(QVector2D(0, 0.18)), cvScalar(255, 0, 0), 1);
    cvLine(houghWorldDisplay, convertToHough(QVector2D(0, 0.18)), convertToHough(QVector2D(0, 0)), cvScalar(255, 0, 0), 1);

}


/**
  * Ensure that all angles are between 0 and 180 deg
  */
void VesselSearch::normalizeGroupByAngle(LineGroup * lineGroup)
{
    float cmpAngle = lineGroup->lines.at(0)->theta;

    for(int i = 1; i < lineGroup->lines.size(); i++) {
        float diff = lineGroup->lines.at(i)->theta - cmpAngle;

        if (diff > M_PI - maxThetaDiff) {
            lineGroup->lines.at(i)->theta -= M_PI;
            lineGroup->lines.at(i)->rho = -lineGroup->lines.at(i)->rho;
        } else if (diff < -M_PI + maxThetaDiff) {
            lineGroup->lines.at(i)->theta += M_PI;
            lineGroup->lines.at(i)->rho = -lineGroup->lines.at(i)->rho;
        }
    }
}

/**
  * Groups lines based on similar angle
  */
void VesselSearch::groupLinesByAngle(void)
{
    LineGroup lineGroup;
    lineGroups.clear();

    for(int i = 0; i < lines->rows; i++) {

        if (worldLines[i].checked) {
            continue;
        }


        int numAdded = 1;
        float sumOrigTheta = lines->data.fl[worldLines[i].srcLineNum * 2 + 1];

        lineGroup.lines.clear();
        worldLines[i].checked = true;
        lineGroup.lines.append(&(worldLines[i]));


        // qDebug() << "checking " << i << " theta = " << worldLines[i].theta * 180 / M_PI;

        for(int j = 0; j < lines->rows; j++) {

            if (i == j || worldLines[j].checked) {
                continue;
            }

            double diff = worldLines[i].theta - worldLines[j].theta;
            if (diff < 0) {
                diff = -diff;
            }


            if (diff < maxThetaDiff || diff > M_PI - maxThetaDiff) {
                //qDebug() << i << "(theta: " << worldLines[i].theta * 180 / M_PI << " " << worldLines[i].theta << " rad) is parallel to " << j << " (theta: " << worldLines[j].theta * 180 / M_PI << " " << worldLines[j].theta << " rad" << " diff = " << diff * 180 / M_PI;
                worldLines[j].checked = true;
                lineGroup.lines.append(&(worldLines[j]));
                sumOrigTheta += lines->data.fl[worldLines[j].srcLineNum * 2 + 1];
                numAdded++;
            } else {
                //qDebug() << "\t" << i << "(theta: " << worldLines[i].theta * 180 / M_PI << " " << worldLines[i].theta << " rad) is not parallel to " << j << " (theta: " << worldLines[j].theta * 180 / M_PI << " " << worldLines[j].theta << " rad" << " diff = " << diff * 180 / M_PI;
            }
        }


        normalizeGroupByAngle(&lineGroup);

        float sumTheta = 0;
        for(int i = 0; i < lineGroup.lines.size(); i++) {
            sumTheta += lineGroup.lines.at(i)->theta;
        }

        float avgTheta = sumTheta / lineGroup.lines.size();

        lineGroup.avgOrigTheta = sumOrigTheta / (float) numAdded;
        lineGroup.avgTheta = avgTheta;
        lineGroups.append(lineGroup);

    }

    /* qDebug() << "Result:";
    for (int i = 0; i < lineGroups.size(); i++) {
        qDebug() << "LineGroup " << i;
        for (int j = 0; j < lineGroups.at(i).lines.size(); j++) {
            qDebug() << "\t " << j << " rho: " << lineGroups.at(i).lines.at(j)->rho << " theta: " << lineGroups.at(i).lines.at(j)->theta * 180 / M_PI;
        }
    }*/
}

float VesselSearch::findBestLineGroupMatch(QVector <LineGroup> lineGroups, int &matchGroup, bool &success)
{
    float maxMatch = 0;
    int maxMatchGroup = 0;

    int len = lineGroups.size();
    float * matchPos = new float[len];

    success = false;

    for (int i = 0; i < len; i++) {
        float match = computeLineGroupMatch((LineGroup *)&(lineGroups.at(i)), matchPos[i]);
        
        if (match > 0.0f) {
            if (match > maxMatch) {
                maxMatch = match;
                maxMatchGroup = i;
                success = true;
            }
        }
        //qDebug() << "checking group " << i << " first angle is " << lineGroups.at(i).lines.at(0)->theta * 180 / M_PI << " avg angle was " << lineGroups.at(i).avgTheta * 180 / M_PI << " match was " << match;
    }






    matchGroup = maxMatchGroup;
    float maxMatchPos = matchPos[maxMatchGroup];
    delete[] matchPos;

    return maxMatchPos;
}

/**
  * Computes the match between a line group and a predefined template
  */
float VesselSearch::computeLineGroupMatch(LineGroup * lineGroup, float &matchPos)
{

    // Check if the theta for the group is in the interested interval
    bool disqualified = false;

    if (lineGroup->avgTheta < minTheta || lineGroup->avgTheta > maxTheta) {
        disqualified = true;
    }

    /*if (lineGroup->avgTheta > 15.0 * M_PI / 180.0f && lineGroup->avgTheta < 165.0 * M_PI / 180.0f ) {
        disqualified = true;
    }*/

    int searchLength = (searchEnd - searchStart) / step;
    bool * search = new bool[searchLength];

    for(int i = 0; i < searchLength; i++) {
        search[i] = false;
    }

    for(int i = 0; i < lineGroup->lines.size(); i++) {
        Line * line = lineGroup->lines.at(i);
        float rho = line->rho;
        if (rho >= searchStart && rho <= searchEnd) {
            int index = (rho - searchStart) / step;
            search[index] = true;
        }
    }

    /*for(int i = 0; i < searchLength; i++) {
        qDebug() << (int) search[i] << ",";
    }*/

    /* for(int i = 0; i < templLength; i++) {
        qDebug() << templ[i] << ",";
    }*/

    float * result = new float[searchLength - templLength];

    float resultMax = 0;
    int resultMaxPos = -1;

    for(int i = 0; i < searchLength - templLength; i++) {
        float acc = 0;

        bool leftEdgePresent = false;
        bool rightEdgePresent = false;

        for(int j = 0; j < templLength; j++) {
            acc = templ[j] * search[i + j];

            if ( (j < deviation * 4 / step) && search[i + j] ) {
                leftEdgePresent = true;
            }
            if ( (j > (deviation * 4 + spacing) / step) && search[i + j]) {
                rightEdgePresent = true;
            }
        }

        if (leftEdgePresent == false || rightEdgePresent == false )
            acc = 0;


        if (acc > resultMax) {
            resultMax = acc;
            resultMaxPos = i;
        }
        //qDebug() << acc << ",";
        result[i] = acc;
    }


    delete[] search;
    delete[] result;

    matchPos = searchStart + (resultMaxPos * step) + deviation * 2;
    lineGroup->matchRho = matchPos;
    if (lineGroup->matchRho < minRho || lineGroup->matchRho > maxRho) {
        return 0;
    }

    if (disqualified) {
        return 0;
    }

    return resultMax;
}

/**
  * Draws the Hough WCS lines in a plot
  */
CvPoint VesselSearch::convertToHough(QVector2D point)
{
    return cvPoint(point.x() * houghWidth  / 0.18 / 4  + 100 * 2, houghHeight - (point.y() * houghHeight  / 0.18 / 4 + 100 * 2));
}

/**
  * Computes the x and y (z = 0) position in the WCS given u and v in the picture
  */
QPointF VesselSearch::mapToXY(QPointF pixel)
{
    double u = pixel.x();
    double v = pixel.y();
    CvMat * C = cameraMatrix;

    double x_w = u * (cvGetReal2D(C, 2, 3) * cvGetReal2D(C, 1, 1) - cvGetReal2D(C, 1, 3) * cvGetReal2D(C, 2, 1)) +
                 v * (cvGetReal2D(C, 0, 3) * cvGetReal2D(C, 2, 1) - cvGetReal2D(C, 2, 3) * cvGetReal2D(C, 0, 1)) +
                      cvGetReal2D(C, 1, 3) * cvGetReal2D(C, 0, 1) - cvGetReal2D(C, 0, 3) * cvGetReal2D(C, 1, 1);

    double y_w = u * (cvGetReal2D(C, 2, 0) * cvGetReal2D(C, 1, 3) - cvGetReal2D(C, 1, 0) * cvGetReal2D(C, 2, 3)) +
                 v * (cvGetReal2D(C, 0, 0) * cvGetReal2D(C, 2, 3) - cvGetReal2D(C, 2, 0) * cvGetReal2D(C, 0, 3)) +
                      cvGetReal2D(C, 1, 0) * cvGetReal2D(C, 0, 3) - cvGetReal2D(C, 0, 0) * cvGetReal2D(C, 1, 3);

    double det = u * (cvGetReal2D(C, 1, 0) * cvGetReal2D(C, 2, 1) - cvGetReal2D(C, 2, 0) * cvGetReal2D(C, 1, 1)) +
                 v * (cvGetReal2D(C, 2, 0) * cvGetReal2D(C, 0, 1) - cvGetReal2D(C, 0, 0) * cvGetReal2D(C, 2, 1)) +
                      cvGetReal2D(C, 0, 0) * cvGetReal2D(C, 1, 1) - cvGetReal2D(C, 1, 0) * cvGetReal2D(C, 0, 1);

    return QPointF(x_w / det, y_w / det);
}



void VesselSearch::setCannyParam1(int val)
{
    qDebug() << "param1 " << val;
    cannyParam1 = val;
}

void VesselSearch::setCannyParam2(int val)
{
    qDebug() << "param2 " << val;
    cannyParam2 = val;
}

void VesselSearch::setHoughThres(int val)
{
    qDebug() << "hough " << val;
    houghThres = val;
}

void VesselSearch::setCameraMatrix(CvMat * matrix)
{
    cvCopy(matrix, cameraMatrix);
    XYclip.clear();
    XYclip.append(mapToXY(QPointF(0, 0)));
    XYclip.append(mapToXY(QPointF(frameSize.width(), 0)));
    XYclip.append(mapToXY(QPointF(frameSize.width(), frameSize.height())));
    XYclip.append(mapToXY(QPointF(0, frameSize.height())));
}

void VesselSearch::setDebugLine(double rho, double theta)
{
    this->debugRho = rho;
    this->debugTheta = theta;

}

/**
  * Maps an unbounded line to image and clips it and returns a bounded line
  */
QLine VesselSearch::mapToImageAndClip(QPointF p, QVector2D dir, QRect window, bool debug)
{
    QPoint p1Img = mapToImage(p);
    QPoint p2Img = mapToImage(p + dir.toPointF());

    QVector2D dirImg(p2Img - p1Img);
    QLineF line = clipLineAgainstRect(p1Img, dirImg, window);
    return line.toLine();
}

QLine VesselSearch::mapToImageAndClip(QPointF p, QVector2D dir, QRect window)
{
    return mapToImageAndClip(p, dir, window, false);

}


/**
  * Finds the pixel coordinates of a point in the XY plane
  */
QPoint VesselSearch::mapToImage(QPointF qPoint)
{

    CvPoint3D32f point = cvPoint3D32f(qPoint.x(), qPoint.y(), 0);

    CvMat * world = cvCreateMat(4, 1, CV_32F);

    world->data.fl[0] = point.x;
    world->data.fl[1] = point.y;
    world->data.fl[2] = point.z;
    world->data.fl[3] = 1;

    CvMat * image = cvCreateMat(3, 1, CV_32F);
    cvGEMM(cameraMatrix, world, 1.0f, NULL, 1.0f, image, 0);

    float u = (float) image->data.fl[0] / fabs((float) image->data.fl[2]);
    float v = (float) image->data.fl[1] / fabs((float) image->data.fl[2]);

    CvPoint p = cvPoint(u, v);

    cvReleaseMat(&world);
    cvReleaseMat(&image);

    return QPoint(p.x, p.y);
}

/**
  * Maps a line with endpoints to image
  */
QLine VesselSearch::mapToImage(QLineF line)
{
    return QLine(mapToImage(line.p1()), mapToImage(line.p2()));
}


void VesselSearch::reset()
{
    successiveMatches = 0;
}

//Compute the bit code for a point (x, y) using the clip rectangle
//bounded diagonally by (xmin, ymin), and (xmax, ymax)
int VesselSearch::computeOutCode (double x, double y)
{

    int ymin = 0;
    int xmin = 0;
    int xmax = frame->width;
    int ymax = frame->height;
    int code = 0;
    if (y > ymax)              //above the clip window
        code |= TOP;
    else if (y < ymin)         //below the clip window
        code |= BOTTOM;
    if (x > xmax)              //to the right of clip window
        code |= RIGHT;
    else if (x < xmin)         //to the left of clip window
        code |= LEFT;
    return code;
}

//Cohen-Sutherland clipping algorithm clips a line from
//P0 = (x0, y0) to P1 = (x1, y1) against a rectangle with
//diagonal from (xmin, ymin) to (xmax, ymax).
QLine VesselSearch::cohenSutherlandLineClip(QLine line)
{

    double x0 = line.p1().x();
    double y0 = line.p1().y();
    double x1 = line.p2().x();
    double y1 = line.p2().y();

    int ymin = 0;
    int xmin = 0;
    int xmax = frame->width;
    int ymax = frame->height;

    //Outcodes for P0, P1, and whatever point lies outside the clip rectangle
    int int0, int1, intOut;
    bool accept = false, done = false;

    //compute ints
    int0 = computeOutCode (x0, y0);
    int1 = computeOutCode (x1, y1);

    do{
        if (!(int0 | int1))      //logical or is 0. Trivially accept and get out of loop
        {
            accept = true;
            done = true;
        }
        else if (int0 & int1)//logical and is not 0. Trivially reject and get out of loop
        {
            done = true;
        }

        else
        {
            //failed both tests, so calculate the line segment to clip
            //from an outside point to an intersection with clip edge
            double x, y;
            //At least one endpoint is outside the clip rectangle; pick it.
            intOut = int0? int0: int1;
            //Now find the intersection point;
            //use formulas y = y0 + slope * (x - x0), x = x0 + (1/slope)* (y - y0)
            if (intOut & TOP)          //point is above the clip rectangle
            {
                x = x0 + (x1 - x0) * (ymax - y0)/(y1 - y0);
                y = ymax;
            }
            else if (intOut & BOTTOM)  //point is below the clip rectangle
            {
                x = x0 + (x1 - x0) * (ymin - y0)/(y1 - y0);
                y = ymin;
            }
            else if (intOut & RIGHT)   //point is to the right of clip rectangle
            {
                y = y0 + (y1 - y0) * (xmax - x0)/(x1 - x0);
                x = xmax;
            }
            else if (intOut & LEFT)                         //point is to the left of clip rectangle
            {
                y = y0 + (y1 - y0) * (xmin - x0)/(x1 - x0);
                x = xmin;
            }
            //Now we move outside point to intersection point to clip
            //and get ready for next pass.
            if (intOut == int0)
            {
                x0 = x;
                y0 = y;
                int0 = computeOutCode (x0, y0);
            }
            else
            {
                x1 = x;
                y1 = y;
                int1 = computeOutCode (x1, y1);
            }
        }
    }while (!done);

    if (accept)
    {
        //Following functions are left for implementation by user based on his platform(OpenGL/graphics.h etc.)
        return QLine(x0, y0, x1, y1);
    }
    return QLine(0, 0, 0, 0);

}


QLine  VesselSearch::clipLineAgainstRect(QPoint p, QVector2D dir, QRect window) {
    return this->clipLineAgainstRect(p, dir, window, false);
}

QLine VesselSearch::clipLineAgainstRect(QPoint p, QVector2D dir, QRect window, bool debug)
{
    debug = false;

    // Intersection left edge
    int leftY = -1;
    if (dir.x() != 0) {
        leftY = p.y() + (window.left() - p.x()) / dir.x() * dir.y();
        if  (leftY < window.top() || leftY > window.bottom()) {
            leftY = -1;
        }
    }

    // Intersection right edge
    int  rightY = -1;
    if (dir.x() != 0) {
        rightY = p.y() + (window.right() - p.x()) / dir.x() * dir.y();
        if  (rightY < window.top() || rightY > window.bottom()) {
            rightY = -1;
        }
    }

    // Intersection top edge
    int  topX = -1;
    if (dir.y() != 0) {
        topX = p.x() + (window.top() - p.y()) / dir.y() * dir.x();
        if (topX < window.left() || topX > window.right()) {
            topX = -1;
        }
    }

    // Intersection bottom edge
    int  bottomX = -1;
    if (dir.y() != 0) {
        bottomX = p.x() + (window.bottom() - p.y()) / dir.y() * dir.x();
        if (bottomX < window.left() || bottomX > window.right()) {
            bottomX = -1;
        }
    }

    // Find two intersections
    int found = 0;

    QPoint pt1, pt2;

    if (leftY != -1) {
        pt1 = QPoint(window.left(), leftY);
        found++;
    }

    if (rightY != -1) {
        if (found == 0) {
            pt1 = QPoint(window.right(), rightY);
        } else {
            pt2 = QPoint(window.right(), rightY);
        }
        found++;
    }

    if (topX != -1) {
        if (found == 0) {
            pt1 = QPoint(topX, window.top());
        } else {
            pt2 = QPoint(topX, window.top());
        }
        found++;
    }

    if (bottomX != -1) {
        if (found == 0) {
            pt1 = QPoint(bottomX, window.bottom());
        } else {
            pt2 = QPoint(bottomX, window.bottom());
        }
        found++;
    }

    QVector2D vec(pt2 - pt1);

    if (debug) {
        qDebug() << "direction is " << dir << "\t resulting vec is: " << vec;
    }

    if (QVector2D::dotProduct(vec.normalized(), dir.normalized()) > 0.8) {
        return QLine(pt1, pt2);
    }
    
    return QLine(pt2, pt1);
}

QLineF VesselSearch::clipLineAgainstRect(QPointF p, QVector2D dir, QRectF window) {
    return this->clipLineAgainstRect(p, dir, window, false);
}

/**
  * TODO: Does not direct the line properly, Fix as done in the integer variant above.
  */
QLineF VesselSearch::clipLineAgainstRect(QPointF p, QVector2D dir, QRectF window, bool debug)
{
    debug = false;
    // Intersection left edge
    float leftY = -1;
    if (dir.x() != 0) {
        leftY = p.y() + (window.left() - p.x()) / dir.x() * dir.y();

        if  (leftY < window.top() || leftY > window.bottom()) {
            leftY = -1;
        }
    }

    // Intersection right edge
    float rightY = -1;
    if (dir.x() != 0) {
        rightY = p.y() + (window.right() - p.x()) / dir.x() * dir.y();


        if  (rightY < window.top() || rightY > window.bottom()) {
            rightY = -1;
        }
    }

    // Intersection top edge
    float topX = -1;
    if (dir.y() != 0) {
        topX = p.x() + (window.top() - p.y()) / dir.y() * dir.x();


        if (topX < window.left() || topX > window.right()) {
            topX = -1;
        }
    }

    // Intersection bottom edge
    float bottomX = -1;
    if (dir.y() != 0) {
        bottomX = p.x() + (window.bottom() - p.y()) / dir.y() * dir.x();


        if (bottomX < window.left() || bottomX > window.right()) {
            bottomX = -1;
        }
    }

    // Find two intersections
    int found = 0;

    QPointF pt1, pt2;

    if (leftY != -1) {
        pt1 = QPointF(window.left(), leftY);
        found++;
    }

    if (rightY != -1) {
        if (found == 0) {
            pt1 = QPointF(window.right(), rightY);
        } else {
            pt2 = QPointF(window.right(), rightY);
        }
        found++;
    }

    if (topX != -1) {
        if (found == 0) {
            pt1 = QPointF(topX, window.top());
        } else {
            pt2 = QPointF(topX, window.top());
        }
        found++;
    }

    if (bottomX != -1) {
        if (found == 0) {
            pt1 = QPointF(bottomX, window.bottom());
        } else {
            pt2 = QPointF(bottomX, window.bottom());
        }
        found++;
    }

    QVector2D vec(pt2 - pt1);

    
    QVector2D diff = vec.normalized() - dir.normalized();

    if (debug) {
        qDebug() << "vec: " << vec.normalized();
        qDebug() << "dir: " << dir.normalized();
        qDebug() << "x fabs: " << fabs(diff.x()) << " x fabs: " << fabs(diff.y());
    }

    QLineF ret;

    if (fabs(diff.x()) < 0.01 && fabs(diff.y()) < 0.01) {
        if (debug) {
            qDebug() << "returning pt1 -> pt2";
        }
        ret = QLineF(pt1, pt2);
    }
    if (debug) {
        qDebug() << "returning pt2 -> pt1";
    }
    ret = QLineF(pt2, pt1);

    if (debug) {
        qDebug() << "ret is " << ret;
    }
    return ret;

}

