#ifndef VESSELSEARCH_H
#define VESSELSEARCH_H

#include <QObject>
#include <QPointF>
#include <QVector>
#include <QVector2D>
#include <QPolygon>
#include <QLine>
#include <QMutex>

#include <cv.h>
#include <highgui.h>

#include <houghsvg.h>

const int RIGHT = 8;  // 1000
const int TOP = 4;    // 0100
const int LEFT = 2;   // 0010
const int BOTTOM = 1; // 0001

class ImageProcessing;

struct Line {
    double rho;
    double theta;
    bool checked;
    int srcLineNum;
};

struct LineGroup {
    QVector <Line *> lines;
    float avgTheta;
    float avgOrigTheta;
    float matchRho;
};

class VesselSearch : public QObject
{
    Q_OBJECT;

public:
    VesselSearch(ImageProcessing * parent, QSize, int);
    ~VesselSearch();

    void setVideoSize(QSize videoSize);
    void setCameraMatrix(CvMat *);
    void setFrame(IplImage *);
    void setDebugLine(double, double);
    IplImage * getLineImage();
    void reset();
    QLine cohenSutherlandLineClip(QLine line);

    QLineF clipLineAgainstRect(QPointF p, QVector2D dir, QRectF window);
    QLineF clipLineAgainstRect(QPointF p, QVector2D dir, QRectF window, bool debug);

    QLine clipLineAgainstRect(QPoint p, QVector2D dir, QRect window);
    QLine clipLineAgainstRect(QPoint p, QVector2D dir, QRect window, bool debug);

    QVector <QPolygonF> getVesselPolygons();
    QVector <QPolygon> getVesselPolygonsImage();
    ImageProcessing * getParent();
    QPoint mapToImage(QPointF Point);
    QVector2D getForwardVec();
    QPointF getCenterLinePt();
    QPointF mapToXY(QPointF);
    QPolygon getVesselPolygon();
    CvMat * getLines();
    Line * getWorldLines();
    void setLinesVisible(bool visible);


private:
    HoughSVG * houghSvg;
    ImageProcessing * parent;
    IplImage * frame;
    IplImage * frameGray;
    IplImage * frameEdges;
    IplImage * frameEdgesRGB;
    IplImage * houghWorldDisplay;
    QSize frameSize;
    QRect frameWindow;
    CvMat * lines;
    Line * worldLines;
    int houghThres;
    int cannyParam1;
    int cannyParam2;
    CvMat * cameraMatrix;
    int houghWidth, houghHeight;
    QVector <LineGroup> lineGroups;
    float maxThetaDiff;
    float maxRhoDiff;
    int id;
    QString worldDisplayText;
    QString houghLinesText;
    float vesselWidth;
    float deviation;
    float step;
    float spacing;
    float searchStart;
    float searchEnd;
    int successiveMatches;
    LineGroup prevMatchGroup;
    float * templ;
    int templLength;
    QSize videoSize;
    bool inited;
    float minTheta;
    float maxTheta;
    float minRho;
    float maxRho;
    QVector <QPolygonF>  vesselPolygons;
    QPolygonF XYclip;
    QVector2D forwardVec;
    QPointF centerLinePt;
    QMutex dataMutex;
    QPolygon vesselPolygon;
    bool linesVisible;


    CvPoint convertToHough(QVector2D);
    void houghLines();
    void transformHoughLines(void);
    void groupLinesByAngle(void);
    double debugRho, debugTheta;
    void normalizeGroupByAngle(LineGroup *);
    float computeLineGroupMatch(LineGroup *, float &);
    float findBestLineGroupMatch(QVector <LineGroup>, int &, bool &);
    bool checkMatchConsistency(LineGroup lineGroup);


    QLine mapToImage(QLineF line);
    QLine mapToImageAndClip(QPointF p, QVector2D dir, QRect window);
    QLine mapToImageAndClip(QPointF p, QVector2D dir, QRect window, bool debug);
    QPointF lineIntersection(QPointF p1, QVector2D v1, QPointF p2, QVector2D v2);

    void createVesselPolygon(LineGroup lineGroup);
    int computeOutCode (double x, double y);
    void drawMatchLines(IplImage * frame, float rho, float theta, CvScalar color);


signals:
    void vesselPolygonFound(QPolygon);
    void vesselPolygonsFound(QVector <QPolygonF>);
    void vesselEdgesFound(QLine, QLine);

public slots:
    void setCannyParam1(int);
    void setCannyParam2(int);
    void setHoughThres(int);

};

#endif // VESSELSEARCH_H
