#include <QDebug>

#include "videodisplay.h"
#include "camera.h"
#include "imageprocessing.h"
#include "videoframe.h"
#include "videographicsitem.h"
#include "vesselsearch.h"



VideoDisplay::VideoDisplay(Camera* camera)
{
    this->camera = camera;
    videoFrame = new VideoFrame();
    videoFrameFiltered = new VideoFrame();

    textItem = new QGraphicsTextItem(QString("CAMERA ") + camera->getName(), this);
    textItem->setDefaultTextColor(QColor(255, 255, 255));
    textItem->setFont(QFont(QString("fixed"), 40));
    addToGroup(textItem);
    addToGroup(videoFrame);
    addToGroup(videoFrameFiltered);

    connect(videoFrame, SIGNAL(pointSelected(QPoint)), camera, SLOT(selectPoint(QPoint)));
    connect(videoFrame, SIGNAL(rectangleSelected(QRect)), camera, SLOT(rectangleSelected(QRect)));
    connect(videoFrame, SIGNAL(mouseMoved(QPoint)), camera, SLOT(setMousePosition(QPoint)));
    connect(camera, SIGNAL(cameraMatrixChanged(CvMat*)), videoFrame, SLOT(setCameraMatrix(CvMat*)));
    connect(camera->getImageProcessing(), SIGNAL(templateMatchAlongLineDone(QPolygon, QLine)), this, SLOT(setCorrespondenceVisuals(QPolygon, QLine)));
    connect(camera->getImageProcessing()->getVesselSearch(), SIGNAL(vesselEdgesFound(QLine, QLine)), this, SLOT(vesselFound(QLine, QLine)));
}

Camera * VideoDisplay::getCamera()
{
    return camera;
}

int VideoDisplay::updateFrames(int left)
{
    IplImage * frame = camera->getFrame();
    videoFrame->setFrameCV(frame);

    IplImage * filteredFrame = camera->getImageProcessing()->getFilteredFrameCV();
    videoFrameFiltered->setFrameCV(filteredFrame);

    videoFrame->setCrosshairPositions(camera->getImageProcessing()->getTargetPositions());
    videoFrame->setPolygon(camera->getImageProcessing()->getVesselPolygon());

    setPos(left, 0);

    videoFrame->setPos(0, textItem->boundingRect().bottom());
    videoFrameFiltered->setPos(0, textItem->boundingRect().bottom() + frame->height + 20);

    return videoFrame->boundingRect().width();
}

void VideoDisplay::vesselFound(QLine, QLine)
{
    videoFrame->videoItem->setPolygons(camera->getImageProcessing()->getVesselSearch()->getVesselPolygons());
}

void VideoDisplay::setCorrespondenceVisuals(QPolygon polygon, QLine line)
{
    videoFrame->videoItem->setPolygon2D(polygon);
    videoFrame->videoItem->setLine(line);
}

void VideoDisplay::mousePressEvent ( QGraphicsSceneMouseEvent * event )
{
    if (!videoFrame->contains(mapToItem(videoFrame, event->pos()))) {
        return;
    }
    videoFrame->pressEvent(event);
}

void VideoDisplay::mouseMoveEvent ( QGraphicsSceneMouseEvent * event )
{
    if (!videoFrame->contains(mapToItem(videoFrame, event->pos()))) {
        return;
    }
    videoFrame->moveEvent(event);
}

void VideoDisplay::mouseReleaseEvent(QGraphicsSceneMouseEvent * event )
{

    if (!videoFrame->contains(mapToItem(videoFrame, event->pos()))) {
        return;
    }
    videoFrame->releaseEvent(event);
}

QRectF VideoDisplay::boundingRect()
{
    return QRectF(0, 0, videoFrame->getVideoSize().width(), videoFrame->getVideoSize().height() * 2 + 20 + textItem->boundingRect().height());
}


VideoDisplay::~VideoDisplay()
{
    camera->disconnect(videoFrame, SIGNAL(pointSelected(QPoint)), camera, SLOT(pointSelected(QPoint)));
    camera->disconnect(videoFrame, SIGNAL(rectangleSelected(QRect)), camera->getImageProcessing(), SLOT(addTracker(QRect)));
    camera->disconnect(videoFrame, SIGNAL(mouseMoved(QPoint)), camera, SLOT(setMousePosition(QPoint)));
    delete videoFrame;
    delete videoFrameFiltered;
}

VideoFrame * VideoDisplay::getVideoFrameFiltered()
{
    return videoFrameFiltered;
}

VideoFrame * VideoDisplay::getVideoFrame()
{
    return videoFrame;
}
