#include "videofile.h"
#include <QDebug>
#include <QColor>

#include <cv.h>
#include <highgui.h>



VideoFile::VideoFile(QString filename)
{
    av_register_all();

    this->filename = filename;

    videoInitialized = false;

    videoStream = -1;
    pFrameBGR = NULL;
    pFrame = NULL;
    buffer = NULL;
    frameNo = 0;
    frameRotated = NULL;

}

VideoFile::~VideoFile()
{
    close();
}

bool VideoFile::init()
{
    if (videoInitialized) {
        close();
    }

    // Open video file
    if (av_open_input_file(&pFormatCtx, filename.toAscii().data(), NULL, 0, NULL)!=0) {
        qDebug() << "Unable to open file";
        return false;
    }
    if (av_find_stream_info(pFormatCtx) < 0) {
        qDebug() << "Couldn't find stream information";
        return false;
    }

    dump_format(pFormatCtx, 0, filename.toAscii().data(), 0);

    int i;

    // Find the first video stream
    videoStream = -1;
    for(i = 0; i < pFormatCtx->nb_streams; i++)
        if(pFormatCtx->streams[i]->codec->codec_type == CODEC_TYPE_VIDEO) {
            videoStream = i;
            qDebug() << "videoStream = " << i;
            break;
        }
    if (videoStream==-1) {
        qDebug() << "Didn't find a video stream";
        return false;
    }

    // Get a pointer to the codec context for the video stream
    pCodecCtx = pFormatCtx->streams[videoStream]->codec;

    AVCodec *pCodec;

    // Find the decoder for the video stream
    pCodec = avcodec_find_decoder(pCodecCtx->codec_id);
    if(pCodec==NULL) {
        qDebug() << "Codec not found";
        return false;
    }
    // Open codec
    if(avcodec_open(pCodecCtx, pCodec)<0) {
        qDebug() << "Could not open codec";
        return false;
    }

    // Allocate video frame
    pFrame = avcodec_alloc_frame();

    // Allocate an AVFrame structure
    pFrameBGR = avcodec_alloc_frame();

    if(pFrameBGR == NULL) {
        qDebug() << "Could not allocate an AVFrame structure";
        return false;
    }

    // Determine required buffer size and allocate buffer
    int numBytes = avpicture_get_size(PIX_FMT_RGB24, pCodecCtx->width, pCodecCtx->height);
    buffer = (uint8_t *) av_malloc(numBytes * sizeof(uint8_t));

    // Assign appropriate parts of buffer to image planes in pFrameBGR
    // Note that pFrameBGR is an AVFrame, but AVFrame is a superset
    // of AVPicture
    avpicture_fill((AVPicture *)pFrameBGR, buffer, PIX_FMT_RGB24,
                    pCodecCtx->width, pCodecCtx->height);

    frame = cvCreateImageHeader(cvSize(pCodecCtx->width, pCodecCtx->height), IPL_DEPTH_8U, 3);
    frame->imageData = (char *) pFrameBGR->data[0];

    // Software scaler context
    pSWSCtx = sws_getContext(pCodecCtx->width, pCodecCtx->height, pCodecCtx->pix_fmt, pCodecCtx->width, pCodecCtx->height, PIX_FMT_BGR24, SWS_FAST_BILINEAR, NULL, NULL, NULL);

    videoInitialized = true;

    return true;
}


bool VideoFile::captureFrame()
{
    if (!videoInitialized) {
        qDebug() << "Video not initliazed";
    }

    int frameFinished;
    AVPacket packet;

    while(av_read_frame(pFormatCtx, &packet) >= 0) {

        // Is this a packet from the video stream?
        if(packet.stream_index == videoStream) {

            avcodec_decode_video(pCodecCtx, pFrame, &frameFinished, packet.data, packet.size);

            if(frameFinished) {

                frameNo++;
                // Convert the image from its native format to BGR

                sws_scale(pSWSCtx, pFrame->data, pFrame->linesize, 0, pCodecCtx->height,  pFrameBGR->data, pFrameBGR->linesize);

                // Free the packet that was allocated by av_read_frame
                av_free_packet(&packet);
                return true;
            }
        }

        // Free the packet that was allocated by av_read_frame
        av_free_packet(&packet);
    }
    close();
    return false;
}

IplImage * VideoFile::getFrame()
{
    if (false) {
        if (frameRotated == NULL) {
            frameRotated = cvCreateImage(cvSize(frame->height, frame->width), IPL_DEPTH_8U, 3);
        }
        cvTranspose(frame, frameRotated);
        return frameRotated;
    }
    return frame;
}

void VideoFile::close()
{
    if (pFrame) {
        av_freep(&pFrame);
    }

    if (pFrameBGR) {
        av_freep(&pFrameBGR);
    }

    if (buffer) {
        av_freep(&buffer);
    }

    /* Fails if this is called. Unsufficient  documentation of swscaler to find out if
       it should be called or not. Source to memory leak. */
    //sws_freeContext(pSWSCtx);

    videoInitialized = false;
}

QSize VideoFile::getVideoSize()
{
    return QSize(frame->width, frame->height);
}

