#include <QDebug>
#include <QPaintEngine>
#include <QGraphicsScene>
#include <QGraphicsView>
#include <QVector2D>


#include "videographicsitem.h"

VideoGraphicsItem::VideoGraphicsItem(QGraphicsItem *parent = 0)
{
    glInited = false;
    newImageSize = true;
    frameSet = false;

    videoSize = QSize(0, 0);

    pixelFormat = GL_BGR;
    depth = 3;

    highlightSquareVisible = false;


    memset(projMat, 0, sizeof(GLfloat) * 16);
    memset(modelviewMat, 0, sizeof(GLfloat) * 16);

    squares = QSize(14, 5);
    squareSize = 0.027;

    deckCenter = QVector3D(0, 0, 0);
    deckAngles = QVector3D(0, 0, 0);

    deckVisible = false;
    highlightedSquare = QPoint(0, 0);

    gizmoVisible = true;

    polygon2DAlpha = 1.0;
    lineAlpha = 1.0;
    aftLineAlpha = 1.0;
}

void VideoGraphicsItem::setLine(QLine line)
{
    this->line = line;
    lineAlpha = 1.0;    
}



void VideoGraphicsItem::setGizmoVisible(bool visible)
{
    this->gizmoVisible = visible;
}


void VideoGraphicsItem::setDeckSize(QSize squares, float squareSize)
{
    this->squares = squares;
    this->squareSize = squareSize;
}

void VideoGraphicsItem::setHighlightedSquare(QPoint square, bool visible)
{
    highlightedSquare = square;
    highlightSquareVisible = visible;
}

void VideoGraphicsItem::setHighlightedSquareColor(QColor color)
{
    highlightedColor = color;
}

void VideoGraphicsItem::setPositions(QVector <QVector3D> positions)
{
    this->positions = positions;
}

void VideoGraphicsItem::drawGizmo()
{
    float gizSize = 0.06f;
    float radius = gizSize / 10;
    int resolution = 10;

    glDisable(GL_TEXTURE);
    glLineWidth(gizSize * 50);

    //RED
    glColor3f(1,0,0);
    glBegin(GL_LINES);
    glVertex3f(0,0,0);
    glVertex3f(gizSize,0,0);
    glEnd();

    glPushMatrix();
    glTranslatef(gizSize, 0, 0);
    glRotatef(90.0f, 0, 1, 0);
    glBegin(GL_TRIANGLE_STRIP);
    for(int i = 0; i < resolution; i++) {
        glVertex3f(0, 0, radius * 2);
        glVertex3f(radius * cos(i * 2 * M_PI / resolution), radius * sin(i * 2 * M_PI / resolution), 0.0);
        glVertex3f(radius * cos((i + 1) * 2 * M_PI / resolution), radius * sin((i + 1) * 2 * M_PI / resolution ), 0.0);
    }
    glEnd();
    glPopMatrix();


    //GREEN
    glColor3f(0,1,0);
    glBegin(GL_LINES);
    glVertex3f(0,0,0);
    glVertex3f(0,gizSize,0);
    glEnd();

    glPushMatrix();
    glTranslatef(0, gizSize, 0);
    glRotatef(-90.0f, 1, 0, 0);
    glBegin(GL_TRIANGLE_STRIP);
    for(int i = 0; i < resolution; i++) {
        glVertex3f(0, 0, radius * 2);
        glVertex3f(radius * cos(i * 2 * M_PI / resolution), radius * sin(i * 2 * M_PI / resolution), 0.0);
        glVertex3f(radius * cos((i + 1) * 2 * M_PI / resolution), radius * sin((i + 1) * 2 * M_PI / resolution ), 0.0);
    }
    glEnd();
    glPopMatrix();

    //BLUE
    glColor3f(0,0,1);
    glBegin(GL_LINES);
    glVertex3f(0,0,0);
    glVertex3f(0,0,gizSize);
    glEnd();

    glPushMatrix();
    glTranslatef(0, 0, gizSize);
    glBegin(GL_TRIANGLE_STRIP);
    for(int i = 0; i < resolution; i++) {
        glVertex3f(0, 0, radius * 2);
        glVertex3f(radius * cos(i * 2 * M_PI / resolution), radius * sin(i * 2 * M_PI / resolution), 0.0);
        glVertex3f(radius * cos((i + 1) * 2 * M_PI / resolution), radius * sin((i + 1) * 2 * M_PI / resolution ), 0.0);
    }
    glEnd();
    glPopMatrix();

    glEnable(GL_TEXTURE);
    glLineWidth(1);
}

void VideoGraphicsItem::drawNormalVector()
{
    glColor4f(1, 1, 1, 1);
    glLineWidth(5);
    glBegin(GL_LINES);
    glVertex3f(deckCenter.x(), deckCenter.y(), deckCenter.z());
    QVector3D normalVector(this->normalVector.x(), this->normalVector.y(), this->normalVector.z());
    normalVector *= 0.1;
    glVertex3f(deckCenter.x() + normalVector.x(), deckCenter.y() + normalVector.y(), deckCenter.z() + normalVector.z());
    glEnd();
    glLineWidth(1);
}

void VideoGraphicsItem::drawPoint(QVector3D point)
{
    glPushMatrix();
    glColor4f(1, 0.6, 0.6, 1);
    glTranslatef(point.x(), point.y(), point.z());

    glLineWidth(4);

    glBegin(GL_LINES);
    glVertex3f(-0.005, 0, 0);
    glVertex3f(0.005, 0, 0);

    glVertex3f(0, -0.005, 0);
    glVertex3f(0, 0.005, 0);

    glVertex3f(0, 0, -0.005);
    glVertex3f(0, 0, 0.005);
    glEnd();


    glLineWidth(1);
    glPopMatrix();
}

/**
  * Draws the video texture using an OpenGL Pixel Buffer Object
  */
void VideoGraphicsItem::drawVideo()
{
    if (!frameSet) {
        return;
    }

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    if (!glInited) {
        glGenTextures(1, &texture);
        glBindTexture(GL_TEXTURE_2D, texture);

        // Various parameres regarding texture scaling
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
        glBindTexture(GL_TEXTURE_2D, 0);
        glPixelStorei(GL_UNPACK_ALIGNMENT, depth);

        glInited = true;

        /* glGenBuffersARB = (PFNGLGENBUFFERSARBPROC)wglGetProcAddress("glGenBuffersARB");
        glBindBufferARB = (PFNGLBINDBUFFERARBPROC)wglGetProcAddress("glBindBufferARB");
        glBufferDataARB = (PFNGLBUFFERDATAARBPROC)wglGetProcAddress("glBufferDataARB");
        glBufferSubDataARB = (PFNGLBUFFERSUBDATAARBPROC)wglGetProcAddress("glBufferSubDataARB");
        glDeleteBuffersARB = (PFNGLDELETEBUFFERSARBPROC)wglGetProcAddress("glDeleteBuffersARB");
        glGetBufferParameterivARB = (PFNGLGETBUFFERPARAMETERIVARBPROC)wglGetProcAddress("glGetBufferParameterivARB");
        glMapBufferARB = (PFNGLMAPBUFFERARBPROC)wglGetProcAddress("glMapBufferARB");
        glUnmapBufferARB = (PFNGLUNMAPBUFFERARBPROC)wglGetProcAddress("glUnmapBufferARB"); */

    }

    // Checks if the frame size has changed since last draw
    if (newImageSize) {
        if (buffer != 0) {
            glDeleteBuffersARB(1, &buffer);
        }

        glGenBuffersARB(1, &buffer);

        // Allocate the buffer size according to the video frame
        glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, buffer);
        glBufferDataARB(GL_PIXEL_UNPACK_BUFFER_ARB, videoSize.width() * videoSize.height() * depth, NULL, GL_STREAM_DRAW_ARB);
        glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, 0);

        // Sets the texture size same as the video frame size
        glBindTexture(GL_TEXTURE_2D, texture);
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, videoSize.width(), videoSize.height(), 0, pixelFormat, GL_UNSIGNED_BYTE, (GLvoid*) NULL);
        glBindTexture(GL_TEXTURE_2D, 0);

        newImageSize = false;
    }


    glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, buffer);
    GLubyte * ptr = (GLubyte*) glMapBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, GL_WRITE_ONLY_ARB);

    if(ptr) {
        memcpy(ptr, frame->imageData, videoSize.width() * videoSize.height() * depth);

        // The following code must be used on some OpenGL implementations.
        // Set depth = 4 and glFormat = GL_BGRA an replace the above memcpy with the
        // following for-loop.

        /*
        int length = videoSize.width() * videoSize.height();
        int bufPos = 0;
        int framePos = 0;

        for(int i = 0; i < length; i++) {
            bufPos += depth;
            framePos += 3;
            ptr[bufPos] = frame->imageData[framePos ];
            ptr[bufPos + 1] = frame->imageData[framePos + 1];
            ptr[bufPos + 2] = frame->imageData[framePos + 2];
            ptr[bufPos + 3] = 255;
        }
        */
        glUnmapBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB); // release pointer to mapping buffer
    }

    // Copy pixels from PBO to texture object
    glBindTexture(GL_TEXTURE_2D, texture);
    glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, videoSize.width(), videoSize.height(), pixelFormat, GL_UNSIGNED_BYTE, 0);
    glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, 0);

    glColor4f(1, 1, 1, 1);
    glBegin(GL_QUADS);
    glNormal3f(0, 0, 1);
    glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f, 1.0f, 0.0f);
    glTexCoord2f(1.0f, 0.0f); glVertex3f(1.0f, 1.0f, 0.0f);
    glTexCoord2f(1.0f, 1.0f); glVertex3f(1.0f,  -1.0f, 0.0f);
    glTexCoord2f(0.0f, 1.0f); glVertex3f(-1.0f,  -1.0f, 0.0f);
    glEnd();
    glDisable(GL_TEXTURE_2D);
}

/**
  * Draw line
  */
void VideoGraphicsItem::drawLine()
{
    QPointF p1((float) line.p1().x() * 2 / videoSize.width() - 1, 1 - (float) line.p1().y() * 2 / videoSize.height());
    QPointF p2((float) line.p2().x() * 2 / videoSize.width() - 1, 1 - (float) line.p2().y() * 2 / videoSize.height());

    glLineWidth(3.0f);

    glColor4f(1.0, 1.0, 0.0, lineAlpha);
    glBegin(GL_LINES);
    glVertex3f(p1.x(), p1.y(), 0);
    glVertex3f(p2.x(), p2.y(), 0);
    glEnd();

    lineAlpha *= 0.95f;
    aftLineAlpha *= 0.95f;
}

void VideoGraphicsItem::drawPolygon2D()
{
    glBegin(GL_QUADS);
    glColor4f(1.0, 0.0, 0.0, polygon2DAlpha);
    for(int i = 0; i < polygon2D.size(); i++) {
        QPointF p(polygon2D.at(i).x() * 2.0f / videoSize.width() - 1.0f, 1.0f - polygon2D.at(i).y() * 2.0f / videoSize.height());
        glVertex3f(p.x(), p.y(), 0.0f);
    }

    if (polygon2D.size() > 0) {
        QPointF p(polygon2D.first().x() * 2.0f / videoSize.width() - 1.0f, 1.0f - polygon2D.first().y() * 2.0f / videoSize.height());
        glVertex3f(p.x(), p.y(), 0.0f);
    }

    glEnd();

    polygon2DAlpha *= 0.95f;
}

/**
  * Draws 2D polygons onto the XY plane.
  */
void VideoGraphicsItem::drawPolygons()
{
    float color = 0;
    for(int i = 0; i < polygons.size(); i++) {
        glBegin(GL_QUADS);
        if (i % 2 == 0) {
            glColor4f(1.0, 0.0, 0, polygonAlpha);
        } else {
            glColor4f(0.0, 1.0, 0, polygonAlpha);
        }
        QPolygonF poly;
        poly = polygons.at(i);
        for(int j = 0; j < poly.size(); j++) {
            glVertex3f(poly.at(j).x(), poly.at(j).y(), 0.0f);
        }
        glEnd();

        glBegin(GL_LINES);
        glVertex3f(poly.at(0).x(), poly.at(0).y(), 0.0f);
        for(int j = 1; j < poly.size(); j++) {
            glVertex3f(poly.at(j).x(), poly.at(j).y(), 0.0f);
        }
        glVertex3f(poly.at(0).x(), poly.at(0).y(), 0.0f);
        glEnd();
    }
    polygonAlpha *= 0.95f;
}

void VideoGraphicsItem::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget)
{

    painter->beginNativePainting();

    glEnable(GL_TEXTURE_2D);
    glEnable (GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    QPolygon test = scene()->views().at(0)->mapFromScene(mapToScene(this->boundingRect()));
    QRect sceneRect(test.at(0), test.at(2));

    float sceneHeight = scene()->views().at(0)->viewport()->height();
    float distanceFromBottom =  sceneHeight - (sceneRect.top() + sceneRect.height());

    glViewport(sceneRect.left(), distanceFromBottom, sceneRect.width(), sceneRect.height());



    /* glInfo * glinfo = new glInfo();
    glinfo->getInfo();
    glinfo->printSelf();
    delete glinfo; */
    //char * str = (char*) glGetString(GL_EXTENSIONS);
    //qDebug() << "extsize: " << strlen(str);


    drawVideo();

    //painter->drawImage(QRect(0, 0, frame->width, frame->height), qImage);

    drawPolygon2D();
    drawLine();

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glMatrixMode(GL_PROJECTION);

    glLoadMatrixf(projMat);

    if (gizmoVisible) {
        drawGizmo();
    }


    glPushMatrix();

    for(int i = 0; i < positions.size(); i++) {
        drawPoint(positions.at(i));
    }
    glTranslatef(deckCenter.x(), deckCenter.y(), deckCenter.z());

    glRotatef(deckAngles.z() * 180 / M_PI, 0, 0, 1);
    glRotatef(deckAngles.y() * 180 / M_PI, 0, 1, 0);
    glRotatef(deckAngles.x() * 180 / M_PI, 1, 0, 0);


    // START render in body coordinates
    if (gizmoVisible) {
        drawGizmo();
    }

    if (deckVisible) {
        drawDeck(squares.height(), squares.width(), squareSize);
    }

    // END render in body coordinates
    glPopMatrix();

    drawPolygons();


    painter->endNativePainting();
}

void VideoGraphicsItem::setDeckVisible(bool visible)
{
    deckVisible = visible;
}

void VideoGraphicsItem::setDeck(QVector3D pos, QVector3D ang)
{
    deckCenter = pos;
    deckAngles = ang;
}

void VideoGraphicsItem::setNormalVector(QVector3D vec)
{
    normalVector = vec;
}

void VideoGraphicsItem::drawDeck(int squaresX, int squaresY, float squareSize)
{
    float halfLength = squaresX * squareSize /  2;
    float halfWidth = squaresY * squareSize /  2;

    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glShadeModel(GL_SMOOTH);
    glEnable(GL_COLOR_MATERIAL);

    glLineWidth(2.5);
    GLfloat light_direction[] = {0.0, 0.0, 1.0};
    GLfloat light_position[] = {0.0, 0.0, -0.5, 0.0};

    glShadeModel (GL_SMOOTH);

    glLightfv(GL_LIGHT0, GL_POSITION, light_position);
    glLightfv(GL_LIGHT0, GL_SPOT_DIRECTION, light_direction);
    glLightf(GL_LIGHT0, GL_SPOT_CUTOFF, 120); // angle is 0 to 180
    glLightf(GL_LIGHT0, GL_SPOT_EXPONENT, 64); // exponent is 0 to 128


    glEnable (GL_LINE_SMOOTH);
    glEnable (GL_BLEND);
    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glHint (GL_LINE_SMOOTH_HINT, GL_DONT_CARE);

    // Grid
    float xMin = -halfLength;
    float xMax = halfLength;
    float xStep = squareSize;

    float yMin = -halfWidth;
    float yMax = halfWidth;
    float yStep = squareSize;

    glColor4f(0.0, 0.0, 0.0, 0.8);

    glBegin(GL_LINES);
    for(float i = xMin; i < xMax; i += xStep) {
        glVertex3f(i, yMin, 0);
        glVertex3f(i, yMax, 0);
    }
    for(float i = yMin; i < yMax; i += yStep) {
        glVertex3f(xMin, i, 0);
        glVertex3f(xMax, i, 0);
    }
    glEnd();

    // Main plane
    glColor4f(1, 0.2, 0.2, 0.36);

    glBegin(GL_QUADS);
    glNormal3f(0, 0, -1);
    glVertex3f(-halfLength, -halfWidth, 0);
    glVertex3f(-halfLength, halfWidth, 0);
    glVertex3f(halfLength, halfWidth, 0);
    glVertex3f(halfLength, -halfWidth, 0);
    glEnd();

    // Rear edge
    glBegin(GL_QUADS);
    glNormal3f(-1, 0, 0);
    glVertex3f(-halfLength, -halfWidth, -0.002f);
    glVertex3f(-halfLength, halfWidth, -0.002f);
    glVertex3f(-halfLength, halfWidth, 0.002f);
    glVertex3f(-halfLength, -halfWidth, 0.002f);
    glEnd();

    // Front edge
    glBegin(GL_QUADS);
    glNormal3f(1, 0, 0);
    glVertex3f(halfLength, -halfWidth, -0.002f);
    glVertex3f(halfLength, halfWidth, -0.002f);
    glVertex3f(halfLength, halfWidth, 0.002f);
    glVertex3f(halfLength, -halfWidth, 0.002f);
    glEnd();

    // Left edge
    glBegin(GL_QUADS);
    glNormal3f(0, -1, 0);
    glVertex3f(halfLength, -halfWidth, -0.002f);
    glVertex3f(halfLength, -halfWidth, 0.002f);
    glVertex3f(-halfLength, -halfWidth, 0.002f);
    glVertex3f(-halfLength, -halfWidth, -0.002f);
    glEnd();

    // Right edge
    glBegin(GL_QUADS);
    glNormal3f(0, 1, 0);
    glVertex3f(halfLength, halfWidth, -0.002f);
    glVertex3f(halfLength, halfWidth, 0.002f);
    glVertex3f(-halfLength, halfWidth, 0.002f);
    glVertex3f(-halfLength, halfWidth, -0.002f);
    glEnd();

    // Highlight square
    if (highlightSquareVisible) {
        glColor4f(highlightedColor.redF(), highlightedColor.greenF(), highlightedColor.blueF(), 0.6);
        glLineWidth(1);

        QPointF upperLeft( (-highlightedSquare.y() + squaresX / 2.0f) * squareSize, (highlightedSquare.x() - squaresY / 2.0f ) * squareSize);

        glBegin(GL_QUADS);
        glNormal3f(0, 0, -1);
        glVertex3f(upperLeft.x(), upperLeft.y(), 0);
        glVertex3f(upperLeft.x() - squareSize, upperLeft.y(), 0);
        glVertex3f(upperLeft.x() - squareSize, upperLeft.y() + squareSize, 0);
        glVertex3f(upperLeft.x(),  upperLeft.y() + squareSize, 0);
        glEnd();

        glColor4f(highlightedColor.redF(), highlightedColor.greenF(), highlightedColor.blueF(), 0.8);
        glBegin(GL_LINES);
        glVertex3f(upperLeft.x(), upperLeft.y(), 0);
        glVertex3f(upperLeft.x() - squareSize, upperLeft.y(), 0);

        glVertex3f(upperLeft.x() - squareSize, upperLeft.y(),0);
        glVertex3f(upperLeft.x() - squareSize, upperLeft.y() + squareSize,0);

        glVertex3f(upperLeft.x() - squareSize, upperLeft.y() + squareSize, 0);
        glVertex3f(upperLeft.x(),  upperLeft.y() + squareSize, 0);

        glVertex3f(upperLeft.x(),  upperLeft.y() + squareSize, 0);
        glVertex3f(upperLeft.x(), upperLeft.y(), 0);
        glEnd();
    }

    glDisable(GL_LIGHTING);
    glDisable(GL_LIGHT0);
    glDisable(GL_COLOR_MATERIAL);
}



void VideoGraphicsItem::setCameraMatrix(CvMat * m)
{

    cparam[0][0] = cvGetReal2D(m, 0, 0);
    cparam[1][0] = cvGetReal2D(m, 1, 0);
    cparam[2][0] = cvGetReal2D(m, 2, 0);

    cparam[0][1] = cvGetReal2D(m, 0, 1);
    cparam[1][1] = cvGetReal2D(m, 1, 1);
    cparam[2][1] = cvGetReal2D(m, 2, 1);

    cparam[0][2] = cvGetReal2D(m, 0, 2);
    cparam[1][2] = cvGetReal2D(m, 1, 2);
    cparam[2][2] = cvGetReal2D(m, 2, 2);

    cparam[0][3] = cvGetReal2D(m, 0, 3);
    cparam[1][3] = cvGetReal2D(m, 1, 3);
    cparam[2][3] = cvGetReal2D(m, 2, 3);

    computeProjMat();
}

void VideoGraphicsItem::computeProjMat()
{

    /*CvMat * intrinsic = cvCreateMat(3, 3, CV_32F);
    CvMat * rotMat = cvCreateMat(3, 3, CV_32F);
    CvMat * trans = cvCreateMat(4, 1, CV_32F);

    cvDecomposeProjectionMatrix(m, intrinsic, rotMat, trans);

    modelviewMat[0]  = cvGetReal2D(rotMat, 0, 0);
    modelviewMat[1]  = cvGetReal2D(rotMat, 1, 0);
    modelviewMat[2]  = cvGetReal2D(rotMat, 2, 0);
    modelviewMat[3]  = 0;

    modelviewMat[4]  = cvGetReal2D(rotMat, 0, 1);
    modelviewMat[5]  = cvGetReal2D(rotMat, 1, 1);
    modelviewMat[6]  = cvGetReal2D(rotMat, 2, 1);
    modelviewMat[4]  = 0;

    modelviewMat[8]  = cvGetReal2D(rotMat, 0, 2);
    modelviewMat[9]  = cvGetReal2D(rotMat, 1, 2);
    modelviewMat[10]  = cvGetReal2D(rotMat, 2, 2);
    modelviewMat[11]  = 0;

    modelviewMat[12]  = cvGetReal2D(trans, 0, 0);
    modelviewMat[13]  = cvGetReal2D(trans, 1, 0);
    modelviewMat[14]  = cvGetReal2D(trans, 2, 0);
    modelviewMat[15]  = 1;

    printMatrix(rotMat);
    printMatrix(trans);
    cvReleaseMat(&intrinsic);
    cvReleaseMat(&rotMat);
    cvReleaseMat(&trans);

    return;*/
    /* float div = cvGetReal2D(intrinsic, 2, 2);
    cvSetReal2D(intrinsic, 0, 0, cvGetReal2D(intrinsic, 0, 0) / div);
    cvSetReal2D(intrinsic, 0, 1, cvGetReal2D(intrinsic, 0, 1) / div);
    cvSetReal2D(intrinsic, 0, 2, cvGetReal2D(intrinsic, 0, 2) / div);

    cvSetReal2D(intrinsic, 1, 1, cvGetReal2D(intrinsic, 1, 1) / div);
    cvSetReal2D(intrinsic, 1, 2, cvGetReal2D(intrinsic, 1, 2) / div);
    cvSetReal2D(intrinsic, 2, 2, cvGetReal2D(intrinsic, 2, 2) / div);*/


    /*this->printMatrix(intrinsic);
    this->printMatrix(rotMatr);
    this->printMatrix(trans);*/



    double glMat[16];

    /*cparam[0][2] *= -1.0;
    cparam[1][2] *= -1.0;
    cparam[2][2] *= -1.0;*/

    argConvGLcpara2(cparam, videoSize.width(), videoSize.height(), 0.1f, 3.0f, glMat);

    //for(int i = 0; i < 16; i++) {

    //}

    projMat[0] = glMat[0];
    projMat[1] = -glMat[1];
    projMat[2] = glMat[2];
    projMat[3] = glMat[3];

    projMat[4] = glMat[4];
    projMat[5] = -glMat[5];
    projMat[6] = glMat[6];
    projMat[7] = glMat[7];

    projMat[8] = glMat[8];
    projMat[9] = -glMat[9];
    projMat[10] = glMat[10];
    projMat[11] = glMat[11];

    projMat[12] = glMat[12];
    projMat[13] = -glMat[13];
    projMat[14] = glMat[14];
    projMat[15] = glMat[15];


    /*qDebug() << "width: " << width << " height: " << height;
    projMat[0] = cvGetReal2D(m, 0, 0) / width;
    projMat[4] = cvGetReal2D(m, 0, 1) / width;
    projMat[8] = cvGetReal2D(m, 0, 2) / width;
    projMat[12] = (cvGetReal2D(m, 0, 3) - width / 2)/ width;

    projMat[1] = cvGetReal2D(m, 1, 0) / height;
    projMat[5] = cvGetReal2D(m, 1, 1) / height;
    projMat[9] = cvGetReal2D(m, 1, 2) / height;
    projMat[13] = (cvGetReal2D(m, 1, 3) - height / 2) / height;

    projMat[2] = cvGetReal2D(m, 2, 0);
    projMat[6] = cvGetReal2D(m, 2, 1);
    projMat[10] = cvGetReal2D(m, 2, 2);
    projMat[14] = cvGetReal2D(m, 2, 3);

    projMat[3] = 0;
    projMat[7] = 0;
    projMat[11] = -4;
    projMat[15] = 0;*/



    /*projMat[0] = 1;
    projMat[4] = 0;
    projMat[8] = 0;
    projMat[12] = 0;

    projMat[1] = 0;
    projMat[5] = 1;
    projMat[9] = 0;
    projMat[13] = 0;

    projMat[2] = 0;
    projMat[6] = 0;
    projMat[10] = 1;
    projMat[14] = 0;

    projMat[3] = 0;
    projMat[7] = 0;
    projMat[11] = 0;
    projMat[15] = 0;*/

    /*qDebug() << "projMat is now: ";
    qDebug() << projMat[0] << "\t" << projMat[4] << "\t" << projMat[8] << "\t" << projMat[12] << "\n" <<
                projMat[1] << "\t" << projMat[5] << "\t" << projMat[9] << "\t" << projMat[13] << "\n" <<
                projMat[2] << "\t" << projMat[6] << "\t" << projMat[10] << "\t" << projMat[14] << "\n" <<
                projMat[3] << "\t" << projMat[7] << "\t" << projMat[11] << "\t" << projMat[15];*/



}

void VideoGraphicsItem::setPolygon2D(QPolygon polygon)
{
    polygon2D = polygon;
    polygon2DAlpha = 0.5;
}

void VideoGraphicsItem::setPolygons(QVector <QPolygonF> polygons)
{
    this->polygons = polygons;
    polygonAlpha = 0.35;
}

void VideoGraphicsItem::setFrame(IplImage * frame)
{
    if (frame == NULL) {
        frameSet = false;
        return;
    }

    if (frame->width != videoSize.width() || frame->height != videoSize.height()) {
        newImageSize = true;
        videoSize = QSize(frame->width, frame->height);
        computeProjMat();
    }

    this->frame = frame;
    frameSet = true;
    update();
}

QRectF VideoGraphicsItem::boundingRect() const
{
    return QRect(0, 0, videoSize.width(), videoSize.height());
}

void VideoGraphicsItem::printMatrix(CvMat * L)
{
    int rows = L->rows;
    int cols = L->cols;

    QString output;
    output.append("[");
    for(int i = 0; i < rows; i++) {
        for(int j = 0; j < cols; j++) {
            QString num;
            num.setNum(cvGetReal2D(L, i, j));
            output.append(num);
            output.append("\t");
        }
        output.append(";\n");
    }
    output.append("]");
    qDebug() << output;
}


int VideoGraphicsItem::arParamDecompMat( double source[3][4], double cpara[3][4], double trans[3][4] )
{
    int       r, c;
    double    Cpara[3][4];
    double    rem1, rem2, rem3;

    if( source[2][3] >= 0 ) {
        for( r = 0; r < 3; r++ ){
            for( c = 0; c < 4; c++ ){
                Cpara[r][c] = source[r][c];
            }
        }
    }
    else {
        for( r = 0; r < 3; r++ ){
            for( c = 0; c < 4; c++ ){
                Cpara[r][c] = -(source[r][c]);
            }
        }
    }


    for( r = 0; r < 3; r++ ){
        for( c = 0; c < 4; c++ ){
            cpara[r][c] = 0.0;
        }
    }

    cpara[2][2] = norm( Cpara[2][0], Cpara[2][1], Cpara[2][2] );
    trans[2][0] = Cpara[2][0] / cpara[2][2];
    trans[2][1] = Cpara[2][1] / cpara[2][2];
    trans[2][2] = Cpara[2][2] / cpara[2][2];
    trans[2][3] = Cpara[2][3] / cpara[2][2];

    cpara[1][2] = dot( trans[2][0], trans[2][1], trans[2][2],
                       Cpara[1][0], Cpara[1][1], Cpara[1][2] );
    rem1 = Cpara[1][0] - cpara[1][2] * trans[2][0];
    rem2 = Cpara[1][1] - cpara[1][2] * trans[2][1];
    rem3 = Cpara[1][2] - cpara[1][2] * trans[2][2];
    cpara[1][1] = norm( rem1, rem2, rem3 );
    trans[1][0] = rem1 / cpara[1][1];
    trans[1][1] = rem2 / cpara[1][1];
    trans[1][2] = rem3 / cpara[1][1];

    cpara[0][2] = dot( trans[2][0], trans[2][1], trans[2][2],
                       Cpara[0][0], Cpara[0][1], Cpara[0][2] );
    cpara[0][1] = dot( trans[1][0], trans[1][1], trans[1][2],
                       Cpara[0][0], Cpara[0][1], Cpara[0][2] );
    rem1 = Cpara[0][0] - cpara[0][1]*trans[1][0] - cpara[0][2]*trans[2][0];
    rem2 = Cpara[0][1] - cpara[0][1]*trans[1][1] - cpara[0][2]*trans[2][1];
    rem3 = Cpara[0][2] - cpara[0][1]*trans[1][2] - cpara[0][2]*trans[2][2];
    cpara[0][0] = norm( rem1, rem2, rem3 );
    trans[0][0] = rem1 / cpara[0][0];
    trans[0][1] = rem2 / cpara[0][0];
    trans[0][2] = rem3 / cpara[0][0];

    trans[1][3] = (Cpara[1][3] - cpara[1][2]*trans[2][3]) / cpara[1][1];
    trans[0][3] = (Cpara[0][3] - cpara[0][1]*trans[1][3]
                               - cpara[0][2]*trans[2][3]) / cpara[0][0];

    for( r = 0; r < 3; r++ ){
        for( c = 0; c < 3; c++ ){
            cpara[r][c] /= cpara[2][2];
        }
    }

    return 0;
}

void VideoGraphicsItem::argConvGLcpara2( double cparam[3][4], int width, int height, double gnear, double gfar, double m[16] )
{
    double   icpara[3][4];
    double   trans[3][4];
    double   p[3][3], q[4][4];
    int      i, j;

    if( arParamDecompMat(cparam, icpara, trans) < 0 ) {
        printf("gConvGLcpara: Parameter error!!\n");
        exit(0);
    }

    for( i = 0; i < 3; i++ ) {
        for( j = 0; j < 3; j++ ) {
            p[i][j] = icpara[i][j] / icpara[2][2];
        }
    }
    q[0][0] = (2.0 * p[0][0] / width);
    q[0][1] = (2.0 * p[0][1] / width);
    q[0][2] = ((2.0 * p[0][2] / width)  - 1.0);
    q[0][3] = 0.0;

    q[1][0] = 0.0;
    q[1][1] = (2.0 * p[1][1] / height);
    q[1][2] = ((2.0 * p[1][2] / height) - 1.0);
    q[1][3] = 0.0;

    q[2][0] = 0.0;
    q[2][1] = 0.0;
    q[2][2] = (gfar + gnear)/(gfar - gnear);
    q[2][3] = -2.0 * gfar * gnear / (gfar - gnear);


    q[3][0] = 0.0;
    q[3][1] = 0.0;
    q[3][2] = 1.0;
    q[3][3] = 0.0;

    for( i = 0; i < 4; i++ ) {
        for( j = 0; j < 3; j++ ) {
            m[i+j*4] = q[i][0] * trans[0][j]
                     + q[i][1] * trans[1][j]
                     + q[i][2] * trans[2][j];
        }
        m[i+3*4] = q[i][0] * trans[0][3]
                 + q[i][1] * trans[1][3]
                 + q[i][2] * trans[2][3]
                 + q[i][3];
    }
}

double VideoGraphicsItem::norm( double a, double b, double c )
{
    return( sqrt( a*a + b*b + c*c ) );
}

double VideoGraphicsItem::dot( double a1, double a2, double a3, double b1, double b2, double b3 )
{
    return( a1 * b1 + a2 * b2 + a3 * b3 );
}
