#include "waveestimator.h"
#include <QDebug>

WaveEstimator::WaveEstimator()
{
    rotMatrix = cvCreateMat(3, 3, CV_32F);
    cvSetIdentity(rotMatrix);
    size = QSize(5, 11);
    squareSize = 0.034;
    timeConstantSet = 0.1;
    timeConstantReset = 0.001;
    allocate();
}

WaveEstimator::~WaveEstimator()
{
    delete[] zPositions;
    delete[] zPositionsMin;
    delete[] zPositionsMax;
    delete[] zAmplitude;
    cvReleaseMat(&rotMatrix);
}

float WaveEstimator::getSquareSize()
{
    return squareSize;
}

void WaveEstimator::allocate()
{
    zPositions = new float[size.width() * size.height()];
    memset(zPositions, 0, sizeof(float) * size.width() * size.height());

    zPositionsMin = new float[size.width() * size.height()];
    memset(zPositionsMin, 0, sizeof(float) * size.width() * size.height());

    zPositionsMax = new float[size.width() * size.height()];
    memset(zPositionsMax, 0, sizeof(float) * size.width() * size.height());

    zAmplitude = new float[size.width() * size.height()];
    memset(zAmplitude, 0, sizeof(float) * size.width() * size.height());
}

void WaveEstimator::clear()
{
    memset(zPositions, 0, sizeof(float) * size.width() * size.height());
    memset(zPositionsMin, 0, sizeof(float) * size.width() * size.height());
    memset(zPositionsMax, 0, sizeof(float) * size.width() * size.height());
    memset(zAmplitude, 0, sizeof(float) * size.width() * size.height());


}

void WaveEstimator::setPosition(QVector3D pos, CvMat * rotMatrix)
{
    position = pos;
    cvCopy(rotMatrix, this->rotMatrix);


    for(int i = -(size.height() - 1)/ 2; i <= (size.height() - 1) / 2; i++) {
        for(int j = -(size.width() - 1) / 2; j <= (size.width() - 1) / 2; j++) {

            QPointF planePos = getPlanePos(QPoint(j, i));
            float z = computeZ(planePos);


            int zPosX = j  + (size.width() - 1) / 2;
            int zPosY = i  + (size.height() - 1) / 2;
            int index = zPosX + zPosY * size.width();

            zPositions[index] = z;

            if (z < zPositionsMin[index]) {
                zPositionsMin[index] = z;
            }

            if (z > zPositionsMax[index ]) {
                zPositionsMax[index] = z;
            }

            lowpass(z, zPositionsMax[index], timeConstantReset);
            lowpass(z, zPositionsMin[index], timeConstantReset);

            lowpass(zPositionsMax[index] - zPositionsMin[index], zAmplitude[index], timeConstantSet);
        }
    }
}

/**
  * Updates "output" according to input. When t -> infinity =>  output = input
  */
void WaveEstimator::lowpass(float input, float &output, float timeconstant)
{
    output += (input - output) * timeconstant;
}

/**
  * Returns the z-positions of the plane and updates the derived data
  */
float * WaveEstimator::getZ()
{
    return zPositions;
}

/**
  * Returns the mean amplitude of the z-position
  */
float * WaveEstimator::getAmplitude()
{
    return zAmplitude;
}

float WaveEstimator::computeZ(QPointF planePos)
{
    // minus Y-screen is X-forward, X-screen is Y-starboard
    QVector3D plane(-planePos.y(), planePos.x(), 0);
    QVector3D pos = position + rotateVector(plane, rotMatrix);
    //QVector3D pos = rotateVector(plane, rotMatrix);
    return pos.z();
}

/**
  * Transforms a vector given a 3x3 rotation matrix. returns rot * vec.
  */
QVector3D WaveEstimator::rotateVector(QVector3D qVec, CvMat * rot)
{
    QVector3D res;
    res.setX( cvGetReal2D(rot, 0, 0) * qVec.x() + cvGetReal2D(rot, 0, 1) * qVec.y() + cvGetReal2D(rot, 0, 2) * qVec.z()) ;
    res.setY( cvGetReal2D(rot, 1, 0) * qVec.x() + cvGetReal2D(rot, 1, 1) * qVec.y() + cvGetReal2D(rot, 1, 2) * qVec.z());
    res.setZ( cvGetReal2D(rot, 2, 0) * qVec.x() + cvGetReal2D(rot, 2, 1) * qVec.y() + cvGetReal2D(rot, 2, 2) * qVec.z());
    return res;
}

QPointF WaveEstimator::getPlanePos(QPoint block)
{
    if (abs(block.x()) > (size.width() - 1)/ 2) {
        qDebug() << "block num outside range, max is: " << (size.width() - 1) / 2 << " was: " << block.x();
    }

    if (abs(block.y()) > (size.height() - 1)/ 2) {
        qDebug() << "block num outside range, max is: " << (size.height() - 1) / 2 << " was: " << block.y();
    }

    return QPointF(squareSize * block.x(), squareSize * block.y());
}

void WaveEstimator::setSize(QSize size)
{

    if ( size.width() % 2 == 0) {
        qDebug() << "width is even, must be odd";
        return;
    }

    if ( size.height() % 2 == 0) {
        qDebug() << "width is even, must be odd";
        return;
    }

    if (size != this->size) {
        delete[] zPositions;
        this->size = size;
        allocate();
    }
}

QSize WaveEstimator::getSize()
{
    return size;
}


void WaveEstimator::printMatrix(CvMat * L)
{
    int rows = L->rows;
    int cols = L->cols;

    QString output;
    output.append("[");
    for(int i = 0; i < rows; i++) {
        for(int j = 0; j < cols; j++) {
            QString num;
            num.setNum(cvGetReal2D(L, i, j));
            output.append(num);
            output.append("\t");
        }
        output.append(";\n");
    }
    output.append("]");
    qDebug() << output;
}
