#include "wavemap.h"
#include <QGraphicsScene>
#include <QDebug>

#include <math.h>

Wavemap::Wavemap()
{
    size = QSize(230, 300);
    legendWith = 0.2;
    cols = 1;
    rows = 1;
    scaleStart = 0;
    scaleEnd = 0.10;
    ticks = 10;
    spacing = 0.08;

    values = new float[cols * rows];
    for(int i = 0; i < rows; i++) {
        float posIdx = 1 - (float) i / (float) rows;
        for(int j = 0; j < cols; j++) {
            values[j + i * cols] = (scaleEnd - scaleStart) * posIdx + scaleStart;
        }
    }

    linearGradient = QLinearGradient (0, size.height(), 0, 0);
    linearGradient.setColorAt(0.0, Qt::darkBlue);
    linearGradient.setColorAt(0.3, Qt::cyan);
    linearGradient.setColorAt(0.6, Qt::yellow);
    linearGradient.setColorAt(1.0, Qt::red);
}

void Wavemap::setScaleFactor(float scale, QString unit)
{
    scaleFactor = scale;
    this->unit = unit;
    update();
}

void Wavemap::setWavemap(float * data, int cols, int rows)
{
    if (cols != this->cols || rows != this->rows) {
        this->cols = cols;
        this->rows = rows;
    
        delete[] values;
        values = new float[cols * rows];
    }

    float min = 1000;
    float max = -1000;

    memcpy(values, data, cols * rows * sizeof(float));

    for(int i = 0; i < cols * rows; i++) {
        if (values[i] < min) {
            min = values[i];
        }
        if (values[i] > max) {
            max = values[i];
        }
    }

    update();
}

/**
  * Finds the color for a given value based on the linearGradient colors
  */
QColor Wavemap::getGradientColor(float value)
{
    if (value < scaleStart) {
        value = scaleStart;
    }

    if (value > scaleEnd) {
        value = scaleEnd;
    }

    float pos = (value - scaleStart) / (scaleEnd - scaleStart);

    QGradientStops stops = linearGradient.stops();

    QGradientStop prevStop = stops.at(0);
    QGradientStop stop;

    for(int i = 0; i < stops.size(); i++) {
        stop = stops.at(i);
        if(stop.first > pos) {
            break;
        }
        prevStop = stop;
    }

    QColor from = prevStop.second;
    QColor to = stop.second;


    float normalizedPos = 0;

    if (stop.first - prevStop.first != 0) {
        normalizedPos = (pos - prevStop.first) / (stop.first - prevStop.first);
    }

    int r = from.red() * (1 - normalizedPos) + to.red() * normalizedPos;
    int g = from.green() * (1 - normalizedPos) + to.green() * normalizedPos;
    int b = from.blue() * (1 - normalizedPos) + to.blue() * normalizedPos;
    int a = from.alpha() * (1 - normalizedPos) + to.alpha() * normalizedPos;

    //qDebug() << " from: " << from << " to: " << to << " norpos: " << normalizedPos;
    //qDebug() << r << " " << g << " " << b << " " << a;
    return QColor(r, g, b, a);
}

void Wavemap::paint(QPainter *painter, const QStyleOptionGraphicsItem * option, QWidget * widget)
{
    painter->setRenderHint(QPainter::Antialiasing);
    painter->setBrush(Qt::NoBrush);
    painter->setPen(QPen(QColor(255, 255, 255)));

    int legendWidthPix = legendWith * size.width();
    int mapLeft = (legendWith + spacing) * size.width();
    int mapRight = spacing * size.width();


    float blockWidth = (size.width() - mapLeft - mapRight) / cols;
    float blockHeight = size.height() / rows;
    float blockSize = blockHeight;

    if (blockWidth < blockHeight) {
        blockSize  = blockWidth;
    }
    int mapTop = size.height() / 2 - blockSize * rows / 2;

    QBrush brush(linearGradient);

    QFont font(QString("arial"), 10);
    painter->setFont(font);
    QFontMetrics fontMetrics(font);
    painter->setBrush(brush);

    int scalePadStart = fontMetrics.height();
    int scalePadEnd = fontMetrics.height();

    // Gradient scale
    painter->drawRect(QRect(legendWidthPix * 4 / 5, scalePadStart , legendWidthPix * 1 / 5, size.height() - scalePadEnd - scalePadStart));
    for(int i = 0; i < ticks + 1; i++) {
        float posIdx = (float) i / (float) ticks;
        float posValue = (scaleEnd - scaleStart) * posIdx + scaleStart;
        int pos = scalePadStart + (1 - posIdx)  * (size.height() - scalePadEnd - scalePadStart);

        QString tickLabel;
        tickLabel.setNum(posValue * fabs(scaleFactor), 'f', 0);
        tickLabel.append(unit);
        painter->drawText(QRect(0, pos - fontMetrics.height() / 2, legendWidthPix * 3 / 5, size.height() / ticks), Qt::AlignRight, tickLabel);
        painter->drawLine(QPointF(legendWidthPix * 4 / 5, pos), QPointF(legendWidthPix * 4.2f / 5, pos));
        painter->drawLine(QPointF(legendWidthPix * 4.8 / 5, pos), QPointF(legendWidthPix, pos));
    }

    painter->setPen(QPen(QColor(0, 0, 0)));
    font.setPointSize(7);
    painter->setFont(font);

    // Squares
    for(int i = 0; i < rows; i++) {
        for(int j = 0; j < cols; j++) {
            QString num;
            num.setNum(values[j + i * cols] * fabs(scaleFactor), 'f', 2);
            QColor rectColor = getGradientColor(values[j + i * cols]);
            painter->setBrush(QBrush(rectColor));
            painter->setPen(QPen(QColor(Qt::black)));
            painter->drawRect(QRect(mapLeft + blockWidth * j, mapTop + blockSize * i, blockWidth, blockSize));
            int mean = (rectColor.red() + rectColor.green() + rectColor.blue()) / 3;

            /* QColor textColor(0, 0, 0);
            if (mean < 127) {
                textColor = QColor(255, 255, 255);
            }

            painter->setPen(QPen(textColor));
            painter->drawText(QRect(mapLeft + blockWidth * j, mapTop + blockSize * i, blockSize, blockSize), Qt::AlignCenter, num); */
        }
    }
}


QRectF Wavemap::boundingRect() const
{
    return QRectF(0, 0, size.width(), size.height());
}

